=begin
#------------------------------------------------------------------------------------------------------------
#************************************************************************************************************
# Copyright 2008 Fredo6 - Designed and written December 2008 by Fredo6
#
# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for commercial purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:   Lib6Core_40.rb
# Original Date	:   10 Sep 2008 - version 3.0
# Type			:   Ruby Library
# Description	:   Plugin Management for all Library utilities of Fredo6's scripts
# Note			:   This file is always loaded first
#				:	The rest of the code resides in other Lib6***.rb files, which are loaded afterward
#------------------------------------------------------------------------------------------------------------
#************************************************************************************************************
=end

require 'sketchup.rb'
require 'extensions.rb'

module Traductor
@@file_trace6 = nil

#-------------------------------------------------------------------------------------------------------------------------------			 				   
#-------------------------------------------------------------------------------------------------------------------------------			 				   
# T6  Language Translation Management (only  top routines) - Rest in LibTraductor_xx.rb 
#-------------------------------------------------------------------------------------------------------------------------------			 				   
#-------------------------------------------------------------------------------------------------------------------------------			 				   

T6MOD_LNG_DEFAULT = '--'			#Code for Language by Default
T6MOD_RKEY_Langpref = "langpref"	#Registry key for preferred languages
RUN_ON_MAC = (RUBY_PLATFORM =~ /darwin/i) ? true : false
SU_MAJOR_VERSION = (Sketchup.version[/(\d+)/]).to_i

#computing the current language
@@langdef = (defined?(TRADUCTOR_DEFAULT) && TRADUCTOR_DEFAULT.strip =~ /^\w\w$/) ? $&.upcase : Sketchup.get_locale[0..1]
@@lang = @@langdef
@@patlang = Regexp.new '\|' + @@lang + '\|', Regexp::IGNORECASE

def Traductor.get_language 
	@@lang
end	

def Traductor.set_language(lang=nil) 
	@@lang = (lang && lang.strip != "") ? lang[0..1]: @@langdef
	@@patlang = Regexp.new '\|' + @@lang + '\|', Regexp::IGNORECASE
end	

#---------------------------------------------------------------------------------------------------------------------------------
# Retina Factor
#---------------------------------------------------------------------------------------------------------------------------------

def Traductor.retina_factor
	@retina = MYDEFPARAM[:DEFAULT_PaletteRetina]
	@retina_factor = (@retina) ? 2.0 : 1.0
end


#Make sure the core libraries are loaded
def Traductor.use_libraries
	@hsh_loaded_libraries = {} unless @hsh_loaded_libraries
	ls = ["G6", "Translate", "Spline", "WebHtml", "Tool", "Algo", "Contour", "Offset3d", "Transform", "Mark"]
	ls.each { |libname| Traductor.use_single_library libname }
end

#Require a fix library of LibFredo6
def Traductor.use_single_library(libname)
	return if @hsh_loaded_libraries[libname]
	@hsh_loaded_libraries[libname] = true
	basename = "Lib6#{libname}.rb"
	body_name = "body_#{basename}"
	fname = File.join LibFredo6.folder, body_name
	t0 = Time.now
	begin
		require fname
		delta = Time.now - t0
		MYPLUGIN.add_load_time_second delta
		LibFredo6.log "#{basename}: Loading body code from #{body_name} [#{(delta * 1000).to_i} ms]"
	rescue Exception => e
		LibFredo6.log "?#{basename}: Error Loading body code from #{body_name}"
	end
end

class T6Mod

Traductor_T6Root = Struct.new("Traductor_T6Root", :rootname, :path, :hsh_t6, :t6edit) 

@@langpref = nil
@@hsh_t6 = {}
@@hsh_root = {}

#Set the Preferred languages (in registry)
def T6Mod.set_langpref(llang)
	llang = [] unless llang
	@@langpref = llang
	Traductor.set_language llang[0]
	ll = Sketchup.write_default "LibFredo6", T6MOD_RKEY_Langpref, llang.join(" ")
end

def T6Mod.init_langpref
	ll = Sketchup.read_default "LibFredo6", T6MOD_RKEY_Langpref
	if ll
		@@langpref = ll.split(" ")
	else
		@@langpref = [Sketchup.get_locale[0..1]]
	end	
	Traductor.set_language @@langpref[0]
end

#Create a language translation context for the module hmod
def initialize(hmod, path, rootname)
	#Default Languages
	T6Mod.init_langpref unless @@langpref
	
	#initialization
	@hmod = hmod
	@modname = hmod.name	
	@hstrings = {}
	@hlng = {}
	@path = path
	@rootname = rootname
	@hloaded = {}
	@@hsh_t6[@modname] = self
	
	#Storing the handler for its root and path
	hroot = @@hsh_root[rootname]
	unless hroot
		hroot = Traductor_T6Root.new
		hroot.hsh_t6 = {}
		hroot.rootname = rootname
		hroot.path = path
		hroot.t6edit = nil
		@@hsh_root[rootname] = hroot
	end	
	hroot.hsh_t6[@modname] = self
	self
end

def get_hstrings
	@hstrings
end
	
def T6Mod.get_langpref 
	@@langpref
end	

#Set and declare a symbolic string and its translation
def []=(symb, sval)
	self.set symb, sval
end

def set(symb, sval)
	return nil unless symb
	symb = symb.to_s.strip 
	sval = symb unless sval
		
	#Parsing the string specification, either a sring or a list of string
	lsval = (sval.class == Array) ? sval : [sval]
	lsval.each do |spec|
		rest = spec.strip
		lng = T6MOD_LNG_DEFAULT
		nodefault = false
		while (rest =~ /\|(\w\w)\|(.*)/)
			if $`.strip.length == 0
				nodefault = true
			else
				store_value lng, symb, $`.strip
				if nodefault
					store_value T6MOD_LNG_DEFAULT, symb, $`.strip
					nodefault = false
				end
			end	
			lng = $1
			rest = $2.strip
		end
		store_value lng, symb, rest
		store_value T6MOD_LNG_DEFAULT, symb, rest if nodefault
	end	
	
	true
end

#Get the translation of a symbol according to current preferred language
#the method works with Hash, Keylist and simple list of symbols
def [](symb, *args)
	self.get symb, *args
end

def get(symb, *args)
	#Symbol is actually from Traductor
	if @hmod != Traductor && symb.class == Symbol && (symb.id2name =~ /\AT_/ || symb.id2name =~ /\AT6_/)
		return Traductor::T6[symb, *args]
	end	
	
	#Hash table and [key, value] lists - Translate the symbols
	if symb.class == Hash
		hsh = {}
		symb.each { |key, val| hsh[key] = get(val, *args) }
		return hsh	
	elsif symb.class == Array && symb.length > 0 && symb[0].class == Array && symb[0].length == 2
		lst = []
		symb.each { |l| lst.push [l[0], get(l[1], *args)] }
		return lst	
	elsif symb.class == Array && symb.length > 0 && symb[0].class == Symbol
		lst = []
		symb.each { |l| lst.push get(l, *args) }
		return lst	
	end
	
	#Old style With Traductor
	return Traductor[symb, *args] unless symb.class == Symbol
	
	#Checking if translation file is loaded
	check_loaded
	
	#Retrieving the string
	symb = symb.id2name
	hsh = @hstrings[symb]
	return symb unless hsh
	
	#Retrieving the hash table
	sval = nil
	@@langpref.each do |lng|
		break if (sval = hsh[lng])
	end
	sval = hsh[T6MOD_LNG_DEFAULT] unless sval
	return symb unless sval
	
	#Performing argument substitution
	for i in 0..args.length-1 do
		sval = sval.gsub "%#{i+1}", args[i].to_s
	end
	
	#returning the value
	sval
end

#Technical storage of the string according to language
def store_value(lng, symb, sval)
	#removing the '~' for leading or trailing space preservation
	val = " " + $' if ((val =~ /\A~\s/) == 0)		# '~' followed by at least one space at beginning of string
	val = $` + " " if (val =~ /\s~\z/)				# '~' preceded by at least one space at end of string
	lng = lng.upcase
	
	#Storing the string
	symb = symb.to_s
	hln = @hstrings[symb]
	hln = (@hstrings[symb] = {}) unless hln
	@hlng[lng] = 0 unless @hlng[lng]
	@hlng[lng] += 1 unless hln[lng]
	sval = sval.gsub /\\n/, "\n" if sval
	hln[lng] = UTF.real(sval)
end

def T6Mod.handler(modulename)
	@@hsh_t6[modulename]
end

#Transfer of global value to avoid diacritic characters in Eval
def T6Mod.store_path_66(path) ; @plugin_dir_66 = path ; end
def T6Mod.store_rootname_66(rootname) ; @rootname_66 = rootname ; end
def T6Mod.get_path_66 ; @plugin_dir_66 ; end
def T6Mod.get_rootname_66 ; @rootname_66 ; end

#Enabling usage of T6 notation in order to call via T6[]= and T6[] in the module <hmod>
def T6Mod.enable(hmod, path, rootname)
	return if hmod.const_defined?('T6')
	
	#Using temporary global variables due to non-ascii characters
	T6Mod.store_path_66(path)
	T6Mod.store_rootname_66(rootname)
	
	#Creating the new instance for Language handling
	txt = "T6 = Traductor::T6Mod.new(#{hmod}, T6Mod.get_path_66, T6Mod.get_rootname_66) " 
	hmod.module_eval txt
end

def T6Mod.all_supported_languages
	llg = []
	@@hsh_t6.each { |mod, t6| llg += t6.supported_languages }
	llg.uniq
end

#Save a translation to file in a given language - if lang == nil, then all supported language)
def T6Mod.save_to_file(rootname, lang=nil, purge=false)
	hroot = @@hsh_root[rootname]
	return 0 unless hroot
	
	#checking if language supported
	if (lang)
		lst_lng = [lang]
	else
		lst_lng = []
		hroot.hsh_t6.each do |key, t6| 
			lst_lng += t6.supported_languages
		end	
		return 0 if lst_lng.length == 0
		lst_lng = lst_lng.uniq
		lst_lng.delete T6MOD_LNG_DEFAULT
	end	
	
	#Writing to file
	nstrings = 0
	lst_lng.each do |lng|
		file = File.join hroot.path, rootname + "_#{lng}.lang"
		File.open(file, "w") do |f| 
			f.puts T6[:T_WARNING_File]
			hroot.hsh_t6.each { |key, t6| nstrings += t6.write_to_file(f, lng, purge) }
		end
	end	
	
	#Propagating the laoded status
	hroot = @@hsh_root[rootname]
	hroot.hsh_t6.each { |key, t6| t6.set_loaded(lst_lng, false) }

	nstrings
end

#Load a file in a given language
def T6Mod.load_file(rootname, lang)
	lang = lang.upcase
	hroot = @@hsh_root[rootname]
	return 0 unless hroot

	#Checking file existence
	file = File.join hroot.path, rootname + "_#{lang}.lang"
	hroot.hsh_t6.each { |hmod, t6| t6.set_loaded(true, lang) }
	return 0 unless FileTest.exist?(file)
	
	#Reading the file
	t6 = nil
	nstrings = 0
	IO.foreach(file) do |line| 	
		line = line.strip
		if line =~ /__Module\s*=\s*(.*)/i
			curmodule = $1.strip
			t6 = @@hsh_t6[curmodule]
		elsif t6 &&	line =~ /([^=]*)=(.*)/i
			symb = $1.strip
			sval = $2.strip
			sval = $' if sval =~ /\A\"/
			sval = $` if sval =~ /\"\Z/
			t6.store_value(lang, symb, sval)
			nstrings += 1
		end	
	end
	
	#Propagating the laoded status
	hroot.hsh_t6.each { |hmod, t6| t6.set_loaded(true, lang) }

	nstrings
end

#Edit the Language Translation via a Web dialog table
def T6Mod.visual_edition(rootname)
	hroot = @@hsh_root[rootname]
	return unless hroot
	plugin = Traductor::Plugin.myplugin(rootname)
	plugin.load_all_body_files
	plugin.load_second_phase_rubies
	t6edit = hroot.t6edit
	t6edit = T6ModEdit.new(hroot, rootname) unless t6edit
	t6edit.show_dialog
end

#Dialog Box for choosing Preferred Languages (limited version with traditional dialog box)
def T6Mod.dialog_preferred_languages

	#Computing the current Preferred Languages
	key_none = "None"
	hlang = {}
	Langue.each { |lng| hlang[lng] = Langue.nicer lng, true }
	llang = hlang.keys
	llpref = @@langpref & llang
	llang = [key_none] + llang.sort
	hlang[key_none] = T6[:T_STR_None]
	
	#Preparing the Dialog Box
	nl = 3
	hsh_params = {}
	dlg = Traductor::DialogBox.new T6[:T_HELP_MenuLanguages] 
	tl = T6[:T_STR_PreferredLanguages]
	for i in 0..nl-1
		lng = llpref[i]
		space = "                                ~"
		dlg.field_enum "#{i}", tl + " #{i+1}" + space, (lng) ? lng : key_none, hlang, llang
	end	
	
	#Calling the Dialog box
	hsh_params = dlg.show hsh_params
	return unless hsh_params
	
	#Parsing the new lang Pref
	lnew = []
	hsh_params.each do |key, lng|
		i = key.to_i
		lnew[i] = lng unless lng == key_none
	end
	lnew.uniq!
	
	#Setting the new preferred languages
	T6Mod.set_langpref lnew unless (lnew == llpref && @@langpref == llpref)		
end

#Return the list of supported language (alphabetic order)
def supported_languages
	ls = []
	@hlng.each do |code, val| 
		next if val == 1
		ls.push code 
	end	
	ls.sort { |a, b | a <=> b }
end

#Enumeration method for all symbolic string in alphabetic order --> usage: <t6>.each.symb { |symb, hsh| .... }
def each_symb(alpha_order=true)
	ls = (alpha_order) ? @hstrings.sort { |a, b| a[0] <=> b[0] } : @hstring.to_a
	ls.each do |lsymb|
		yield lsymb[0], lsymb[1]
	end	
end

#Purge unused strings (symbol defined, but no value in default language, i.e. program)
def purge_unused
	#Loop on Modules
	@hstrings.each do |symb, hsh|
		@hstrings.delete symb unless hsh[T6MOD_LNG_DEFAULT]
	end	
	nil
end

#Return the load status for a given language
def is_loaded?(lng)
	@hloaded[lng]
end

#Set the load status for all or some languages
def set_loaded(status, lst_lng=nil)
	lst_lng = @@langpref unless lst_lng
	lst_lng = [lst_lng] unless lst_lng.class == Array
	lst_lng.each do |lng|
		@hloaded[lng] = status 
	end	
	@loaded = status
end

#Make sure all language files are loaded according to the environment @@langpref
def check_loaded
	return if @loaded
	
	#Loading the translation files for the preferred languages
	@@langpref = [Sketchup.get_locale[0..1]] unless @@langpref
	@@langpref.each do |lang|
		T6Mod.load_file @rootname, lang unless @hloaded[lang]
	end
end

#Generate the file for a given language
def write_to_file (file, lng, purge=false)
	return 0 unless @hlng[lng]
	nstrings = 0
	file.puts "__Module = #{@modname}"
	each_symb do |symb, hsh|
		if purge && !hsh[T6MOD_LNG_DEFAULT]
			hsh.delete lng
			next
		end	
		val = hsh[lng]
		if val
			val = val.gsub /[\n]/, "\\n"
			file.puts "#{symb} = \"#{UTF.flatten(val)}\""
			nstrings += 1
		end	
	end	
	return nstrings
end

end	#class T6Mod

#--------------------------------------------------------------------------------------------------------------
# Class Langue: Hold Language definitions
#--------------------------------------------------------------------------------------------------------------			 

class Langue

@@hlang = {}
@@prefix = "__Langue_" 
@@file = "Langues.def"
@@keepval = false

def Langue.add(code, english_name, native_name)
	return unless code && code.strip != ""
	code = code[0..1]
	@@hlang[code] = [english_name, native_name]
	symb = (@@prefix + code).intern
	T6[symb] = english_name
	T6.store_value(code, symb.to_s, native_name)
	code
end

def Langue.load_file
	file = File.join LibFredo6.path, @@file
	return unless FileTest.exist?(file)
	
	#Reading the file
	IO.foreach(file) do |line| 	
		line = line.strip
		if line =~ /\A(\w\w)\s*=\s*(.*)/
			code = $1
			ls = $2.split(';')
			english_name = UTF.real(ls[0])
			native_name = UTF.real(ls[1])
			Langue.add(code, english_name, native_name)
		end	
	end
end

def Langue.save_to_file
	file = File.join LibFredo6.path, @@file
	File.open(file, "w") do |f| 
		f.puts T6[:T_WARNING_File]
		@@hlang.each do |code, ls|
			f.puts "#{code} = #{UTF.flatten(ls[0])};#{UTF.flatten(ls[1])}"
		end	
	end
end

def Langue.list_codes
	@@hlang.keys.sort
end

#Enumeration method for all symbolic string in alphabetic order --> usage: <t6>.each.symb { |symb, hsh| .... }
def Langue.each
	Langue.list_codes.each { |code| yield code }
end

def Langue.nicer(code, native=false)
	lcur = Langue.current_name(code)
	leng = Langue.english_name(code)
	lnat = Langue.native_name(code)
	s = "#{code}: #{lcur}"
	s += " / #{leng}" if leng != lcur
	s += " / #{lnat}" if native && lnat != lcur && lnat != code && lnat != leng
	s
end

def Langue.pretty(code, native=false)
	"#{code}: #{Langue.current_name(code)}" + ((native) ? " / #{Langue.native_name(code)}" : "")
end

def Langue.list_for_display
	lg = []
	@@hlang.each { |code, ls| lg.push Langue.pretty(code, true) }
	lg.sort
end

def Langue.english_name(code)
	return '--' unless code
	ls = @@hlang[code]
	(ls && ls[0]) ? ls[0] : code
end

def Langue.native_name(code)
	return '--' unless code
	ls = @@hlang[code]
	(ls && ls[1]) ? ls[1] : Langue.english_name(code)
end

def Langue.current_name(code)
	return code unless code
	symb = (@@prefix + code[0..1]).intern
	val = T6[symb]
	(val != symb.to_s) ? val : Langue.english_name(code)
end

end	#class Langue

#--------------------------------------------------------------------------------------------------------------
# Class UTF: Helpers to manage UTF strings
#--------------------------------------------------------------------------------------------------------------			 

class UTF

#Convert a string to another pure ASCII string where special characters are encoded !code! 
def UTF.flatten(s)
	return s unless s && s != ""
	begin
		begin
			ls = s.unpack("U*")
		rescue
			ls = []
			for i in 0..s.length-1
				ls.push s[i]
			end	
		end	
		s2 = ""
		ls.each { |c| s2 += ((c >= 128) ? "!#{c}!" : c.chr) }
		return s2
	rescue
		return s
	end
end

#Take an encoded string with !number! convention and build a UTF compatible string
def UTF.real(s)
	return s unless s && s != ""
	s = UTF.flatten s
	s.gsub(/!(\d{3,5})!/) { |u| [$1.to_i].pack("U") }
end

#Take an encoded string with !number! convention and build a UTF compatible string
def UTF.from_iso(s)
	return s unless s && s != ""
	begin
		ls = s.unpack("C*")
		s2 = ls.pack("U*")
		return UTF.real(s2)
	rescue
		return s
	end	
end

end	#class UTF

#--------------------------------------------------------------------------------------------------------------
# Default Parameters Management (only the top routines) - Rest in LibTraductor_xx.rb
#--------------------------------------------------------------------------------------------------------------			 				   

class DefaultParameters

def initialize(file, hmod)
	@file = file
	@rootname = File.basename file
	@hmod = hmod
	@hparam = {}
	@lparam = []
	@wdlg = nil
	@herror = {}
	@lst_no_default_check = [:T_DEFAULT_IconVisible]
end

def DefaultParameters.get_dir
	LibFredo6.defparam_dir
end

end	#class DefaultParameters

#--------------------------------------------------------------------------------------------------------------
# PLugin Management (only the top routines) - Rest in LibTraductor_xx.rb
#--------------------------------------------------------------------------------------------------------------			 				   

class Plugin

attr_reader :plugin_dir, :version, :load_time, :plugin_name, :folder, :loaded_rubys, :defparam,
            :main_module, :lst_obsolete_files, :name, :load_time, :load_time_startup, :load_time_second

@@hsh_plugins = {} unless defined?(@@hsh_plugins)
@@lst_files_error = [] unless defined?(@@lst_files_error)

def Plugin.hsh_plugins ; @@hsh_plugins ; end

def initialize
	@list_commands = []
	@list_handlers = []
	@handler_menu = nil
	@hsh_commands = {}
	@hsh_tpc = {}
	@default_icons_visible = []
	@default_handlers_visible = []
end
	
#Register a plugin as an extension, based on its declaration file .plugin - does not do more	
# <folder> is given by reference to the Sketchup plugin directory
# <rootname> is the name to be used for files (declaration, translation, def param), without extensions
def load_from_config(rootname, folder, plugin_name=nil)
	#Checking the rootname and Plugin_name
	return false unless folder
	rootname = plugin_name unless rootname && rootname.strip.length > 0
	plugin_name = rootname unless plugin_name && plugin_name.strip.length > 0
	return false unless rootname && rootname.strip.length > 0

	#Checking if plugin already loaded
	if @@hsh_plugins[@plugin_name]
		return true
	end	
	
	#Checking the folder and root directory
	@folder = folder
	@old_conv = (folder =~ /(.+_Dir_)(\d\d)\Z/)
	@su_plugin_dir = LibFredo6.sudir
	@plugin_dir = File.join @su_plugin_dir, folder
	
	#Checking the declaration file for the plugin
	@rootname = rootname
	@plugin_name = plugin_name
	file = File.join @plugin_dir, (rootname + '.plugin')
	return false unless FileTest.exist?(file)
	
	#Loading the declaration file
	sulng = Sketchup.get_locale
	lst_var = ['modules', 'version', 'date', 'name', 'description', 'copyright', 'creator', 
	           'ruby_files', 'ext_load', 'web_site_name', 'web_site_link', 'web_support_link', 
			   'credits', 'old_file', 'old_dir', 'picture_prefix', 'startup', 'donation',
			   'icon_conv', 'cursor_conv', 'toolbar', 'video', 'libfredo6', 'sketchup_version',
			   'web_repository', 'web_repository_link', 'bootstrap', 'upd_website', 'upd_url', 
			   'doc_pattern', 'doclinks']
	lst_var.each { |var| eval "@lst_" + var + "=[]" }		   
	IO.foreach(file) do |line| 
		line = line.force_encoding("UTF-8") if defined?(Encoding)
		lst_var.each do |var| 
			next unless Regexp.new('\A' + var + '\s*=\s*"*(.*)', Regexp::IGNORECASE).match(line)
			sval = $1.strip
			sval = $`.strip if sval =~ /"|#/
			#sval = UTF.from_iso sval if Sketchup.version.to_i <= 13
			eval "@lst_#{var}.push " + '"' + sval + '"'
			if sval =~ /\|(\w\w)\|(.*)/ 
				sval = $2
				if ($1.upcase == sulng[0..1].upcase)
					eval '@' + var + '="' + sval + '"'
					next
				end	
			end	
			eval '@' + var + '="' + sval + '" unless @' + var
		end	
	end	
	
	#Checking modules
	@lst_modules = [[rootname]] unless @lst_modules.length > 0
	@lst_modules = super_flatten @lst_modules
	@main_module = @lst_modules[0]
	@lst_ruby_files = super_flatten @lst_ruby_files
	@lst_old_file = super_flatten @lst_old_file
	@lst_old_dir = super_flatten @lst_old_dir
	@ext_load = (@ext_load =~ /true/i) ? true : false
	@nice_name = @name
	
	#Obsolete files
	@lst_obsolete_files = [] unless @lst_obsolete_files
	@lst_obsolete_files |= ["ZLoader__#{@plugin_name}.rb"]
	@lst_obsolete_files |= @lst_old_file if @lst_old_file
	@lst_obsolete_files |= @lst_old_dir if @lst_old_dir
	if rootname == "LibFredo6"
		@lst_obsolete_files += ["LibFredo6.rb", "DEFPARAM_Dir", "DOCUMENTATION_Dir"]
	end	
	
	#Other parameters
	@date = "??" unless @date
	@ext_load = true if @ext_load == nil
		
	#Storing the plugin instance
	@@hsh_plugins[@plugin_name] = self
	
	#Preparing modules with the variable @@myplugin defined and enabling T6 language translation
	@lst_modules.each do |m|
		Object.module_eval "module #{m} ; MYPLUGIN = Traductor::Plugin.myplugin('#{@plugin_name}') ; end"
	end
	@hmod_main_module = eval(@main_module)
	@plugin_title_symb = @plugin_name
	
	#Creating the loader file
	if false && @plugin_name == "LibFredo6"
		effective_load
	else
		floader = File.join @plugin_dir, "__loader.rb"
		unless FileTest.exist?(floader)
			File.open(floader, "w") do |f| 
				f.puts "Traductor::Plugin.myplugin('#{@plugin_name}').effective_load"
			end
		end	
	end	

	#Registering the plugin as an extension of Sketchup
	ext = SketchupExtension.new "#{@creator} #{@name}", File.join(@folder, File.basename(floader))
	ext.creator = @creator 
	ext.description = @description 
	ext.version = @version + " - " + @date 
	ext.copyright = @creator + " - " + @copyright 
	Sketchup.register_extension ext, true

	#Loading all Plugins using LibFredo6
	Traductor::Plugin.load_all_plugins if rootname =~ /LibFredo6/i
	
	self	
end

def super_flatten(list)
	return nil if list == nil
	list = [list] unless list.class == Array
	return list if list.length == 0
	lm = []
	list.each { |m| lm.push m.split(/;\s*|,\s*/) }
	lm.flatten
end

#Return the class instance of a Plugin given by its name
def Plugin.myplugin(plugin_name)
	@@hsh_plugins[plugin_name]
end

def Plugin.libfredo6_big_version
	@@hsh_plugins["LibFredo6"].get_big_version
end

#Return the name of a plugin
def get_name
	@plugin_name
end

#Return the full version as a string
def get_version
	@version
end

#Return the major single integre version
def get_big_version
	@version =~ /\A(\d+)\.(\d)/
	$1+$2
end

def get_upd_info
	[@upd_website, @upd_url]
end
	
#Create the Default parameter class for the plugin
def create_def_param(hmod)
	unless @defparam
		filedef = File.join(DefaultParameters.get_dir, @rootname) + ".def"
		@defparam = Traductor::DefaultParameters.new filedef, hmod
	end	
	@defparam
end

#Check if the plugin required version of LibFredo6 is correct
def libfredo6_version_valid?
	if @libfredo6 && @libfredo6.to_i > Plugin.libfredo6_big_version.to_i
		text = T6[:T_ERROR_VersionModule, @plugin_name, @libfredo6, Plugin.libfredo6_big_version] 
		LibFredo6.log "?#{text}"
		UI.messagebox text
		return false
	end	
	true
end

#Check if the plugin required version of Sketchup is correct
def sketchup_version_valid?
	return true unless @sketchup_version && @sketchup_version.class == String
	curv = Sketchup.version
	curv = $1 + $2 if curv =~ /(\d+\.\d+)\.(\d+)/
	minv = @sketchup_version
	minv = 6 unless minv.class == String
	minv = $1 + $2 if minv =~ /(\d+\.\d+)\.(\d+)/
	if curv.to_f < minv.to_f
		text = T6[:T_ERROR_VersionSketchup, @plugin_name, @sketchup_version, Sketchup.version] 
		LibFredo6.log "?#{text}"
		UI.messagebox text 
		return false
	end	
	true
end

#Get the LibFredo6 required version (as a string with 2 digis version, i.e "30", not "3.0") 
def get_libfredo6_version
	@libfredo6
end

#Start the loading process for the Plugin
def effective_load	
	return if @effective_loaded
	
	#Testing the Default Dir directory
	defdir = DefaultParameters.get_dir
	unless defdir
		text = "Cannot load plugin #{@plugin_name} because the script cannot create DEFPARAM_Dir folder in:\n "
		text += "#{LibFredo6.sudir}\n\n"
		text += "Please create it manually, respecting the case"
		UI.messagebox text
		return
	end	
	
	#Loading Other LibFredo6 module if not already loaded
	@effective_loaded = true
	
	#Loading the required plugin
	self.effective_load_part2 if libfredo6_version_valid? && sketchup_version_valid?
	
	#Warning if the load is performed outside of the LibFredo6 loading cycle
	if LibFredo6.finished_loading_cycle
		UI.messagebox "Script #{@plugin_name} is loaded but menus will show up only at next startup of Sketchup"
	end	
end

#Transfer of global value to avoid diacritic characters in Eval
def Plugin.store_path_66(path) ; @plugin_dir_66 = path ; end
def Plugin.store_rootname_66(rootname) ; @rootname_66 = rootname ; end
def Plugin.get_path_66 ; @plugin_dir_66 ; end
def Plugin.get_rootname_66 ; @rootname_66 ; end


#Effective load of a plugin (follow up)
def effective_load_part2
	t0 = Time.now.to_f
	
	#Using temporary global variables due to non-ascii characters
	Plugin.store_path_66(@plugin_dir)
	Plugin.store_rootname_66(@rootname)
	
	#Enabling all modules for T6
	@lst_modules.each do |m|
		txt = "module #{m} ;"
		txt += "Traductor::T6Mod.enable(self, Plugin.get_path_66, Plugin.get_rootname_66) ; "
		txt += "MYDEFPARAM = MYPLUGIN.create_def_param(self) ; "
		txt += "SU_MAJOR_VERSION = (Sketchup.version[/(\\d+)/]).to_i	; "
		txt += "PC_OR_MAC = (RUBY_PLATFORM =~ /darwin/i) ? 'MAC' : 'PC' ;"
		txt += "RUN_ON_MAC = #{RUN_ON_MAC.inspect} ;"
		txt += "end;"
		Object.module_eval txt
	end
	
	#Loading specified ruby files
	load_full_rubies unless load_bootstrap_rubies

	#Invoke the startup commands
	t1 = Time.now.to_f
	err = nil
	@lst_startup = ["startup"] if @lst_startup.length == 0
	@lst_startup.each do |m|
		begin
			mm = Object.const_get(@main_module)
			mm.send(m)
		rescue Exception => e
			text = "#{@plugin_name}  #{@version}: Error starting up plugin (creating menus and icons)"
			err = LibFredo6.mask_dir LibFredo6.exception_message(e)
			LibFredo6.log "?#{text}", err
			LibFredo6.log_messagebox text, err
		end
	end
	@load_time_startup = Time.now.to_f - t1
	
	unless err
		delta = (@load_time_startup * 1000).to_i
		LibFredo6.log "#{@plugin_name}  #{@version}: Startup (creating menus and icons)   [#{delta} ms]"
	end	
	
	#Storing the load time
	@load_time = Time.now.to_f - t0
	
	#Registering the Plugin for common services
	hargs = { :version => @version, :date => T6[@lst_date], :author => @creator, :dir => @plugin_dir,
              :link_info => @web_support_link, :description => Traductor[@lst_description],
			  :load_time => @load_time, :load_time_startup => @load_time_startup, 
			  :load_time_files => (@load_time - @load_time_startup),
			  :plugin6 => self,
              :required => ((@libfredo6) ? "LibFredo6 v#{@libfredo6}" : nil) }
	AllPlugins.register @plugin_name, hargs
	Plugin.contribute_load_time(load_time)	
end

#Load bootstrap ruby
def load_bootstrap_rubies
	t0 = Time.now
	@loaded_bootstrap = []
	#@lst_bootstrap = [] unless @lst_bootstrap
	@lst_bootstrap = ["bootstrap*"] if @lst_bootstrap.length == 0
	lst_errors = []
	@lst_bootstrap.each do |rb|
		rb = rb + '.rb' unless rb =~ /\.rb\Z/i
		if FileTest.exist?(rb)
			lsrb = [rb]
		else	
			lsrb = Dir[File.join(@plugin_dir, rb)]
		end	
		lsrb.each do |f| 
			bf = File.basename(f)
			begin
				status = require f
				@loaded_bootstrap.push bf if status
			rescue Exception => e
				err = LibFredo6.mask_dir LibFredo6.exception_message(e)
				LibFredo6.log "?#{@plugin_name}  #{@version}: ERROR in Loading ruby bootstrap file #{bf}", err
				lst_errors.push bf
			end	
		end	
	end
	if lst_errors.length > 0
		@@lst_files_error += lst_errors
	else
		delta = ((Time.now - t0) * 1000).to_i
		LibFredo6.log "#{@plugin_name} #{@version}: Bootstrap Ruby files loaded         [#{delta} ms]" if @loaded_bootstrap.length > 0
	end
	@loaded_bootstrap.length > 0
end

#Load all other rubies
def load_full_rubies
	t0 = Time.now
	@loaded_rubys = []
	@lst_ruby_files = "*" if @lst_ruby_files.length == 0
	lst_errors = []
	@lst_ruby_files.each do |rb|
		rb += '.rb' unless rb =~ /\.rb\Z/i
		if FileTest.exist?(rb)
			lsrb = [rb]
		else	
			lsrb = Dir[File.join(@plugin_dir, rb)]
		end	
		lsrb.each do |f| 
			bf = File.basename(f)
			begin
				status = require f
				@loaded_rubys.push bf if status
			rescue Exception => e
				err = LibFredo6.mask_dir LibFredo6.exception_message(e)
				LibFredo6.log "?#{@plugin_name} #{@version}: ERROR in Loading ruby file #{bf}", err
				lst_errors.push bf
			end	
		end	
	end	
	if lst_errors.length > 0
		@@lst_files_error += lst_errors
	else
		delta = ((Time.now - t0) * 1000).to_i
		LibFredo6.log "#{@plugin_name} #{@version}: Ruby files loaded         [#{delta} ms]" if @loaded_rubys.length > 0
	end
	@fully_loaded = true
end

#Second phase loading of rubies
def load_second_phase_rubies
	return if @fully_loaded
	
	Sketchup.set_status_text T6[:T_HELP_LoadingModules, @plugin_name]
	t0 = Time.now.to_f
	@@lst_files_error = []
	load_full_rubies
	@load_time_second = Time.now.to_f - t0
	AllPlugins.update @plugin_name, { :load_time_second => @load_time_second }
	text = "#{(@load_time_second * 1000.0).to_i} ms"
	Sketchup.set_status_text T6[:T_HELP_LoadedModules, @plugin_name, text]
end

#Find the list of plugins to load and load them
def Plugin.load_all_plugins
	@@lst_files_error = []
	t0 = Time.now.to_f
	hsh_new_conv0 = {}
	hsh_new_conv = {}
	hsh_plug = {}
	
	#Scanning the directories to find subfolders with a .plugin file
	Dir[File.join(LibFredo6.sudir, "*")].each do |f|
		s = File.basename f
		lplu = Dir[File.join(f, "*.plugin")]
		next if lplu.empty?
		pluname = File.basename lplu[0], ".plugin"
		next if pluname =~ /LibFredo6/i
		
		#Old naming convention
		if s =~ /\A(.+)_Dir_(\d\d)/i
			key = $1
			next if key == "LIBFREDO6"
			oldver = hsh_plug[pluname] 
			if oldver
				hsh_plug[pluname][1] = s if s > oldver[1]
			else
				hsh_plug[pluname] = [pluname, s]
			end	
		
		#New naming convention without explicit version
		elsif s =~ Regexp.new("_#{pluname}\\Z", Regexp::IGNORECASE)
			next unless FileTest.exist?(File.join(File.dirname(f), "#{s}.rb"))
			hsh_new_conv0[pluname] = [pluname, s]
			
		#New naming convention with explicit version
		elsif s =~ Regexp.new("_#{pluname}_\\d\\d\\Z", Regexp::IGNORECASE)
			###next unless FileTest.exist?(File.join(File.dirname(f), "#{$`}_#{pluname}.rb"))
			oldver = hsh_new_conv[pluname] 
			if !oldver || s > oldver[1]
				hsh_new_conv[pluname] = [pluname, s]
			end	
			
		end	
	end	
	
	#Skipping plugin with old conventions if new convention exists too
	hsh_new_conv.each { |key, val| hsh_plug[key] = val }
	hsh_new_conv0.each { |key, val| hsh_plug[key] = val }
		
	#Loading the plugins with old convention
	lplug = hsh_plug.values.sort { |a, b| a[0] <=> b[0] }
	lplug.each { |a| Traductor::Plugin.new.load_from_config *a }
	
	lplug
end

#Loading of a plugin from its top rb file
#Argument should be __FILE__
def Plugin.load_plugin_from_top_rb(filepath)
	folder = File.dirname filepath
	filename = File.basename filepath, ".rb"
	pat = Regexp.new "#{filename}_\\d\\d\\Z"
	
	#Checking if there a directory with same name associated to the rb file
	if FileTest.directory?(File.join(folder, filename))
		dl = [File.join(folder, filename)]
	else	
		dl = Dir[File.join(folder, "#{filename}_??")]
	end	
	return nil if dl.empty?
	fulldir = nil
	ls = []
	dl.each do |d|
		f = File.basename d
		if f == filename
			fulldir = f
			break
		elsif f =~ pat
			ls.push f
		end	
	end
	ls = ls.sort	
	fulldir = ls.last
	return nil unless fulldir
	
	#Checking if it contains a .plugin file
	lplu = Dir[File.join(folder, fulldir, "*.plugin")]
	return nil if lplu.empty?
	pluname = File.basename lplu[0], ".plugin"
	
	#Loading the plugin if not already loaded
	if !@@hsh_plugins[pluname]
		Traductor::Plugin.new.load_from_config pluname, fulldir
	#elsif LibFredo6.finished_loading_cycle
	#else
		#UI.messagebox "Script #{pluname} will be loaded at next startup of Sketchup"
	end	
end

#Signal the errors in Plugins loading phase
def Plugin.signal_error_in_loading
	return unless @@lst_files_error.length > 0
	T6[:MSG_Error_LoadingRuby] = "The following files are in ERROR and cannot be loaded:"
	T6[:MSG_Error_OpenTraceLog] = "Do you want to open the Trace Log?"
	text = "Message from LibFredo6\n\n"
	text += T6[:MSG_Error_LoadingRuby] + "\n" + @@lst_files_error.join("\n") + "\n\n" + T6[:MSG_Error_OpenTraceLog]
	status = UI.messagebox text, MB_YESNO
	if status == 6
		TraceLogDialog.invoke
	end	
end

#Store the contribution to the total load time
def Plugin.contribute_load_time(load_time)
	#Retrieve the current list of times
	key = "Average_load_times"
	nmax = 11
	s = Sketchup.read_default "LibFredo6", key, ""
	ls = s.split ";"
	lstime = []
	n = ls.length
	for i in 0..n-1
		a = ls[n-i-1]
		break unless a && i < nmax
		lstime.push ls[i].to_i
	end
	
	#Setting the origin when first plugin loaded
	lstime.push 0 if AllPlugins.get_all_plugins.length <= 1
	
	#Adding current load time
	lstime[-1] = (lstime[-1] + load_time * 1000).round
	
	#Store back
	lstime.shift if lstime.length >= nmax
	Sketchup.write_default "LibFredo6", key, lstime.join(';')	
end

#Retrieve or store the average plugin time
def Plugin.average_load_time
	key = "Average_load_times"
	nmax = 10
	
	#Retrieve the current list of times
	s = Sketchup.read_default "LibFredo6", key, ""
	ls = s.split ";"
	lstime = []
	n = ls.length
	for i in 0..n-1
		a = ls[n-i-1]
		break unless a && i < nmax
		lstime.push ls[i].to_i
	end
	
	#Retrieve the Current average
	return nil if lstime.empty?
	x = 0
	lstime.each { |t| x += t }
	[lstime.length, (x * 1.0 / lstime.length).round]
end

#Compute the load time at SU startup of all registered plugins, including LibFredo6
def Plugin.total_load_time
	ttot = 0
	@@hsh_plugins.each do |key, plugin|
		ttot += plugin.load_time.to_f
	end
	(ttot * 1000).to_i
end

def Plugin.get_total_load_time
	"#{Plugin.total_load_time} ms"
end

end	#class Plugin


#For troubleshooting purpose
def Traductor.trace6(*args)
	unless @@file_trace6
		file = File.join LibFredo6.sudir, "Fredo6Trace.txt"
		@@file_trace6 = File.open(file, "w")
		@@file_trace6.puts "RUBY_PLATFORM = #{RUBY_PLATFORM}"
	end
	@@file_trace6.puts *args
	@@file_trace6.flush
end


#--------------------------------------------------------------------------------------------------------------
# Class AllPlugins: class initialization part
#--------------------------------------------------------------------------------------------------------------			 

#Class variables initialization
unless defined?(self::AllPlugins)
class AllPlugins
	@@hsh_allplugins = {}	
end
end

class AllPlugins

#--------------------------------------------------------------------------------------------------------------
# Class methods: Manage plugin registration
#Valid fields are
# 	:name
#	:author
#	:description
#	:load_time
#	:load_time_startup
#	:load_time_files
#	:load_time_second --> Phase 2 loading time
#	:lst_obsolete_files --> List of obsolete files (just the basenames)
#--------------------------------------------------------------------------------------------------------------

#Plugin Registration: Parameters are passed in one or several hash arrays
def AllPlugins.register(plugin_name, *hargs)
	return unless plugin_name.class == String && plugin_name.length > 0
	hsh = @@hsh_allplugins[plugin_name]
	return if hsh
	@@hsh_allplugins[plugin_name] = hsh = {} unless hsh
	hargs.each do |h| 
		h.each do |key, val| 
			val = Traductor.encode_to_ruby val
			hsh[key] = (val.class == String || val.class == Symbol) ? T6[val] : val
		end	
	end	
	hsh[:name] = plugin_name
	hsh[:author] = T6[:T_TXT_Anonymous] unless hsh[:author]
	hsh[:installed] = true
	hsh
end

#Plugin Registration: Parameters are passed in one or several hash arrays
def AllPlugins.update(plugin_name, *hargs)
	return unless plugin_name.class == String && plugin_name.length > 0
	hsh = @@hsh_allplugins[plugin_name]
	@@hsh_allplugins[plugin_name] = hsh = {} unless hsh
	hargs.each do |h| 
		h.each do |key, val| 
			val = Traductor.encode_to_ruby val
			hsh[key] = (val.class == String || val.class == Symbol) ? T6[val] : val
		end	
	end	
	hsh
end

#Log information on version for SCF statistics
def AllPlugins.scf_version_log
	return unless defined?(SCFapi) && defined?(SCFapi.storekeyvalue)
	
	#Getting the past information
	dico = "LibFredo6_SCF"
	section = "VersionOfPlugins"
	sparam = Default.read dico, section
	hsh_version = {}
	begin
		hsh_version = eval(sparam) if sparam
	rescue
		hsh_version = {}
	end

	#Checking if there is a new version versus the one registered
	author = "FrwGx547CiKm9qploXCD"
	changed = false
	@@hsh_allplugins.each do |plugin_name, hsh|
		next unless hsh[:link_info]
		version = hsh[:version]
		next unless version && plugin_name !~ /::/
		version = 'v' + version.gsub(/\./, '')
		next if hsh_version[plugin_name] == version
		changed = true
		hsh_version[plugin_name] = version
		begin
			###SCFapi.storekeyvalue author, plugin_name, "Version", "#{plugin_name} - #{version}"
		rescue
		end
	end
	
	#Storing the information
	if changed
		sparam = hsh_version.inspect.gsub('"', "'")
		Default.store dico, section, sparam	
	end
	
	nil
end

def self.cleanup_string_encoding(s)
	return s unless s.class == String
	begin
		s.unpack 'U*'
	rescue
		s = s.unpack('C*').pack('U*')
	end
	s
end

#Fetch external plugin registration
def self.get_all_plugins ; @@hsh_allplugins ; end

#Finding out if module has registered for LibFredo6
def self.get_all_registered_plugins
	for m in Object.constants
		next if m == :Config
		m = Object.const_get m
		next unless m.class == Module
		next unless m.respond_to?(:register_plugin_for_LibFredo6)
		begin
			method = m.method "register_plugin_for_LibFredo6"
			next unless m
			if method.arity == 1
				lang = T6Mod.get_langpref[0]
				lang = "" unless lang
				harg = method.call lang
			elsif method.arity == 0	
				harg = method.call
			else
				next
			end	
			next unless harg
			name = harg[:name]
			name = harg[:plugin] unless name
			AllPlugins.register name, harg if name
		rescue
		end
	end
	@@hsh_allplugins
end

#Declare that a plugin is obsolete
def self.declare_obsolete_plugin(plugin_name, text)
	return unless plugin_name.class == String && plugin_name.length > 0
	hsh = @@hsh_allplugins[plugin_name]
	@@hsh_allplugins[plugin_name] = hsh = {} unless hsh
	hsh[:obsolete] = text
end

#Return load time info in ms
def self.load_time_info(plugin_name)
	hsh = @@hsh_allplugins[plugin_name]
	return [] unless hsh
	t = []
	t[0] = hsh[:load_time_files]
	t[1] = hsh[:load_time_startup]
	t[2] = hsh[:load_time]
	t[3] = hsh[:load_time_second]
	if t[2] || t[3]
		t[4] = 0
		t[4] += t[2] if t[2]
		t[4] += t[3] if t[3]
	else
		t[4] = nil
	end	
	t = t.collect { |a| (a * 1000).to_i if a}
	t
end

#Compute the load time at SU startup of all registered plugins, including LibFredo6
def self.total_load_time
	ttot = 0
	@@hsh_allplugins.each do |key, hsh|
		t = hsh[:load_time]
		ttot += t if t
	end
	(ttot * 1000).to_i
end

#Compute a tooltip text with plugin name, version, date and author
def self.compute_tooltip(plugin_name)
	hsh = @@hsh_allplugins[plugin_name]
	return [] unless hsh
	plugin6 = hsh[:plugin6]
	return plugin6.compute_tooltip if plugin6
	"#{@plugin_name} v#{hsh[:version]} - #{hsh[:date]} - #{hsh[:author]}"
end

#Get the list of obsolete files for a plugin
def self.get_obsolete_files(plugin_name)
	hsh = @@hsh_allplugins[plugin_name]
	return [] unless hsh
	plugin6 = hsh[:plugin6]

	#Registered files
	return plugin6.check_obsolete_files if plugin6
	
	#Other plugins
	ls = hsh[:lst_obsolete_files]
	return [] unless ls
	ls = [ls] unless ls && ls.class == Array
	return [] unless ls.length > 0
	
	lfiles = []
	$:.each do |sudir|
		ls.each do |f|
			lfiles += Dir[File.join(sudir, f)]
		end	
	end
	lfiles
end

end	#class AllPlugins

#--------------------------------------------------------------------------------------------------------------
#--------------------------------------------------------------------------------------------------------------
# Mixin module MixinCallBack: 
#     - mechanism of sub classing callback
#     - Common methods to ALL tools 
#--------------------------------------------------------------------------------------------------------------			 
#--------------------------------------------------------------------------------------------------------------			 

module MixinCallBack

#Register the caller class instance
def _register_caller(caller)
	@_calleR = caller
end

#Call the caller methgod, with convention 'sub_xxxx'
def _sub(func, *args)
	return unless @_calleR && func
	smeth = 'sub_' + func.to_s
	symb = smeth.intern
	if @_calleR.respond_to? symb
		return @_calleR.send(symb, *args)
	elsif self.respond_to? symb
		return self.send(symb, *args)
	elsif block_given?
		return yield
	end	
	nil
end

end	#mixin module MixinCallBack

end	#module Traductor

	