=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Designed Dec. 2008 by Fredo6

# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for commercial purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:  Lib6Registry.rb
# Original Date	:  20 Jan 2011
# Type			:  Script library part of the LibFredo6 shared libraries
# Description	:  A utility library to store and retrieve parameters in the SU Registry.
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end

module Traductor

#**************************************************************************************************************			 
#--------------------------------------------------------------------------------------------------------------
# REGISTRY: Helper for storing web dialog information in Sketchup registry
#--------------------------------------------------------------------------------------------------------------			 
#**************************************************************************************************************			 

class Registry

#--------------------------------------------------------------------------------------------------------------
# REGISTRY: Generic routines
#--------------------------------------------------------------------------------------------------------------			 

def Registry.store(regkey, *args)
	#Sketchup.write_default "LibFredo6", regkey, args.join(';')
	Default.store "LibFredo6", regkey, args.join(';')
end

def Registry.load(regkey)
	#param = Sketchup.read_default "LibFredo6", regkey
	param = Default.read "LibFredo6", regkey
	return nil unless param && param.length > 0
	param
end

#--------------------------------------------------------------------------------------------------------------
# REGISTRY: Web Dialog
#--------------------------------------------------------------------------------------------------------------			 

#Store the size of the screen display in registry
def Registry.browser_info_store(*args)
	Registry.store "WDLG6-SCREENINFO", *args
end

#Store the size of the screen display in registry
def Registry.browser_info_load()
	param = Registry.load "WDLG6-SCREENINFO"
	return nil unless param
	param = param.split ';'
	browser, sw, sh = param
	[browser, sw.to_i, sh.to_i]
end

#Store the position and size of a Wldg in the registry under key
def Registry.wposition_store(key, *args)
	Registry.store "WDLG6-POS-" + key, *args
end

#Load the position and size of a Wldg in the registry under key - Returns [xpos, ypos, sx, sy]
def Registry.wposition_load(key)
	default = [100, 100, 100, 100]
	param = Registry.load "WDLG6-POS-" + key
	return default unless param && param.length > 0
	begin
		param = param.split(';').collect { |a| a.to_i }
		param.each { |a| return default unless a.class == Fixnum }  
	rescue
		param = default
	end	
	param
end

end #Class Registry

#**************************************************************************************************************			 
#--------------------------------------------------------------------------------------------------------------
# DEFAULT: Storage of Sketchup registry, either in real registry or in a flat file
#--------------------------------------------------------------------------------------------------------------			 
#**************************************************************************************************************			 

class Default

@@dir_data = LibFredo6.appdata_dir
@@file_default = nil unless defined?(@@file_default)
@@hsh_defaults = nil unless defined?(@@hsh_defaults)

def Default.init
	return if @@file_default || !@@dir_data
	@@file_default = File.join @@dir_data, "LibFredo6_all_defaults.dat"
	@@hsh_defaults = {}
	
	#Read the file if it exists
	if FileTest.exist?(@@file_default)
		lines = IO.readlines(@@file_default)
		hsh = nil
		lines.each do |line|
			if line =~ /\A==(.+)\Z/
				dico = $1
				hsh = @@hsh_defaults[dico]
				hsh = @@hsh_defaults[dico] = {} unless hsh
			elsif line =~ /\A(.*)::=(.*)/
				hsh[$1] = $2 if hsh
			end				
		end
	end
end

#DEFAULT: Store information by dico and key
def Default.store(dico, key, val)
	Default.init
	return Sketchup.write_default(dico, key, val) unless @@file_default
	hsh = @@hsh_defaults[dico]
	hsh = @@hsh_defaults[dico] = {} unless hsh	
	hsh[key] = val
	Default.write_to_file
end

#DEFAULT: Read information by dico and key
def Default.read(dico, key, def_val=nil)
	Default.init
	unless @@file_default
		val = Sketchup.read_default(dico, key)
		return def_val if val == nil
	end	
	hsh = @@hsh_defaults[dico]
	val = (hsh && hsh.has_key?(key)) ? hsh[key] : Sketchup.read_default(dico, key) 
	(val == nil) ? def_val : val
end

#DEFAULT: Write to file
def Default.write_to_file
	File.open(@@file_default, "w") do |f| 
		@@hsh_defaults.each do |dico, hsh|
			f.puts "==#{dico}"
			hsh.each { |key, val| f.puts "#{key}::=#{val}" }
		end		
	end	
end

#DEFAULT: Delete a key in the pseudo registry file
def Default.delete_key(dico, key)
	Default.init
	return Sketchup.write_default(dico, key, nil) unless @@file_default
	hsh = @@hsh_defaults[dico]
	return nil unless hsh && hsh.has_key?(key) 
	hsh.delete(key)
	Default.write_to_file
end

#DEFAULT: Delete all entries for a dico
def Default.delete_dico(dico)
	Default.init
	return nil unless @@file_default
	return nil unless @@hsh_defaults.has_key?(dico) 
	@@hsh_defaults.delete(dico)
	Default.write_to_file
end

end	#class Default

end #Module Traductor
