=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Designed by Fredo6 - Copyright April 2009

# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for commercial purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:   Lib6Selmode.rb
# Original Date	:   8 May 2009 - version 1.0
# Description	:   Utility for managing selection mode for other scripts
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end


module Traductor

T6[:T_DEFAULT_Section_EdgeSelection] = "Parameters for Edge Selection"
T6[:T_DEFAULT_Flag_SelectionAperture] = "Aperture for picking in pixel (0 means Sketchup default)"
T6[:T_DEFAULT_Flag_SelectionModifiers] = "Edge Selection mode"
T6[:T_DEFAULT_Flag_SelectionAnglemax] = "Maximum Edge Angle for Follow mode (degree)"

T6[:T_MNU_Extend_None] = "Selection edge by edge"
T6[:T_MNU_Extend_Connected] = "Extend selection to all connected edges (CTRL + SHIFT)"
T6[:T_MNU_Extend_Curve] = "Extend selection to curve (CTRL)"
T6[:T_MNU_Extend_Follow] = "Extend selection to cofacial and aligned edges (SHIFT)"
T6[:T_MNU_Extend_Anglemax] = "Maximum deviation angle in degree (in VCB followed by d)"

T6[:T_OPT_Extend_None] = "Edge by Edge"
T6[:T_OPT_Extend_Connected] = "All connected edge"
T6[:T_OPT_Extend_Curve] = "Curve"
T6[:T_OPT_Extend_Follow] = "Follow mode"

#--------------------------------------
# Valid Mode
#	'N' : Edge by edge
#	'A' : All connected
#	'C' : Curve
#	'F' : Follow (based on an angle max)
#--------------------------------------

class SelMode

#--------------------------------------
# Class Methods
#--------------------------------------

#Contribute to the default parameters
def SelMode.default_param(defparam, symbroot, default=nil, options=nil)
	options = 'NACF' unless options
	default = 'N' unless default
	
	klist = []
	klist.push ['N', T6[:T_OPT_Extend_None]] if options =~ /N/i
	klist.push ['A', T6[:T_OPT_Extend_Connected]] if options =~ /A/i
	klist.push ['C', T6[:T_OPT_Extend_Curve]] if options =~ /C/i
	klist.push ['F', T6[:T_OPT_Extend_Follow]] if options =~ /F/i
	
	text_flag = T6[:T_DEFAULT_Flag_SelectionModifiers]
	text_angle = T6[:T_DEFAULT_Flag_SelectionAnglemax]
	text_aperture = T6[:T_DEFAULT_Flag_SelectionAperture]
	
	defparam.separator :T_DEFAULT_Section_EdgeSelection
	defparam.declare SelMode.default_symb_aperture(symbroot), 0, 'I:>=0<=30', nil, text_aperture
	defparam.declare SelMode.default_symb_modifier(symbroot), 'N', 'H',  klist, text_flag
	defparam.declare SelMode.default_symb_anglemax(symbroot), 30.0, 'F:>=0<=90', nil, text_angle	
end

def SelMode.default_symb_modifier(symbroot)
	s = symbroot.to_s + '__Selmode_modifier'
	s.intern
end

def SelMode.default_symb_anglemax(symbroot)
	s = symbroot.to_s + '__Selmode_anglemax'
	s.intern
end

def SelMode.default_symb_aperture(symbroot)
	s = symbroot.to_s + '__Selmode_aperture'
	s.intern
end

#--------------------------------------
# Class initialization
#--------------------------------------

def initialize(*args)
	MYPLUGIN.load_body(self, __FILE__)
	initialize__(*args)
end

end	# class SelMode

end	#module Traductor

