=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Designed July 2013 by Fredo6

# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for commercial purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:   Lib6FacePicker.rb
# Original Date	:   15 Jul 2013
# Description	:   Manage interactive selection of faces and standlone edges
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end

module Traductor

T6[:TIP_ClickSelectFaces] = "Click to Select faces"
T6[:TIP_ClickUnselectFaces] = "Click to Unselect faces"
T6[:TIP_HistoryAddFaces] = "Add faces"
T6[:TIP_HistoryRemoveFaces] = "Remove faces"

T6[:T_TIP_ClickDefaultOption] = "Click to set Default options"
T6[:T_TIP_TABCycleOption] = "TAB to cycle through options"

#=============================================================================================
#=============================================================================================
# Class FacePicker: main class for the Face Picker Interactive tool
#=============================================================================================
#=============================================================================================

class FacePicker

SelInfo = Struct.new :tr, :parent, :hfaces, :view_tracking_ref, :draw_frame
HoverInfo = Struct.new :tr, :parent, :mode, :faces_id, :faces, :view_tracking_ref, :draw_frame
DrawFrame = Struct.new :frames, :dashed, :contour, :triangles, :ll_frames, :ll_dashed, :ll_contour, :ll_triangles

#INIT: Class initialization
def initialize__(*hargs)
	#Initialization
	@model = Sketchup.active_model
	@selection = @model.selection
	@view = @model.active_view
	@ph = @view.pick_helper	
	@rendering_options = Sketchup.active_model.rendering_options
	@tr_id = Geom::Transformation.new
	@viewtracker = G6::ViewTracker.new
	@button_down = false
	@dragging = false
	@pixel_dragging = 5
	@hmaster_comp = {}
	
	#Parsing the arguments
	hargs.each do |harg|	
		harg.each { |key, value|  parse_args(key, value) } if harg.class == Hash
	end
	
	#Initialization
	text_init
	colors_init
	reset
	history_init
	
	#Default parameters
	@modes_face_selection = [:single, :surface, :connected, :same_color]
	@option_face_selection = :surface unless @option_face_selection
	
	#Initial selection
	selection_init
end

#INIT: Parse the parameters of the Face Picker
def parse_args(key, value)
	skey = key.to_s
	case skey
	when /notify_proc/i
		@notify_proc = value
	when /option_face_selection/i
		@option_face_selection = value
	when /option_ignore_edges/i
		@option_ignore_edge = value
	when /text_drag/i
		@text_drag = value
	when /text_reselect/i
		@text_reselect = value	
	when /make_group_unique_proc/
		@make_group_unique_proc = value
	when /no_tooltip/
		@no_tooltip = value
	end	
end	

def set_no_tooltip(on) ; @no_tooltip = on ; end

#INIT: Return the parameters of the face picker
def get_hparams
	{ :option_face_selection => @option_face_selection }
end

#INIT: Initialize texts
def text_init
	@tip_click_select_faces = T6[:TIP_ClickSelectFaces]
	@tip_click_select_faces += "\n" + @text_drag if @text_drag
	@tip_click_unselect_faces = T6[:TIP_ClickUnselectFaces]
	@tip_click_unselect_faces += "\n" + @text_drag if @text_drag
	
	@tip_history_add_faces = T6[:TIP_HistoryAddFaces]
	@tip_history_remove_faces = T6[:TIP_HistoryRemoveFaces]
	
	prefix = T6[:T_BOX_FaceSelection] + ": "
	@mnu_selection_single = prefix + T6[:T_TIP_FaceSelectionSingle]
	@mnu_selection_surface = prefix + T6[:T_TIP_FaceSelectionSurface]
	@mnu_selection_connected = prefix + T6[:T_TIP_FaceSelectionConnected]
	@mnu_selection_samecolor = prefix + T6[:T_TIP_FaceSelectionSameColor]
end

#INIT: Initialize colors
def colors_init
	@color_hover_frame_add = 'green'
	@color_hover_face_add = Sketchup::Color.new 'lightgreen'
	@color_hover_face_add.alpha = 0.5
	@color_hover_face_add_h = Sketchup::Color.new 'lightgreen'
	@color_hover_face_add_h.alpha = 0.7
	
	@color_hover_frame_remove = 'red'
	@color_hover_face_remove = Sketchup::Color.new 'pink'
	@color_hover_face_remove.alpha = 0.5
	@color_hover_face_remove_h = Sketchup::Color.new 'pink'
	@color_hover_face_remove_h.alpha = 0.7
	@color_selection_frame = 'blue'
	@color_selection_face = 'yellow'
	
	@hsh_boxinfo_add = { :bk_color => 'lightgreen', :fr_color => 'green' }					 
	@hsh_boxinfo_remove = { :bk_color => 'pink', :fr_color => 'red' }					 
end

#HOVERINFO: initialize the environment
def reset
	@current_hoverinfo = nil
	@last_initial_face = nil
	@hsh_hoverinfo_faces = {}
	@sel_modified = false
	selection_init
	history_init
end

#Notify the caller of events in the selection process
def notify_selection(code)
	@notify_proc.call code if @notify_proc 
end
	
#-------------------------------------------
# OPTIONS: Manage the options
#-------------------------------------------

#OPTIONS: Mode for face selection
def get_option_face_selection ; @option_face_selection ; end
def set_option_face_selection(mode=nil)
	mode = :surface unless mode
	@option_face_selection = mode
	@focus_initial_selection = false
end	

#OPTIONS: Cycle thorugh the face selection modes
def cycle_option_face_selection(up=true)
	if @focus_initial_selection
		@focus_initial_selection = false
		return
	end	
	ipos = @modes_face_selection.rindex @option_face_selection
	n = @modes_face_selection.length
	incr = (up) ? +1 : -1
	set_option_face_selection @modes_face_selection[(ipos + incr).modulo(n)]
end

#-------------------------------------------
# Picking and hovering Faces and Edges
#-------------------------------------------

#PICK: Manage actions on a mouse move
def handle_move(flags, x, y, view)
	@xmove = x
	@ymove = y
	hover_current_face(flags, x, y, view)
end
	
#HOVER: Identify which objects (face and edge)  are under the mouse
def hover_under_mouse(flags, x, y, view)
	precision = 0
	@ph.do_pick x, y, precision
	
	elt = ph_face = @ph.picked_face

	#Finding the parent and transformation
	tr = @tr_id
	parent = nil
	return nil unless elt
	for i in 0..@ph.count
		ls = @ph.path_at(i)
		if ls && ls.include?(elt)
			parent = ls[-2]
			tr = @ph.transformation_at(i)
			break
		end
	end	
	tr = @tr_id unless tr
	parent = @model unless parent
	[ph_face, tr, parent]
end
	
#HOVER: Get the master parent if the parent is a component instance	
def hover_master_parent(parent)
	return parent unless parent && parent.instance_of?(Sketchup::ComponentInstance)
	cdef = parent.definition
	cdef_id = cdef.entityID
	master_parent = @hmaster_comp[cdef_id]
	master_parent = @hmaster_comp[cdef_id] = parent unless master_parent
	master_parent
end
	
#HOVER: Interactive pick of edges, faces or container
def hover_current_face(flags, x, y, view)
	@x = x
	@y = y
	
	#Getting the faces under the mouse
	@initial_face, @tr, @parent = hover_under_mouse(flags, x, y, view)
	
	#No face picked
	unless @initial_face
		@initial_face_id = nil
		@current_hoverinfo = nil
		@last_initial_face = nil
		@sel_modified = false
		return
	end
	@initial_face_id = @initial_face.entityID
	
	#Face already hovered at last pick
	if @initial_face == @last_initial_face && @option_face_selection == @last_option_face_selection
		return
	end
	
	#Storing the current information
	@last_initial_face = @initial_face
	@last_option_face_selection = @option_face_selection
	
	#Selection mode is single - Do not optimize with Hover info
	if @focus_initial_selection || (@option_face_selection == :single && @rendering_options["DrawHidden"])
		@hover_add = selection_add?([@initial_face])
		@current_hoverinfo = nil
		@sel_modified = false
		return
	end
	
	#Finding the hover info for the faces
	hoverinfo = hoverinfo_check_face @initial_face
	@sel_modified = false if @current_hoverinfo != hoverinfo
	if hoverinfo
		@hover_add = selection_add?(hoverinfo.faces) if @current_hoverinfo != hoverinfo
		@current_hoverinfo = hoverinfo
		return
	end

	#Getting the other faces based on the selection option
	lfaces = hover_extend_faces @initial_face, @option_face_selection
	
	#Checking if faces must be added or removed
	@hover_add = selection_add?(lfaces)
	
	#Registering the hoverinfo
	@current_hoverinfo = hoverinfo_register lfaces
end

#HOVER: Extending the initial face
def hover_extend_faces(face0, option_face_selection)
	return nil unless face0
	case option_face_selection
	when :connected
		lsfaces = face0.all_connected.find_all { |e| e.instance_of?(Sketchup::Face) } 
	when :same_color
		lsfaces = adjacent_faces_same_material(face0)
	else
		lsfaces = G6.face_neighbours(face0)
	end	
	lsfaces	
end

#HOVER: Find the adjacent faces with same color	
def adjacent_faces_same_material(face0)
	recto0 = front_face_is_visible?(face0)
	mat0 = (recto0) ? face0.material : face0.back_material
	
	hsh_faces = {}
	lface = [face0]	
	while lface.length > 0
		f = lface.shift
		next if hsh_faces[f.entityID]
		hsh_faces[f.entityID] = f
		f.edges.each do |e|
			e.faces.each do |ff| 
				next if ff == f || hsh_faces[ff.entityID]
				recto = front_face_is_visible?(ff)
				mat = (recto) ? ff.material : ff.back_material
				lface.push ff if mat == mat0
			end	
		end
	end	
	hsh_faces.values
end

#HOVER: Determine if the front (or back) face is the visible face
def front_face_is_visible?(face)
	pt2d = @view.screen_coords(@tr * face.vertices[0].position)
	ray = @view.pickray pt2d.x, pt2d.y
	vec = @tr * face.normal
	(vec % ray[1] <= 0)	
end

#---------------------------------------------------------------------------------------------
# HOVERINFO: Management of Hover information
#---------------------------------------------------------------------------------------------

#HOVERINFO: Compute the key for a face
def hover_key(face)
	"#{face.entityID}_#{@option_face_selection}_#{@parent.entityID}_#{@tr.to_a}"
end

#HOVERINFO: Check if a face is referenced in a HoverInfo structure
def hoverinfo_check_face(face)
	@hsh_hoverinfo_faces[hover_key(face)]
end

#HOVERINFO: create a hoverinfo from a list of faces
def hoverinfo_register(faces)
	hoverinfo = HoverInfo.new
	@lst_hoverinfo.push hoverinfo
	hoverinfo.tr = @tr
	hoverinfo.parent = @parent
	hoverinfo.mode = @option_face_selection
	hoverinfo.faces = faces.clone
	hoverinfo.faces_id = faces.collect { |f| f.entityID }
	hoverinfo.view_tracking_ref = nil
	
	faces.each do |face|
		@hsh_hoverinfo_faces[hover_key(face)] = hoverinfo
	end
	hoverinfo
end

#HOVERINFO: Compute the frame drawing information if not computed or view changed
def hoverinfo_dessin_update_when_view_changed(hoverinfo, vtref)
	hoverinfo.draw_frame = drawframe_compute(hoverinfo.faces, hoverinfo.tr) unless hoverinfo.draw_frame
	if hoverinfo.view_tracking_ref != vtref
		drawframe_update_from_view(hoverinfo.draw_frame) if hoverinfo.view_tracking_ref != vtref
		hoverinfo.view_tracking_ref = vtref
	end	
end

#HOVERINFO: Get the picked origin and normal
def hoverinfo_get_origin
	return nil unless @initial_face && @initial_face.valid?

	#Finding the picked point on the picked face
	tr_r = @tr.inverse
	ray = @view.pickray @x, @y
	ray = ray.collect { |pt| tr_r * pt }
	origin = @tr * Geom.intersect_line_plane(ray, @initial_face.plane)

	#Computing the normal
	normal = @tr * @initial_face.normal
	normal = normal.reverse unless front_face_is_visible?(@initial_face)
	[origin, normal]
end

#---------------------------------------------------------------------------------------------
# PARENT: Parent Management
#---------------------------------------------------------------------------------------------

#PARENT: Register a parent
def parent_register(parent)
	#Finding the Definition
	if parent.instance_of?(Sketchup::ComponentInstance)
		cdef = parent.definition
	elsif parent.instance_of?(Sketchup::Group)
		cdef = parent.entities.parent
	else
		return nil
	end	
	cdef_id = cdef.entityID
	ls = @hsh_parent_cdef[cdef_id]
	ls = @hsh_parent_cdef[cdef_id] = [] unless ls
	parent_id = parent.entityID
	ls.push parent_id unless ls.include?(parent_id)
	ls
end

#---------------------------------------------------------------------------------------------
# DRAWFRAME: Management of Drawingg information
#---------------------------------------------------------------------------------------------

#DRAWFRAME: Check if the edge is a border
def edge_is_border(edge, hfaces)
	n = 0
	edge.faces.each do |f|
		n += 1 if hfaces[f.entityID]
		return false if n > 1
	end
	true
end

#DRAWFRAME: Compute the frame drawing information
def drawframe_compute(faces, tr, no_face=false)
	#Initialization
	Traductor::HourGlass.start
	
	draw_frames = []
	draw_dashed = []
	draw_contour = []
	draw_triangles = []
	
	#Checking if too many faces. Then only highlight the picked face
	nbmax = 3000
	if faces.length > nbmax
		picked_faces = [faces[0]]
		partial = true
	else	
		picked_faces = faces
		partial = false
	end
	
	#Computing the drawing elements
	hfaces = {}
	picked_faces.each { |f| hfaces[f.entityID] = true }
	
	if G6.su_capa_color_polygon
		picked_faces.each do |face|
			unless no_face
				mesh = face.mesh
				pts = mesh.points
				mesh.polygons.each { |p| draw_triangles += p.collect { |i| tr * pts[i.abs-1] } }
			end	
			unless partial
				face.edges.each do |edge| 
					Traductor::HourGlass.check?
					if edge_is_border(edge, hfaces)
						draw_contour.push(tr * edge.start.position, tr * edge.end.position) 
					end	
				end	
			end	
		end		
	else
		picked_faces.each do |face|
			face.edges.each do |edge| 
				Traductor::HourGlass.check?
				if edge_is_border(edge, hfaces)
					draw_contour.push(tr * edge.start.position, tr * edge.end.position) unless partial
				elsif G6.edge_plain?(edge)
					draw_frames.push(tr * edge.start.position, tr * edge.end.position) 
				else
					draw_dashed.push(tr * edge.start.position, tr * edge.end.position) 
				end	
			end	
		end
	end	
	
	#Storing in a structure
	drframe = DrawFrame.new
	drframe.frames = draw_frames
	drframe.dashed = draw_dashed
	drframe.contour = draw_contour
	drframe.triangles = draw_triangles
	Traductor::HourGlass.stop
	drframe
end

#DRAWFRAME: update the slight shift of drawing lines and polygon based on current view
def drawframe_update_from_view(drframe)
	drframe.ll_frames = drframe.frames.collect { |pt| G6.small_offset(@view, pt) }
	drframe.ll_dashed = drframe.dashed.collect { |pt| G6.small_offset(@view, pt) }
	drframe.ll_contour = drframe.contour.collect { |pt| G6.small_offset(@view, pt) }
	drframe.ll_triangles = drframe.triangles
end

#DRAWFRAME: Draw the Frame information
def drawframe_draw(view, drframe, style)
	if style == :hover
		if @hover_add
			frcolor = @color_hover_frame_add
			bkcolor = @color_hover_face_add
		else	
			frcolor = @color_hover_frame_remove
			bkcolor = @color_hover_face_remove
		end	
	else
		frcolor = @color_selection_frame
		bkcolor = @color_selection_face	
	end
	
	view.drawing_color = frcolor
	view.line_width = 2
	view.line_stipple = ''
	view.draw GL_LINES, drframe.ll_contour unless drframe.ll_contour.empty?
	
	view.drawing_color = bkcolor
	view.draw GL_TRIANGLES, drframe.ll_triangles unless drframe.ll_triangles.empty?
end

#---------------------------------------------------------------------------------------------
# DRAW: Methods to manage drawing
#---------------------------------------------------------------------------------------------

#DRAW: Top method for drawing in the viewport
def draw(view)
	#Drawing hovered faces
	unless @dragging || @sel_modified
		if @current_hoverinfo
			draw_hoverinfo view, @current_hoverinfo
		else	
			draw_single_face(view, @initial_face) if @initial_face && @initial_face.valid?
		end
	end
	
	#Drawing selection
	draw_selection view
	
	#Draw the tooltip for add or remove face
	return if @no_tooltip
	if !@dragging && !@sel_modified && (@current_hiverinfo || @initial_face)
		if @hover_add
			G6.draw_rectangle_multi_text view, @xmove, @ymove, @tip_click_select_faces, @hsh_boxinfo_add
		else	
			G6.draw_rectangle_multi_text view, @xmove, @ymove, @tip_click_unselect_faces, @hsh_boxinfo_remove
		end
	elsif @initial_face
		G6.draw_rectangle_multi_text view, @xmove, @ymove, @text_reselect, @hsh_boxinfo_add if @text_reselect
	end	

end

#DRAW: draw a single face
def draw_single_face(view, face)
	if @hover_add
		frcolor = @color_hover_frame_add
		bkcolor = @color_hover_face_add
	else	
		frcolor = @color_hover_frame_remove
		bkcolor = @color_hover_face_remove
	end	
	lines = []
	face.edges.each { |e| lines.push e.start.position, e.end.position }
	view.drawing_color = frcolor
	view.line_width = 2
	view.line_stipple = ''
	view.draw GL_LINES, lines.collect { |pt| G6.small_offset(view, @tr * pt) }
	G6.draw_face(view, face, bkcolor, @tr)
end

#DRAW: draw multiple faces based on Hover info
def draw_hoverinfo(view, hoverinfo)
	return unless @initial_face && @initial_face.valid?
	bkcolor = (@hover_add) ? @color_hover_face_add_h : @color_hover_face_remove_h
	G6.draw_face(view, @initial_face, bkcolor, @tr)
	hoverinfo_dessin_update_when_view_changed hoverinfo, @viewtracker.reference
	drawframe_draw view, hoverinfo.draw_frame, :hover
end

#DRAW: Draw the selection
def draw_selection(view)
	vtref = @viewtracker.reference
	draw_component view, @parent if @parent && @parent != @model
	@hsh_selection_info.each do |key, selinfo|
		selection_dessin_update_when_view_changed(selinfo, vtref)	
		drawframe_draw view, selinfo.draw_frame, :selection
	end		
end

#DRAW: draw the boundary box of a component
def draw_component(view, parent)
	return unless parent.valid?
	view.line_stipple = ''
	view.line_width = 1
	view.drawing_color = 'gray'
	llines = G6.grouponent_box_lines(view, parent, @tr, 4)
	view.draw GL_LINES, llines unless llines.empty?
end

#---------------------------------------------------------------------------------------------
# SELECTION: Management of Initial Selection
#---------------------------------------------------------------------------------------------

#SELECTION: Init the selection environment
def selection_init
	@hsh_selection_info = {}
	@lst_hoverinfo = []
	@hsh_parent_cdef = {}
	@hsh_faces_selected = {}
	@initial_face = @last_initial_face = nil
end

#SELECTION: Remove all selection
def selection_unselect_all
	selection_state_save if @hsh_faces_selected.length > 0
	selection_init
	history_init
	@selection.clear
end

#SELECTION: Validate selection from hovering if this corresponds to Add
def selection_validate_if_add
	selection_validate if @hover_add
end

#SELECTION: Save the whole state of selection for future reselection
def selection_state_save
	@hsh_faces_selected_save = @hsh_faces_selected
	@lst_hoverinfo_save = @lst_hoverinfo
	@hsh_selection_info_save = @hsh_selection_info
	@hsh_parent_cdef_save = @hsh_parent_cdef
	@history_save = @history
	@ipos_history_save = @ipos_history
	@initial_face_id_save = @initial_face_id
end

#SELECTION: Restore the whole previous state of selection
def selection_state_restore
	@hsh_faces_selected = @hsh_faces_selected_save
	@lst_hoverinfo = @lst_hoverinfo_save
	@hsh_selection_info = @hsh_selection_info_save
	@hsh_parent_cdef = @hsh_parent_cdef_save
	@history = @history_save
	@ipos_history = @ipos_history_save
	@initial_face_id = @initial_face_id_save
end

#SELECTION: Return the currently selected groups of faces
def selection_get_groupings
	lst_groups = []
	@hsh_selection_info.each do |key, selinfo|
		lst_groups.push [selinfo.hfaces, selinfo.tr, selinfo.parent] if selinfo.hfaces.length > 0
	end		
	lst_groups
end

#SELECTION: Return the currently selected faces, groups and components
def selection_get_nb_info
	nbfaces = @hsh_faces_selected.length
	ngroups = ncomps = 0
	@hsh_selection_info.each do |id, selinfo| 
		parent = selinfo.parent
		ngroups += 1 if parent.instance_of?(Sketchup::Group)		
		ncomps += 1 if parent.instance_of?(Sketchup::ComponentInstance)
	end
	[nbfaces, ngroups, ncomps]
end
	
#SELECTION: ramapping of all faces used based on their ID (after undo)
def selection_remapping_faces_when_geometry(hmap_master)
	@initial_face = hmap_master[@initial_face_id]
end

#SELECTION: ramapping of all faces used based on their ID (after undo)
def selection_remapping_faces_when_undo(hmap_master)
	
	#Restore the selection state
	selection_state_restore
	@initial_face = hmap_master[@initial_face_id]
	
	#Selection buckets of faces
	@hsh_selection_info.each do |key, selinfo|
		hfaces = selinfo.hfaces
		hfaces.each { |face_id, face| hfaces[face_id] = hmap_master[face_id] }
	end		
	
	#Hoverinfo structures
	@lst_hoverinfo.each do |hoverinfo|
		faces = hoverinfo.faces
		hoverinfo.faces_id.each_with_index { |face_id, i| faces[i] = hmap_master[face_id] }
	end
	
	#List of selected faces
	@hsh_faces_selected.each { |face_id, face| @hsh_faces_selected[face_id] = hmap_master[face_id] }
	vfaces = @hsh_faces_selected.values
	@selection.add vfaces unless vfaces.empty?
	
	#History
	@history.each do |ls|
		code, faces, tr, parent, text, faces_id = ls
		faces_id.each_with_index do |face_id, i|
			faces[i] = hmap_master[face_id]
		end
	end	
end

#SELECTION: Return the currently selected groups of faces
def selection_get_picked_face_info
	[@initial_face, @tr, @parent]
end
	
#SELECTION: Handle initial selection
def selection_initial(selection=nil, parent=nil, tr=nil)
	selection = @model.selection unless selection
	@focus_initial_selection = false
	return if selection.length == 0
	parent = @model unless parent
	tr = @tr_id unless tr
	hsh_selection_faces = {}
	selection.each do |e|
		if e.instance_of?(Sketchup::Face)
			hsh_selection_faces[e.entityID] = e
		elsif e.instance_of?(Sketchup::Group)
			faces = e.entities.grep(Sketchup::Face)
			selection_add_remove_faces true, faces, e.transformation, e
		elsif e.instance_of?(Sketchup::ComponentInstance)
			faces = e.definition.entities.grep(Sketchup::Face)
			selection_add_remove_faces true, faces, e.transformation, e
		end	
	end
	@focus_initial_selection = (hsh_selection_faces.length > 0)
	selection_add_remove_faces true, hsh_selection_faces.values, tr, parent
end

#SELECTION: Check if the hovered faces are for ADD or REMOVE
def selection_add?(faces)
	faces.find { |face| !@hsh_faces_selected[face.entityID] }
end

#SELECTION: Validate the selection if any
def selection_validate
	if @current_hoverinfo
		selection_add_remove_faces @hover_add, @current_hoverinfo.faces, @current_hoverinfo.tr, @current_hoverinfo.parent
	elsif @initial_face	
		selection_add_remove_faces @hover_add, [@initial_face], @tr, @parent
	else
		notify_selection :outside	
		return false
	end	
	notify_selection :selection
	true
end

#SELECTION: Check if group should be made unique
def selection_check_group_unique(faces, tr, parent)
	return unless @make_group_unique_proc
	return [faces, tr, parent] unless parent.instance_of?(Sketchup::Group) && !G6.grouponent_unique?(parent)
	
	#Calculating the remapping of faces before making it unique
	hmap = G6.grouponent_map_of_entities_object_id(parent)
	
	#Making the group unique
	@make_group_unique_proc.call parent
	
	#Remapping the faces
	entities = G6.grouponent_entities(parent)
	new_faces = []
	faces.each do |face|
		new_face = entities[hmap[face.object_id]]
		new_faces.push new_face
		if face == @initial_face
			@initial_face = @last_initial_face = new_face
			@initial_face_id = new_face.entityID
		end	
	end
	
	#Resetting the hoverinfo information
	@current_hoverinfo = nil
	@hsh_hoverinfo_faces = {}
	@lst_hoverinfo = []
	
	#Returning the new faces
	[new_faces, tr, parent]
end

#SELECTION: add faces to a selection
def selection_add_remove_faces(flg_add, faces, tr, parent, nostore=false)
	#Making group unique if required
	faces, tr, parent = selection_check_group_unique faces, tr, parent
		
	#New selection info structure
	key = parent.entityID
	@sel_modified = true
	selinfo = @hsh_selection_info[key]
	unless selinfo
		selinfo = @hsh_selection_info[key] = SelInfo.new
		selinfo.parent = parent
		selinfo.tr = tr
		hfaces = selinfo.hfaces = {}
		if flg_add
			modified_faces = selection_add_faces_to_selinfo(selinfo, faces)
			return nil if modified_faces.empty?
			history_store(:add, modified_faces, tr, parent, @tip_history_add_faces) unless nostore
			selection_synchronize_instances selinfo
			return [modified_faces, tr, parent]
		else
			faces.each { |face| hfaces[face.entityID] = face }
		end
	end	
	
	#Existing selection info structure
	if flg_add
		modified_faces = selection_add_faces_to_selinfo(selinfo, faces)
		return nil if modified_faces.empty?
		history_store(:add, modified_faces, tr, parent, @tip_history_add_faces) unless nostore
	else
		modified_faces = selection_remove_faces_from_selinfo(selinfo, faces)
		return nil if modified_faces.empty?
		history_store(:remove, modified_faces, tr, parent, @tip_history_remove_faces) unless nostore
	end
		
	#Updating other instances
	selection_synchronize_instances selinfo
	[modified_faces, tr, parent]
end

#SELECTION: Add a list of faces to a selection grouping
def selection_add_faces_to_selinfo(selinfo, faces)
	new_faces = []
	hfaces = selinfo.hfaces
	faces.each do |face|
		face_id = face.entityID
		next if hfaces[face_id]
		hfaces[face_id] = face
		@hsh_faces_selected[face_id] = face
		new_faces.push face
	end
	@selection.add faces
	selinfo.draw_frame = nil
	new_faces
end

#SELECTION: Add a list of faces to a selection grouping
def selection_remove_faces_from_selinfo(selinfo, faces)
	new_faces = []
	hfaces = selinfo.hfaces
	faces.each do |face|
		face_id = face.entityID
		next unless hfaces[face_id]
		hfaces.delete face_id
		@hsh_faces_selected.delete face_id
		new_faces.push face
	end
	@selection.remove faces
	selinfo.draw_frame = nil
	new_faces
end

#SELECTION: Synchronize the selection across components and groups with same definition
def selection_synchronize_instances(selinfo_master)
	parent = selinfo_master.parent
	return if parent == @model
	ls_pid = parent_register parent
	return if ls_pid.length < 2
	
	#Consolidating the list of faces
	lst_selinfo = ls_pid.collect { |pid| @hsh_selection_info[pid] }
	hfaces = selinfo_master.hfaces
	lst_selinfo.each { |selinfo| selinfo.hfaces.each { |key, val| hfaces[key] = val } } if @hover_add
	
	#Synchronizing all instances
	lst_selinfo.each do |selinfo|
		selinfo.hfaces = hfaces
		selinfo.draw_frame = nil
	end
end

#SELECTION: Compute the frame drawing information if not computed or view changed
def selection_dessin_update_when_view_changed(selinfo, vtref)
	if !selinfo.draw_frame
		selinfo.draw_frame = drawframe_compute(selinfo.hfaces.values, selinfo.tr, true)
		drawframe_update_from_view(selinfo.draw_frame)
	elsif selinfo.view_tracking_ref != vtref
		drawframe_update_from_view(selinfo.draw_frame) if selinfo.view_tracking_ref != vtref
		selinfo.view_tracking_ref = vtref
	end	
end

#---------------------------------------------------------------------------------------------
# KEYBOARD: Keyboard management
#---------------------------------------------------------------------------------------------

def handle_key_down(key)
	history_navigate(key)
end

def handle_key_up(key)
	false
end

#---------------------------------------------------------------------------------------------
# CLICK: Click management
#---------------------------------------------------------------------------------------------

#CLICK: Button click DOWN
def onLButtonDown(flags, x, y, view)
	@button_down = true
	@xdown = x
	@ydown = y
	@already_selected = (@initial_face && @hover_add && @hsh_faces_selected[@initial_face.entityID])
end

#CLICK: Button click UP
def onLButtonUp(flags, x, y, view)
	@button_down = false
	handle_move(flags, x, y, view)	
	selection_validate
	if @initial_face && selection_add?([@initial_face])
		@last_initial_face = @current_hoverinfo = nil
		@sel_modified = false
	end	
	if @already_selected
		notify_selection :reselect
	end	
	@already_selected = false
end

#CLICK: Double Click received
def onLButtonDoubleClick(flags, x, y, view)
	
end

#------------------------------------------------------------------------------------
# DRAGGING: Dragging Management (for future development)
#------------------------------------------------------------------------------------

def dragging_start
	@dragging = true
	#@notify_proc.call :stop_dragging if @notify_proc
end	

def dragging_stop
	@dragging = false
	#@notify_proc.call :stop_dragging if @notify_proc
end	

#------------------------------------------------------------------------------------
# HISTORY: Management of History of selection
#------------------------------------------------------------------------------------
	
#HISTORY: Construct the initial history environment
def history_init
	@history = []
	@ipos_history = 0
	@tip_navig_down = Traductor.encode_tip T6[:T_MNU_History_Clear], :ctrl_arrow_down
	@tip_navig_up = Traductor.encode_tip T6[:T_MNU_History_Restore], :ctrl_arrow_up
	history_compute_texts
end

#HISTORY: Store action for History navigation
def history_store(code, faces, tr, parent, text=nil)
	@history[@ipos_history..-1] = []
	faces_id = faces.collect { |f| f.entityID }
	@history.push [code, faces, tr, parent, text, faces_id]
	@ipos_history += 1
	history_compute_texts
end

#HISTORY: Compute the texts for menu and tooltips
def history_compute_texts	
	if @ipos_history > 0
		txundo = ' --> ' + @history[@ipos_history-1][4]
	else
		txundo = ''
	end	
	@tip_navig_left = Traductor.encode_tip T6[:T_MNU_History_Last, txundo], [:escape, :ctrl_arrow_left]
	
	if @ipos_history < @history.length
		txredo = ' --> ' + @history[@ipos_history][4]
	else
		txredo = ''
	end	
	@tip_navig_right = Traductor.encode_tip T6[:T_MNU_History_Next, txredo], :ctrl_arrow_right
end

#HISTORY: Undo one step from history
def history_undo
	return UI.beep if @ipos_history == 0
	@ipos_history -= 1
	code, faces, tr, parent = @history[@ipos_history]
	case code
	when :add
		selection_add_remove_faces(false, faces, tr, parent, true)
	when :remove
		selection_add_remove_faces(true, faces, tr, parent, true)
	else
		return
	end	
	history_compute_texts
end

#HISTORY: Redo one step from history
def history_redo
	return UI.beep if @ipos_history == @history.length
	code, faces, tr, parent = @history[@ipos_history]
	case code
	when :add
		selection_add_remove_faces(true, faces, tr, parent, true)
	when :remove
		selection_add_remove_faces(false, faces, tr, parent, true)
	else
		return
	end	
	@ipos_history += 1
	history_compute_texts
end

#HISTORY: Clear all from history
def history_clear
	#return UI.beep if @ipos_history == 0
	return if @ipos_history == 0
	for i in 1..@ipos_history
		history_undo
	end	
end

#HISTORY: Restore all from history
def history_restore
	len = @history.length
	return UI.beep if @ipos_history == len
	for i in @ipos_history..len-1
		history_redo
	end	
end

#HISTORY: Notification proc for action of History buttons
def history_proc_action(code)
	case code
	when :undo
		history_undo
	when :redo
		history_redo
	when :clear
		history_clear
	else
		history_restore
	end
end

#HISTORY: Notification proc for tooltips of History buttons
def history_proc_tip(code)
	case code
	when :undo
		@tip_navig_left
	when :redo
		@tip_navig_right
	when :clear
		@tip_navig_down
	else
		@tip_navig_up
	end
end

#HISTORY: Notification proc for state of History buttons
def history_proc_gray(code)
	case code
	when :undo, :clear
		@ipos_history == 0
	when :redo, :restore
		@ipos_history >= @history.length
	else
		false
	end
end

#HISTORY: History_navigation
def history_navigate(key)
	case key
	when VK_LEFT
		history_undo
	when VK_RIGHT
		history_redo
	when VK_DOWN
		history_clear
	when VK_UP
		history_restore
	else
		return false
	end
	true
end

#HISTORY: Check if Undo is possible
def history_undo_possible?
	@ipos_history > 0
end

#------------------------------------------------------------------------------------
# MENU: Contribution to Contextual Menu
#------------------------------------------------------------------------------------

#Contextual menu
def contextual_menu_contribution(cxmenu)
	#Face Selection options
	cxmenu.add_sepa
	cxmenu.add_item(@mnu_selection_single) { set_option_face_selection :single } unless @option_face_selection == :single
	cxmenu.add_item(@mnu_selection_surface) { set_option_face_selection :surface } unless @option_face_selection == :surface
	cxmenu.add_item(@mnu_selection_connected) { set_option_face_selection :connected } unless @option_face_selection == :connected
	cxmenu.add_item(@mnu_selection_samecolor) { set_option_face_selection :same_color } unless @option_face_selection == :same_color
	
	#History
	cxmenu.add_sepa
	mnu_navig_left = Traductor.encode_menu @tip_navig_left
	mnu_navig_right = Traductor.encode_menu @tip_navig_right	
	mnu_navig_down = Traductor.encode_menu @tip_navig_down
	mnu_navig_up = Traductor.encode_menu @tip_navig_up
	cxmenu.add_item(mnu_navig_left) { history_undo } unless history_proc_gray(:undo)
	cxmenu.add_item(mnu_navig_right) { history_redo } unless history_proc_gray(:redo)
	cxmenu.add_item(mnu_navig_down) { history_clear } unless history_proc_gray(:clear)
	cxmenu.add_item(mnu_navig_up) { history_restore } unless history_proc_gray(:restore)
end

#------------------------------------------------------------------------------------
# PALETTE: Palette contribution
#------------------------------------------------------------------------------------

#PALETTE: palette contribution from the Face Picker
def palette_contribution(palette, *hargs)
	#Defaults
	@pal_bk_color = 'lightblue'
	
	#Parsing the arguments
	hargs.each do |harg|
		next unless harg.class == Hash
		harg.each do |key, value|  
			case key
			when :grayed_proc
				@super_grayed_proc = value
			when :bk_color
				@pal_bk_color = value
			end
		end	
	end

	#Creating the palettes
	palette_history palette
	palette_face_options palette
end

#PALETTE: History navigation
def palette_history(palette)
	palette.declare_separator
	tip_proc = proc { |code| history_proc_tip code }
	gray_proc = proc { |code| history_proc_gray(code) || @super_grayed_proc.call }
	hsh = { :grayed_proc => gray_proc, :tip_proc => tip_proc, :compact => true, :key_ctrl => true }
	palette.declare_historical(:t_group_ep_history, hsh) { |code| history_proc_action code }
end

#PALETTE: Options for Face Selection
def palette_face_options(palette)
	palette.declare_separator
	
	#Master Button
	box_text = T6[:T_BOX_FaceSelection]
	w, = G6.simple_text_size(box_text)
	wid = [w / 4, 24].max
	symb_master = :pal_fpick_face_selection
	tip = T6[:T_TIP_ClickDefaultOption] + ' - ' + T6[:T_TIP_TABCycleOption]
	hsh = { :type => 'multi_free', :text => box_text, :bk_color => @pal_bk_color, :grayed_proc => @super_grayed_proc,
	        :height => 16, :tooltip => tip }
	palette.declare_button(symb_master, hsh) { set_option_face_selection }

	hshp = { :parent => symb_master, :width => wid, :draw_proc => @draw_local, :grayed_proc => @super_grayed_proc }
	
	#Single Face
	hsh = { :value_proc => (proc { @option_face_selection == :single }), :tooltip => T6[:T_TIP_FaceSelectionSingle],
            :draw_proc => :std_face_selection_single }
	palette.declare_button(:pal_fpick_face_selection_single, hshp, hsh) { set_option_face_selection :single }
	
	#Surface
	hsh = { :value_proc => (proc { @option_face_selection == :surface }), :tooltip => T6[:T_TIP_FaceSelectionSurface],
	        :draw_proc => :std_face_selection_surface }
	palette.declare_button(:pal_fpick_face_selection_surface, hshp, hsh) { set_option_face_selection :surface }
	
	#All Connected
	hsh = { :value_proc => (proc { @option_face_selection == :connected }), :tooltip => T6[:T_TIP_FaceSelectionConnected],
	        :draw_proc => :std_face_selection_connected }
	palette.declare_button(:pal_fpick_face_selection_connected, hshp, hsh) { set_option_face_selection :connected }
	
	#Same Material
	hsh = { :value_proc => (proc { @option_face_selection == :same_color }), :tooltip => T6[:T_TIP_FaceSelectionSameColor],
	        :draw_proc => :std_face_selection_same_color }
	palette.declare_button(:pal_fpick_face_selection_same_color, hshp, hsh) { set_option_face_selection :same_color }
	
end

end	#class FacePicker

end	#End Module Traductor
