=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Copyright © 2014 Fredo6 - Designed and written Mar 2014 by Fredo6
#
# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for commercial purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:   body_Lib6OriginTargetPicker.rb
# Original Date	:   11 Mar 2014 
# Description	:   Manage the picking of an origin and a target with second-level inferences
# IMPORTANT		:	DO NOT TRANSLATE STRINGS in the source code
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end

module Traductor

#-------------------------------------------
# Text for tooltips of Inferences
#-------------------------------------------

T6[:INFER_InGroup] = "in Group"
T6[:INFER_InComponent] = "in Component"
T6[:INFER_InImage] = "in Image"
T6[:INFER_FromRemote] = "from Inference point"

T6[:INFER_Cpoint] = "Guide Point"
T6[:INFER_Cline] = "On Guide Line"
T6[:INFER_EndCline] = "End of Guide Line"
T6[:INFER_Endpoint] = "Endpoint"
T6[:INFER_Midpoint] = "Midpoint"
T6[:INFER_Twin] = "Intersection of remote directions"
T6[:INFER_CenterFace] = "Center of Face"
T6[:INFER_CenterArc] = "Center of arc"
T6[:INFER_CenterComp] = "Center of Component"
T6[:INFER_CenterCompTop] = "Center of Top Component"
T6[:INFER_CenterGroup] = "Center of Group"
T6[:INFER_CenterGroupTop] = "Center of Top Group"
T6[:INFER_CenterPolygon] = "Center of Circle/Polygon"
T6[:INFER_OnEdge] = "on Edge"
T6[:INFER_OnFace] = "on Face"
T6[:INFER_OnSectionPlane] = "on Section plane"
T6[:INFER_AtOrigin] = "at Origin"
T6[:INFER_OnXAxis] = "on X axis"
T6[:INFER_OnYAxis] = "on Y axis"
T6[:INFER_OnZAxis] = "on Z axis"
T6[:INFER_AlongXAxis] = "Along X axis"
T6[:INFER_AlongYAxis] = "Along Y axis"
T6[:INFER_AlongZAxis] = "Along Z axis"
T6[:INFER_AlongXAxisComp] = "Along X axis of Component"
T6[:INFER_AlongYAxisComp] = "Along Y axis of Component"
T6[:INFER_AlongZAxisComp] = "Along Z axis of Component"
T6[:INFER_PerpEdgeFace] = "Perpendicular to Edge on Face"
T6[:INFER_PerpEdgeAlone] = "Perpendicular to Edge"
T6[:INFER_PerpBisector] = "Bisector line"
T6[:INFER_AlongEdgeDirection] = "Along Edge direction"
T6[:INFER_PerpTwoEdges] = "Perpendicular to two Edges"
T6[:INFER_AlongCline] = "Along Guide line"
T6[:INFER_AlongAxesBbox] = "Along axes of Bounding Box"
T6[:INFER_AlongNormalFace] = "Along Normal to Face"
T6[:INFER_AlongSectionPlane] = "Along Section plane"
T6[:INFER_InterEdgeFace] = "Intersection Edge - Face"
T6[:INFER_InterClineFace] = "Intersection Guide Line - Face"
T6[:INFER_InterClineEdge] = "Intersection Guide Line - Edge"
T6[:INFER_InterClineCline] = "Intersection Guide Line - Guide Line"
T6[:INFER_InterVectorFace] = "Intersection Direction and Face"
T6[:INFER_InterVectorSectionPlane] = "Intersection Direction and Section plane"
T6[:INFER_InterVectorEdge] = "Intersection Direction and Edge"
T6[:INFER_InterVector] = "Intersection with Direction"
T6[:INFER_InterVectorRemote] = "Intersection Direction with Inference direction"
T6[:INFER_InterPlaneRemote] = "Intersection Plane with Inference direction"
T6[:INFER_InterPlaneEdge] = "Intersection Plane and Edge"
T6[:INFER_TextAnchor] = "Text Anchor"
T6[:INFER_OnTextLabel] = "on Text Label"
T6[:INFER_OnTextLeader] = "on Text Leader"
T6[:INFER_OnDimension] = "on Dimension"
T6[:INFER_OnDrawingElement] = "on Drawing Element"
T6[:INFER_AtCornerBbox] = "at Corner of Bounding Box"
T6[:INFER_OnFaceBbox] = "on Face of Bounding Box"
T6[:INFER_ParallelToEdge] = "Parallel to Reference Edge"
T6[:INFER_ParallelToCLine] = "Parallel to Reference Guide line"
T6[:INFER_ProtractorAngle] = "Specified angle"
T6[:INFER_SpecifiedPoint] = "Specified point"
T6[:TIP_ResetInference] = "Reset all Inference Marks"

T6[:TIP_Vector_Title] = "Constrain Move by a Direction specified by its vector"
T6[:TIP_Planar_Title] = "Constrain Move by a Plane specified by its normal"
T6[:TIP_ToggleInferenceComp] = "Toggle inference on Components"
T6[:TIP_ToggleInferenceRemote] = "Toggle inference from Remote points"
T6[:TIP_NoInference] = "Toggle All Inferences"
T6[:TIP_PickPlaneFromModel] = "Pick Plane from model"
T6[:TIP_PickVectorFromModel] = "Pick Direction from model"

#=============================================================================================
#=============================================================================================
# Class OriginTargetPicker: Class for picking an Origin and a Target with inferences
#=============================================================================================
#=============================================================================================

class OriginTargetPicker

#INIT: Class instance initialization
def initialize__(hsh_parameters, *hargs)
	#Initialization
	@tr_id = Geom::Transformation.new
	@model = Sketchup.active_model
	@selection = @model.selection
	@rendering_options = Sketchup.active_model.rendering_options
	@view = @model.active_view
	@ph = @view.pick_helper	
	@ip = Sketchup::InputPoint.new
	@ogl = Traductor::OpenGL_6.new	
	@mode = :origin
	@no_inference = false
	
	#Initialization
	init_colors
	
	#Parsing the arguments
	hargs.each do |harg|	
		harg.each { |key, value|  parse_args(key, value) } if harg.class == Hash
	end
	
	#Init texts and messages
	init_text
		
	#Precision parameters
	@precision_default = 8
	@inference_angle_default = 8.degrees
	@remote_delay_origin_default = 1.0
	@remote_delay_target_default = 1.0
	@precision = MYDEFPARAM[:DEFAULT_OTPicker_PixelPrecision]
	@inference_angle = MYDEFPARAM[:DEFAULT_OTPicker_AlignAngle].degrees
	@inference_distance = [3 * @precision, 25].min
	@inference_distance = 15 if @inference_distance < 15
	@remote_delay_origin = MYDEFPARAM[:DEFAULT_OTPicker_RemoteDelayOrigin]
	@remote_delay_target = MYDEFPARAM[:DEFAULT_OTPicker_RemoteDelayTarget]
	
	@ip_origin = Sketchup::InputPoint.new unless @ip_origin
	@ip_target = Sketchup::InputPoint.new unless @ip_target
	
	#Initialization
	mark_init
	
	#Setting the directions and other options
	hsh_parameters = {} unless hsh_parameters
	@option_planar = hsh_parameters.fetch :option_planar, false
	@option_planar_code = hsh_parameters.fetch :option_planar_code, false
	@option_vector = hsh_parameters.fetch :option_vector, :no
	@option_vector_code = hsh_parameters.fetch :option_vector_code, :no
	@option_inference_comp = hsh_parameters.fetch :option_inference_comp, true
	@option_inference_remote = hsh_parameters.fetch :option_inference_remote, true
	@option_vector_code = :no unless @option_planar_code
	@option_vector_code = :no unless @option_vector_code

	direction_changed
	
	#Creating the protractor shape
	@protractor = G6::ShapeProtractor.new	
end

#INIT: Parse the parameters of the Face Picker
def parse_args(key, value)
	skey = key.to_s
	case skey
	when /precision/i
		@precision = value
	when /inference_angle/i
		@inference_angle = value.degrees
	when /color_box_distance/
		@color_box_distance = value
	when /ignore_selection/i
		@ignore_selection = value
	when /line_mode/
		@line_mode = value
	when /implicit_plane/
		@option_implicit_plane = (value && value.class != Array) ? [value] : value
	when /notify_proc/
		@notify_caller_proc = value
	when /hide_distance/
		@hide_distance = value
	end	
end	

#INIT: Set the precision and inference angle
def set_precision(precision) ; @precision = (precision) ? precision : @precision_default ; end
def get_precision ; @precision ; end
def set_inference_angle(angle) ; @inference_angle = (angle) ? angle.degrees : @inference_angle_default ; end
def set_remote_delay_origin(delay) ; @remote_delay_origin = (delay) ? delay : @remote_delay_origin_default ; end
def set_remote_delay_target(delay) ; @remote_delay_target = (delay) ? delay : @remote_delay_target_default ; end
def remote_delay ; (@mode == :origin) ? @remote_delay_origin : @remote_delay_target ; end

#INIT: Simulate a move of mouse by calling the caller method
def onMouseMove_zero
	if @notify_caller_proc
		notify_caller :onMouseMove_zero
	else
		@view.invalidate
	end
end
	
#INIT: Set the option for implicit plane
def set_option_implicit_plane(loption)	
	@option_implicit_plane = loption
end

def option_implicit_plane_check(val)
	return false unless @option_implicit_plane
	@option_implicit_plane.include?(val)
end
	
#INIT: Set the current mode, Origin or Target
def set_mode(mode)
	@mode = mode
end

#INIT: Set the current mode, Origin or Target
def set_stipple_color_width(stipple=nil, color=nil, width=nil)
	@line_stipple = (stipple) ? stipple : ''
	@line_color = (color) ? color : 'black'
	@line_width = (width) ? width : 1
end
	
#INIT: reset the environment
def reset
	@pt_origin = @origin_info = nil
	@pt_target = @target_info = nil
	@pt_target_prev = nil
	@enter_unreal = false
	@trxy = @tr_id
	@last_plane_found = nil
	@implicit_plane_normal = nil
	remote_reset
	@hi_curve_pts = nil
	@top_parent = @info_parent = nil
	@direction_lock = false
	@protractor_showable = false
	@protractor_direction = nil
	inference_reset
end
	
#INIT: Initialize colors
def init_colors
	#Remote inference
	@color_twin = 'orange'
	@color_remote_fr = 'gold'
	@color_remote = Sketchup::Color.new @color_remote_fr
	@color_remote.alpha = 0.4
	@color_box_distance_remote = ['gold', 'goldenrod']
	
	@line_stipple = ''
	@line_color = 'black'
	@line_width = 1	
	
	#Colors for directions
	lst_color_dir = [ [:free, 'black'], [:lock, 'black'], [:follow_bbox, 'slategray'], [:remote, @color_remote_fr],
	                  [:follow_x, 'red'], [:follow_y, 'green'], [:follow_z, 'blue'], [:perp_edge_face, 'brown'], 
					  [:perp_edge_alone, 'brown'], [:perp_bisector, 'blueviolet'],
					  [:follow_edge, 'magenta'], [:follow_cline, 'purple'], [:follow_normal, 'darkorange'],
					  [:follow_x_comp, 'tomato'], [:follow_y_comp, 'limegreen'], [:follow_z_comp, 'skyblue'],
					  [:plane_of_edges, 'goldenrod'], [:protractor, 'plum'], [:implicit_plane, 'silver'],
					  [:follow_section_plane, 'lightblue']]
	@color_box_distance = ['lightgrey', 'black']
	
	#Color for palette buttons
	@color_but_title = 'lightgrey'
	@color_but_hi = 'lightgreen'	

	#Color for vectors and planes
	@hsh_color_vector = Hash.new 'black'
	@hsh_bkcolor_plane = Hash.new 'gray'
	@hsh_frcolor_plane = Hash.new 'black'
	lst_color_dir.each do |symb, color|
		@hsh_color_vector[symb] = color
		pcolor = Sketchup::Color.new color
		pcolor.alpha = 0.1
		@hsh_bkcolor_plane[symb] = pcolor
		@hsh_frcolor_plane[symb] = color
	end
	
end

#INIT: Initialize texts and messages
def init_text
	@tip_plane_z = T6[:T_TIP_Plane_Z] + " [#{T6[:T_KEY_ArrowUp]}]"
	@tip_plane_x = T6[:T_TIP_Plane_X] + " [#{T6[:T_KEY_ArrowRight]}]"
	@tip_plane_y = T6[:T_TIP_Plane_Y] + " [#{T6[:T_KEY_ArrowLeft]}]"
	@tip_plane_no = T6[:T_TIP_Plane_None] + " [ #{T6[:T_KEY_ArrowDown]}]"
	@tip_vector_z = T6[:T_TIP_Vector_Z] + " [#{T6[:T_KEY_ArrowUp]}]"
	@tip_vector_x = T6[:T_TIP_Vector_X] + " [#{T6[:T_KEY_ArrowRight]}]"
	@tip_vector_y = T6[:T_TIP_Vector_Y] + " [#{T6[:T_KEY_ArrowLeft]}]"
	@tip_vector_no = T6[:T_TIP_Vector_None] + " [#{T6[:T_KEY_ArrowDown]}]"
	@tip_reset_inference = T6[:TIP_ResetInference] + " [#{T6[:T_KEY_Backspace]}]"	
	@tip_no_inference = T6[:TIP_NoInference] + " [#{T6[:T_KEY_ALT]}]"
	skey = T6[:T_KEY_Enter]
	skey += ", " + T6[:T_KEY_Shift] if @line_mode
	@tip_pick_plane = T6[:TIP_PickPlaneFromModel] + " [#{skey}]"
	@tip_pick_vector = T6[:TIP_PickVectorFromModel] + " [#{skey}]"

	@txt_offset = (@line_mode) ? T6[:T_TXT_Length] : T6[:T_TXT_Offset]	
	@txt_distance = T6[:T_TXT_Distance]	
end

#---------------------------------------------------------
# NOTIFY: Notification call back
#---------------------------------------------------------
	
#NOTIFY: Call the notification method of the caller class	
def notify_caller(event, default=nil)
	return default unless @notify_caller_proc
	val = @notify_caller_proc.call(event)
	return default if val == :no_event
	val
end

#---------------------------------------------------------
# PRESELECTION: Manage preselection
#---------------------------------------------------------
	
#PRESELECTION: set a preselection for the picker tool	
def preselection_set(preselection)
	@active_selection = @preselection = preselection
	@cumulative_selection = nil
	suselection_add(@preselection, true)
	@preselection
end

#PRESELECTION: Get the preselection (nil if none)
def preselection_get
	@preselection || @cumulative_selection
end

#PRESELECTION: Restore the preselection if any
def preselection_restore
	@preselection = @cumulative_selection unless @preselection	
	return nil unless @preselection
	preselection = @preselection.find_all { |e| e.valid? }
	preselection_set preselection
end
	
#PRESELECTION: Cumulate the selection
def preselection_cumulate
	@cumulative_bit = @selection.to_a.clone
	@cumulative_selection = @selection.to_a
end

#PRESELECTION: Cumulate the selection
def preselection_cumulate_undo
	return unless @cumulative_selection && @cumulative_bit
	@cumulative_selection = @cumulative_selection - @cumulative_bit
	@cumulative_selection = nil if @cumulative_selection.empty?
	@cumulative_bit = nil
end

#PRESELECTION: Set the preselection when there is a cumulative selection
def preselection_set_when_cumulative
	if @cumulative_selection
		suselection_add @cumulative_selection, true
		preselection_cumulate
	end	
end

#---------------------------------------------------------
# MARK: Manage marks
#---------------------------------------------------------

#MARK: Initialize the instructions for marks (at origin)
def mark_init
	@hsh_marks = {}
	@color_comp = 'mediumorchid'
	@marks_keep_color = [:cpoint, :on_cline, :end_cline, :on_text, :text_leader, :text_anchor, :center_face, 
	                     :center_comp_mark, :center_comp_mark_top, :center_comp, :center_comp_top, :center_group, :center_group_top,
						 :bbox_corner, :bbox_face, :inter_plane_edge, :inter_vector_face, :inter_vector_edge, :inter_vector,
						 :inter_vector_remote, :inter_plane_remote, :inter_vector_section]
	@bbox_lines_indexing = [0, 1, 1, 3, 3, 2, 2, 0, 0, 4, 1, 5, 2, 6, 3, 7, 4, 5, 5, 7, 6, 7, 4, 6]
	
	#Default mark
	dec = 3
	pts = G6.pts_square 0, 0, dec
	@hsh_marks[nil] = [[GL_POLYGON, pts, 'darkred']]

	#Mark for void origin
	dec = 3
	pts = G6.pts_square 0, 0, dec
	@hsh_marks[:void_origin] = [[GL_POLYGON, pts, 'green']]

	#Mark for void target via remoting
	dec = 3
	pts = G6.pts_square 0, 0, dec
	@hsh_marks[:remote] = [[GL_POLYGON, pts, @color_remote_fr]]

	#Mark on Face
	dec = 6
	pts = [[dec, 0], [0, dec], [-dec, 0], [0, -dec]].collect { |x, y| Geom::Point3d.new(x, y, 0) }
	@hsh_marks[:on_face] = [[GL_POLYGON, pts, 'blue'], [GL_LINE_LOOP, pts, 'white', 1, '']]

	#Mark on Section Plane
	dec = 6
	pts = [[dec, 0], [0, dec], [-dec, 0], [0, -dec]].collect { |x, y| Geom::Point3d.new(x, y, 0) }
	@hsh_marks[:on_section_plane] = [[GL_POLYGON, pts, 'lightblue'], [GL_LINE_LOOP, pts, 'white', 1, '']]

	#Mark on Edge
	dec = 4
	pts = G6.pts_square 0, 0, dec
	@hsh_marks[:on_edge] = [[GL_POLYGON, pts, 'red'], [GL_LINE_LOOP, pts, 'white', 1, '']]
	
	#Mark on Text
	dec = 6
	pts = [[dec, 0], [0, dec], [-dec, 0], [0, -dec]].collect { |x, y| Geom::Point3d.new(x, y, 0) }
	@hsh_marks[:on_text] = [[GL_POLYGON, pts, 'orange'], [GL_LINE_LOOP, pts, 'yellow', 1, '']]

	#Mark on Text
	dec = 4
	pts = G6.pts_square 0, 0, dec
	@hsh_marks[:text_leader] = [[GL_POLYGON, pts, 'orange'], [GL_LINE_LOOP, pts, 'yellow', 1, '']]
	
	#Mark on Guide line
	dec = 4
	pts = G6.pts_square 0, 0, dec
	@hsh_marks[:on_cline] = [[GL_POLYGON, pts, 'darkgray'], [GL_LINE_LOOP, pts, 'white', 1, '']]

	#Mark at extremity of Guide line
	dec = 5
	pts = G6.pts_circle 0, 0, dec
	@hsh_marks[:end_cline] = [[GL_POLYGON, pts, 'darkgray'], [GL_LINE_LOOP, pts, 'white', 1, '']]
	
	#Mark at Midpoint
	dec = 5
	pts = G6.pts_circle 0, 0, dec
	pts2 = [[-dec, 0], [dec, 0], [0, -dec], [0, dec]].collect { |x, y| Geom::Point3d.new(x, y, 0) }
	@hsh_marks[:midpoint] = [[GL_POLYGON, pts, 'red'], [GL_LINE_LOOP, pts, 'white', 1, ''], [GL_LINES, pts2, 'white', 1, '']]

	#Mark at Center of Face
	dec = 5
	pts = G6.pts_circle 0, 0, dec
	pts2 = [[-dec, 0], [dec, 0], [0, -dec], [0, dec]].collect { |x, y| Geom::Point3d.new(x, y, 0) }
	@hsh_marks[:center_face] = [[GL_POLYGON, pts, 'blue'], [GL_LINE_LOOP, pts, 'white', 1, ''], [GL_LINES, pts2, 'white', 1, '']]

	#Mark at Vertex
	dec = 5
	pts = G6.pts_circle 0, 0, dec
	@hsh_marks[:vertex] = [[GL_POLYGON, pts, 'green'], [GL_LINE_LOOP, pts, 'white', 1, '']]

	#Mark at Guide Point
	dec = 5
	dec1 = dec - 1
	pts = G6.pts_circle 0, 0, dec
	pts2 = [[-dec1, 0], [dec1, 0], [0, -dec1], [0, dec1]].collect { |x, y| Geom::Point3d.new(x, y, 0) }
	@hsh_marks[:cpoint] = [[GL_POLYGON, pts, 'darkgray'], [GL_LINE_LOOP, pts, 'white', 1, ''], [GL_LINES, pts2, 'black', 2, '']]

	#Mark at Text Anchor
	dec = 5
	pts = G6.pts_circle 0, 0, dec
	@hsh_marks[:text_anchor] = [[GL_POLYGON, pts, 'orange'], [GL_LINE_LOOP, pts, 'white', 1, '']]

	#Mark at origin
	dec = 5
	pts = G6.pts_circle 0, 0, dec
	@hsh_marks[:origin] = [[GL_POLYGON, pts, 'yellow'], [GL_LINE_LOOP, pts, 'white', 1, '']]

	#Mark on axes
	dec = 5
	pts = [[-dec, -dec], [dec, -dec], [0, dec]].collect { |x, y| Geom::Point3d.new(x, y, 0) }
	[[:x, 'red'], [:y, 'green'], [:z, 'blue']].each do |symb, color|
		@hsh_marks[symb] = [[GL_POLYGON, pts, color], [GL_LINE_LOOP, pts, 'white', 1, '']]
	end	

	#Mark at Intersection
	dec = 5
	dec1 = dec-1
	pts = [[-dec, -dec], [dec, dec], [-dec, dec], [dec, -dec]].collect { |x, y| Geom::Point3d.new(x, y, 0) }
	pts2 = [[-dec1, -dec1], [dec1, dec1], [-dec1, dec1], [dec1, -dec1]].collect { |x, y| Geom::Point3d.new(x, y, 0) }
	inter_mark = @hsh_marks[:inter] = [[GL_LINES, pts, 'white', 4, ''], [GL_LINES, pts2, 'red', 2, '']]
	[:inter_edge_face, :inter_cline_face, :inter_cline_edge, :inter_cline_cline].each do |symb|
		@hsh_marks[symb] = inter_mark
	end	
	@hsh_marks[:inter_vector_face] = [[GL_LINES, pts, 'white', 4, ''], [GL_LINES, pts2, 'navy', 2, '']]
	@hsh_marks[:inter_plane_edge] = [[GL_LINES, pts, 'white', 4, ''], [GL_LINES, pts2, 'tomato', 2, '']]
	@hsh_marks[:inter_vector_edge] = [[GL_LINES, pts, 'white', 4, ''], [GL_LINES, pts2, 'red', 2, '']]
	@hsh_marks[:inter_vector] = [[GL_LINES, pts, 'white', 4, ''], [GL_LINES, pts2, 'black', 2, '']]
	@hsh_marks[:inter_vector_remote] = @hsh_marks[:inter_plane_remote] = [[GL_LINES, pts, 'black', 4, ''], [GL_LINES, pts2, @color_remote_fr, 2, '']]
	
	#Mark at Center of Arc curve
	dec = 5
	pts = [[-dec, 0], [dec, 0], [0, -dec], [0, dec]].collect { |x, y| Geom::Point3d.new(x, y, 0) }
	@hsh_marks[:center_arc_mark] = [[GL_LINES, pts, 'red', 1, '']]
	pts2 = G6.pts_circle 0, 0, dec, 6
	@hsh_marks[:center_arc] = @hsh_marks[:center_polygon] = [[GL_POLYGON, pts2, 'red'], [GL_LINES, pts, 'white', 1, '']]

	#Mark for twin intersections
	dec = 5
	pts = [[-dec, 0], [dec, 0], [0, -dec], [0, dec]].collect { |x, y| Geom::Point3d.new(x, y, 0) }
	pts2 = G6.pts_circle 0, 0, dec, 6
	@hsh_marks[:center_twin] = [[GL_POLYGON, pts2, @color_twin], [GL_LINES, pts, 'gray', 1, '']]

	#Mark at Center of Component or Group
	dec = 6
	pts1 = G6.pts_circle 0, 0, dec
	pts2 = G6.pts_circle 0, 0, dec-2
	pts3 = G6.pts_circle 0, 0, dec-4
	[[:center_comp_mark_top, ['red', 'yellow', 'blue']], [:center_comp_top, ['blue', 'yellow', 'red']],
	 [:center_comp, ['black', 'white', 'orange']], [:center_comp_mark, ['orange', 'white', 'black']],
	 [:bbox_corner, ['blue', 'white', 'blue']]].each do |a|
		symb, colors = a
		@hsh_marks[symb] = [[GL_POLYGON, pts1, colors[0]], [GL_POLYGON, pts2, colors[1]], [GL_POLYGON, pts3, colors[2]]]
	end	
	@hsh_marks[:center_group] = @hsh_marks[:center_comp]
	@hsh_marks[:center_group_top] = @hsh_marks[:center_comp_top]
end

#MARK: Compute the instructions 2D for drawing at point pt3d
def mark_draw(view, info)
	pt3d, mark, info_comp = info
	return unless pt3d
	
	#Mark for remote
	mark = :remote if mark.to_s =~ /\Aremote/
	
	#Transformation to 2d
	pt2d = G6.check_screen_coords view, pt3d
	return unless pt2d
	t = Geom::Transformation.translation pt2d
	
	#Original instructions
	return unless @hsh_marks.has_key?(mark)
	ls = @hsh_marks[mark]
	
	#Creating the instructions for drawing
	in_comp = (info_comp && info_comp[1])
	instructions = []
	ls.each do |a|
		code, pts, color, width, stipple = a
		color = @color_comp if in_comp && code == GL_POLYGON && mark && !@marks_keep_color.include?(mark)
		instructions.push [code, pts, color, width, stipple]
	end	
	@ogl.process_draw_GL view, t, instructions
end

#MARK: Get the current tooltip based on inference or direction
def get_view_tooltip
	@view_tooltip
end
	
#---------------------------------------------------------
# EXCLUSION: Manage the elements to be excluded
#---------------------------------------------------------

#EXCLUSION: Register the exclusion
def exclusion_register(lst_elts=nil)
	@lst_exclusion = (lst_elts.class == Array) ? lst_elts.clone : nil
	
	@hsh_exclusion_ids = {}
	@hsh_snapshot_ids = {}
	@hsh_compdef = {}
	@at_top_level = false
	return unless @lst_exclusion
	@lst_exclusion = @lst_exclusion.find_all { |e| e.valid? }
	exclusion_map_entityID(@lst_exclusion)	
	
	#Registering ids at top level
	ee = @model.active_entities
	(ee.grep(Sketchup::Face) + ee.grep(Sketchup::Edge) + ee.grep(Sketchup::ConstructionPoint)).each do |e|
		@hsh_snapshot_ids[e.entityID] = e
	end	
end

#EXCLUSION: Find the new created edges
def exclusion_new_edges
	@model.active_entities.grep(Sketchup::Edge).find_all { |e| !@hsh_snapshot_ids[e.entityID] }
end

#EXCLUSION: Go through the selection and map the entity id		
def exclusion_map_entityID(lentities=nil)
	lentities.each do |e|
		next unless e.valid?
		@hsh_exclusion_ids[e.entityID] = e
		if e.instance_of?(Sketchup::ComponentInstance)
			lst = e.definition.entities.to_a
		elsif e.instance_of?(Sketchup::Group)
			lst = e.entities.to_a
		else 		
			(e.edges + e.faces).each { |ee| @hsh_exclusion_ids[ee.entityID] = ee } if e.instance_of?(Sketchup::Vertex)
			e.faces.each { |ee| @hsh_exclusion_ids[ee.entityID] = ee } if e.instance_of?(Sketchup::Edge)
			@at_top_level = true
			next
		end
		lst = lst.grep(Sketchup::ComponentInstance) + lst.grep(Sketchup::Group) + lst.grep(Sketchup::Image)
		exclusion_map_entityID(lst) unless lst.empty?
	end	
end

#EXCLUSION: Exclude when picked element is part of the preselection
def part_of_exclusion?(lentities, parent)
	#Element is part of a grouponent
	return true if parent && @hsh_exclusion_ids[parent.entityID]
	
	#Element is at toplevel model
	lentities.each do |e|
		next unless e
		id = e.entityID
		return true if @hsh_exclusion_ids[id]
		return true if !parent && !@hsh_snapshot_ids[id] && (e.instance_of?(Sketchup::Face) || e.instance_of?(Sketchup::Edge) || e.instance_of?(Sketchup::ConstructionPoint))
		if e.instance_of?(Sketchup::Vertex)
			return true if e.edges.find { |edge| @hsh_exclusion_ids[edge.entityID] }
			return true if e.faces.find { |face| @hsh_exclusion_ids[face.entityID] }
		end
		if e.instance_of?(Sketchup::Face)
			return true if e.vertices.find { |vx| @hsh_exclusion_ids[vx.entityID] }
		end
		if e.instance_of?(Sketchup::Edge)
			return true if [e.start, e.end].find { |vx| @hsh_exclusion_ids[vx.entityID] }
		end
		return true if (e.instance_of?(Sketchup::Edge) || e.instance_of?(Sketchup::Face)) && e.vertices.find { |vx| @lst_exclusion.include?(vx) }
	end
	false
end

#---------------------------------------------------------
# NO_INFERENCE: Manage No Inference mode
#---------------------------------------------------------

def no_inference?
	@no_inference
end
	
def no_inference_toggle
	@no_inference = !@no_inference
	no_inference_after
end

def no_inference_set(flag)
	@no_inference = flag
	no_inference_after
end

def no_inference_after
	remote_reset
	inference_reset
	onMouseMove_zero
end

#NO_INFERENCE: Compute the position with no inference
def no_inference_compute(view, x, y, type)
	ray = view.pickray(x, y)
	face = @ip.face
	if @forced_plane_dir && @pt_origin
		plane = [@pt_origin, @forced_plane_dir]
	elsif face
		tr = @ip.transformation
		plane = [@ip.position, G6.transform_vector(face.normal, tr)]
		pt = Geom.intersect_line_plane ray, plane	
		if pt && [1, 2, 4].include?(face.classify_point(tr.inverse * pt))
			return [pt, :on_face, [face, nil, tr]]
		end	
	elsif @last_plane_found
		plane = @last_plane_found 
	else
		plane = [@ip.position, view.camera.direction]
	end	
	pt = Geom.intersect_line_plane ray, plane
	[pt, type, nil]	
end
		
#---------------------------------------------------------
# DIRECTION: Manage forced_direction
#---------------------------------------------------------

#DIRECTION: Set the imposed direction as a vector
def set_vector_direction(vec, code=nil)
	@forced_vector_code = code if code
	@forced_plane_dir = nil
	@forced_vector_dir = vec
	@direction_tooltip = anchor_tooltip([nil, code])
end

#DIRECTION: Set the imposed direction as a vector
def set_plane_direction(vec, code=nil)
	@forced_plane_code = code
	@forced_vector_dir = nil
	@forced_plane_dir = vec
	@direction_tooltip = anchor_tooltip([nil, code])
end

#DIRECTION: get the planar direction if any
def direction_plane_active
	@implicit_plane_normal
end
	
#DIRECTION: Show / hide the implicit plane	
def show_implicit_plane(val)
	@implicit_show_plane = val
end
	
#DIRECTION: return the tooltip associated to the direction or nil if none
def direction_tooltip
	@direction_tooltip
end

#DIRECTION: Check is a direction is currently imposed
def direction_forced?
	@forced_vector_dir && @forced_vector_code != :protractor
end

def direction_forced_any?
	@forced_vector_dir
end

#DIRECTION: Compute the target when direction is forced
def direction_adjust_target(view, x, y)
	@pt_target, type, info_comp = @target_info
	@info_drive3d = (@target_info) ? @target_info.clone : nil
	return if @pt_origin == @pt_target

	#Vector direction imposed
	if @forced_vector_dir
	
		pvorig = view.screen_coords @pt_origin
		pv0 = view.screen_coords @pt_origin.offset(@forced_vector_dir, 100)
		ptxy = Geom::Point3d.new x, y, 0
		pv1 = ptxy.project_to_line [pvorig, pv0]
		a = Geom.closest_points [@pt_origin, @forced_vector_dir], view.pickray(pv1.x, pv1.y)
		free_target = a[0]
		@info_drive3d = [free_target, :free, nil] unless @info_drive3d
	
		found = false
		#Position locked on a point, edge, or surface in the model
		if @pt_target
			case type
			
			#Force intersection with section plane
			when :on_section_plane
				section, comp, tr = info_comp
				pt, normal = G6.plane_convert_coords(section.get_plane)
				plane = [tr * pt, G6.transform_vector(normal, tr)]
				ptinter = Geom.intersect_line_plane [@pt_origin, @forced_vector_dir], plane
				if ptinter
					@pt_target = ptinter
					type = :inter_vector_section
					@target_info = [@pt_target, type, [section, comp, tr]]
					found = true
				end	

			#Force intersection with face
			when :on_face, :center_face
				face, comp, tr = info_comp
				plane = [tr * face.vertices[0].position, G6.transform_vector(face.normal, tr)]
				ptinter = Geom.intersect_line_plane [@pt_origin, @forced_vector_dir], plane
				if ptinter && [1, 2, 4].include?(face.classify_point(tr.inverse * ptinter))
					@pt_target = ptinter
					type = :inter_vector_face
					@target_info = [@pt_target, type, [face, comp, tr]]
					found = true
				end	
				
			#Force intersection with an edge	
			when :on_edge, :midpoint
				edge, comp, tr = info_comp
				line_edge = [tr * edge.start.position, tr * edge.end.position]
				ptinter = Geom.intersect_line_line [@pt_origin, @forced_vector_dir], line_edge
				if ptinter
					@pt_target = ptinter
					type = :inter_vector_edge
					@target_info = [@pt_target, type, [edge, comp, tr]]
					found = true
				end	
				
			end
			
			#Remote inferencing
			if @remoting_dir
				line = [@remote_origin, @remoting_dir]
				lpt = Geom.closest_points [@pt_origin, @forced_vector_dir], line
				ptinter, = lpt
				if ptinter
					@pt_target = ptinter
					type = :inter_vector_remote
					@target_info = [@pt_target, type, nil]
					found = true
				end	
			end
			
			#All other cases: do a projection
			@pt_target = @pt_target.project_to_line [@pt_origin, @forced_vector_dir] unless found
			
			#Avoiding double inferencing
			if @pt_target == @pt_origin || (found && !G6.points_close_in_pixel?(view, free_target, @pt_target, 15))
				@pt_target = nil
			end	
		end
		
		#Position is free
		unless @pt_target
			type = :free
			@pt_target = free_target
		end	
		
	#Planar direction imposed	
	elsif @forced_plane_dir
		pt_target0 = @pt_target
		
		#A privileged direction exists. Keep it if on plane
		plane = [@pt_origin, @forced_plane_dir]
		direction_privileged_on_target?(view, x, y)
		pt_target, = @target_info
		return if pt_target && pt_target.on_plane?(plane)
		@pt_target = pt_target0
		
		#No target picked
		unless @pt_target && ![:x, :y, :z].include?(type)
			pt = Geom.intersect_line_plane view.pickray(x, y), [@pt_origin, @forced_plane_dir]
			@pt_target = (pt) ? pt : @pt_origin
			@target_info = @pt_target, :free, info_comp		
			@info_drive3d = @target_info
			return
		end
				
		#No good inference for intersection
		unless [:on_edge, :on_cline, :x, :y, :z, :midpoint, :vertex].include?(type) || @remoting_dir
			@pt_target = @pt_target.project_to_plane plane 
			@target_info = @pt_target, type, info_comp		
			return
		end	
		
		#Forcing intersection of plane with Edge (defined by edge, midpoint or vertex
		if type == :on_edge || type == :midpoint || type == :vertex
			if type == :vertex
				vx, comp, tr = info_comp
				ledges = vx.edges
				if ledges.length > 2
					@pt_target = @pt_target.project_to_plane plane
					@target_info = @pt_target, type, info_comp
					return
				end	
				edge = ledges.first
			else
				edge, comp, tr = info_comp
			end	
			new_target, edge1 = direction_intersect_edge_plane(edge, tr, plane)
			if new_target
				@pt_target = new_target
				type = :inter_plane_edge
				@target_info = [@pt_target, type, [edge1, comp, tr]]
				return
			elsif type == :vertex
				@pt_target = @pt_target.project_to_plane plane
				@target_info = @pt_target, type, info_comp
				return
			else
				line = [tr * edge.start.position, tr * edge.end.position]
			end	

		#Forcing intersection of plane with Guide line
		elsif type == :on_cline || type == :end_cline
			cline, comp, tr = info_comp
			line = [tr * cline.position, G6.transform_vector(cline.direction, tr)]

		#Forcing intersection of axes
		elsif type == :x
			line = [ORIGIN, X_AXIS]
		elsif type == :y
			line = [ORIGIN, Y_AXIS]
		elsif type == :z
			line = [ORIGIN, Z_AXIS]
		elsif @remoting_dir && @remote_origin
			line = [@remote_origin, @remoting_dir]
			type = :inter_plane_remote
		end	
		
		ptinter = Geom.intersect_line_plane line, plane
		@pt_target = (ptinter) ? ptinter : @pt_target.project_to_plane(plane)
		
	#No Inference
	elsif @no_inference
		return
		
	#Handling Privileged directions
	elsif direction_privileged_on_target?(view, x, y)
		return
	end

	#Configuring the new target information
	@target_info = @pt_target, type, info_comp
end

#DIRECTION: Privileged direction for target
def direction_privileged_on_target?(view, x, y)
	@pt_target, type, info_comp = @target_info
	pixels = 2 * @precision
	
	#Parameters for the privileged direction
	info_dir = anchor_follow_direction(view, x, y, @pt_origin)
	ptdir, type_dir = info_dir
	
	#Not a privileged direction
	return false unless type_dir && type_dir.to_s =~ /follow|perp/ 
	
	#Exploring the potential intersections...
	
	#...with a face
	if type == :on_face
		line_dir = [@pt_origin, ptdir]
		face, comp, tr = info_comp
		plane = [@pt_target, G6.transform_vector(face.normal, tr)]
		ptinter = Geom.intersect_line_plane line_dir, plane
		if ptinter && [1, 2, 4].include?(face.classify_point(tr.inverse * ptinter)) && 
		   G6.points_close_in_pixel?(view, ptinter, @pt_target, pixels)
			@target_info = [ptinter, type, info_comp]
			@info_drive3d = [ptinter, type_dir]
			return true
		end
		@info_drive3d = @target_info = info_dir
		return true
	end
	
	#...with a face
	if type == :on_section_plane
		line_dir = [@pt_origin, ptdir]
		section, comp, tr = info_comp
		pt, normal = G6.plane_convert_coords(section.get_plane)
		plane = [tr * pt, G6.transform_vector(normal, tr)]
		ptinter = Geom.intersect_line_plane line_dir, plane
		if ptinter
			@target_info = [ptinter, type, info_comp]
			@info_drive3d = [ptinter, type_dir]
			return true
		end
		@info_drive3d = @target_info = info_dir
		return true
	end
	
	#...with an edge	
	new_type = :inter_vector
	if type == :on_edge
		edge, comp, tr = info_comp
		line = [tr * edge.start.position, tr * edge.end.position]
		
	#...with axes	
	elsif type == :x
		line = [ORIGIN, X_AXIS]
	elsif type == :y
		line = [ORIGIN, Y_AXIS]
	elsif type == :z
		line = [ORIGIN, Z_AXIS]
		
	#...with guide line	
	elsif type == :on_cline
		cline, comp, tr = info_comp
		line = [tr * cline.position, G6.transform_vector(cline.direction, tr)]
		
	#...with remote direction if any	
	elsif @remoting_dir && @remote_origin
		line = [@remote_origin, @remoting_dir]
		new_type = :inter_vector_remote
	else
		line = nil
	end
	
	#If there is line compute the intersection
	if line
		line_dir = [@pt_origin, ptdir]
		ptinter = Geom.intersect_line_line line, line_dir
		if ptinter && G6.points_close_in_pixel?(view, ptinter, @pt_target, pixels)
			@target_info = [ptinter, new_type, info_comp]
			@info_drive3d = [ptinter, type_dir]
			return true	
		end
	end
	
	#No line
	if type == :free || !type
		@info_drive3d = @target_info = info_dir
		return true
	end
	false
end

#DIRECTION: Find the intersection of an edge and neighbours with the forced plane
def direction_intersect_edge_plane(edge, tr, plane)
	pt1 = tr * edge.start.position
	pt2 = tr * edge.end.position
	ptinter = Geom.intersect_line_plane [pt1, pt2], plane
	
	#Intersection found
	if ptinter && G6.point_within_segment?(ptinter, pt1, pt2)
		return [ptinter, edge]
	end	
		
	#Getting edge on the right
	edge_right = edge_left = edge
	
	for i in 0..5
		ptinter = nil
		ptinter, edge_right = direction_intersect_edge_plane_next(edge_right, tr, plane, edge_right.end) if edge_right
		return [ptinter, edge_right] if ptinter
		ptinter, edge_left = direction_intersect_edge_plane_next(edge_left, tr, plane, edge_left.start) if edge_left
		return [ptinter, edge_left] if ptinter
	end
	
	#Prolonging the Edge if part of an alone sequence
	nil	
end

#DIRECTION: Find the intersection of the next edge the forced plane
def direction_intersect_edge_plane_next(edge, tr, plane, vx)
	ledges = vx.edges.find_all { |e| e != edge }
	len = ledges.length
	ledges = ledges.find_all { |e| G6.edge_plain?(e) } if len > 1
	return nil if ledges.length != 1
	edge1 = ledges.first

	pt1 = tr * edge1.start.position
	pt2 = tr * edge1.end.position
	ptinter = Geom.intersect_line_plane [pt1, pt2], plane
	
	#Intersection found
	(ptinter && G6.point_within_segment?(ptinter, pt1, pt2)) ? [ptinter, edge1] : [nil, edge1]
end

#DIRECTION: Compute the symbol of direction based on vector and specs
def direction_symbol(vec, code=nil)
	return 'black' unless vec
	unless code
		return @forced_vector_code if @forced_vector_dir
		return @forced_plane_code if @forced_plane_dir
	end
	
	if vec.parallel?(X_AXIS)
		symb = :x
	elsif vec.parallel?(Y_AXIS)
		symb = :y
	elsif vec.parallel?(Z_AXIS)
		symb = :z
	elsif code
		symb = code
	else	
		symb = :default
	end
	symb
end
	
#DIRECTION: Compute privileged direction from a point of reference	
def compute_privileged_directions(reference_info)
	ptref, type_ref, info_comp_ref = reference_info
	elt, comp, tr = info_comp_ref
	
	#Axes
	ls_priv = [[:follow_z, Z_AXIS], [:follow_x, X_AXIS], [:follow_y, Y_AXIS]]
	
	#Axes of components if different from model axis
	@top_axes = false
	if @top_parent.instance_of?(Sketchup::ComponentInstance)
		tr = @top_parent.transformation
		[[:follow_z_comp, Z_AXIS], [:follow_x_comp, X_AXIS], [:follow_y_comp, Y_AXIS]].each do |symb, vec|
			vec_comp = tr * vec
			unless vec_comp.parallel?(vec)
				ls_priv.push [symb, vec_comp]
				@top_axes = true
			end
		end
	end
	
	#Follow a privileged direction
	symb = nil
	case type_ref
	when :on_edge, :midpoint
		edge, comp, tr = info_comp_ref
		vec_edge = (tr * edge.start.position).vector_to(tr * edge.end.position)
		ls_priv.push [:follow_edge, vec_edge]
		if @forced_plane_dir
			vec = vec_edge * @forced_plane_dir
			ls_priv.push [:perp_edge_alone, vec] if vec.valid?			
		elsif edge.faces.length > 0
			edge.faces.each do |face|
				normal = G6.transform_vector(face.normal, tr)
				vec = vec_edge * normal
				ls_priv.push [:perp_edge_face, vec]
			end
		else
			normal = G6.edge_alone_best_plane_normal(edge, ptref, tr) unless normal
			vec = vec_edge * normal
			ls_priv.push [:perp_edge_alone, vec] if vec.valid?
		end
		
	when :on_face, :center_face
		face, comp, tr = info_comp_ref
		ls_priv.push [:follow_normal, G6.transform_vector(face.normal, tr)]
	
	when :on_section_plane
		section, comp, tr = info_comp_ref
		pt, normal = G6.plane_convert_coords(section.get_plane)
		ls_priv.push [:follow_section_plane, G6.transform_vector(normal, tr)]

	when :vertex
		vx, comp, tr = info_comp_ref
		ls_priv.concat direction_privileged_at_vertex(vx, ptref, tr)
		ls_priv.concat direction_privileged_bisectors(vx, tr)
		
	when :on_cline, :end_cline
		cline, comp, tr = info_comp_ref
		ls_priv.push [:follow_cline, G6.transform_vector(cline.direction, tr)]
		
	when :bbox_corner, :center_comp, :center_group, :center_group_top, :center_comp_top
		if @option_inference_comp
			comp, parent, tr = info_comp_ref
			[X_AXIS, Y_AXIS, Z_AXIS].each do |vec|
				ls_priv.push [:follow_bbox, G6.transform_vector(vec, tr)]
			end	
		end	
	end
	
	@origin_privileged_directions = ls_priv
end
	
#DIRECTION: Compute the privileged directions at a vertex	
def direction_privileged_at_vertex(vx, ptref, tr)
	ls_priv = []
	edges = vx.edges
	edges_plain = edges.find_all { |e| G6.edge_plain?(e) }
	edges_plain = edges if edges_plain.empty?
	
	#Following directions
	edges_plain.each do |edge|
		vec_edge = (tr * edge.start.position).vector_to(tr * edge.end.position)
		ls_priv.push [:follow_edge, vec_edge]
	end
	
	#Perpendicular directions
	edges_plain.each do |edge|
		vec_edge = (tr * edge.start.position).vector_to(tr * edge.end.position)
		if edge.faces.length > 0
			edge.faces.each do |face|
				normal = G6.transform_vector face.normal, tr
				ls_priv.push [:perp_edge_face, vec_edge * normal]
			end	
		else
			normal = G6.edge_alone_best_plane_normal(edge, ptref, tr)
			ls_priv.push [:perp_edge_alone, vec_edge * normal] if normal && normal.valid?
		end	
	end	
	ls_priv
end
	
#DIRECTION: Compute the privileged directions at a vertex	
def direction_privileged_bisectors(vx, tr)
	ls_priv = []
	edges = vx.edges
	edges_plain = edges.find_all { |e| G6.edge_plain?(e) }
	edges_plain = edges if edges_plain.empty?
	
	n = edges_plain.length - 2
	for i in 0..n
		for j in i+1..n+1
			edge_i = edges_plain[i]
			edge_j = edges_plain[j]
			vx_i = edge_i.other_vertex vx
			vx_j = edge_j.other_vertex vx
			vec_i = (tr * vx.position).vector_to(tr * vx_i.position).normalize
			vec_j = (tr * vx.position).vector_to(tr * vx_j.position).normalize
			angle = vec_i.angle_between vec_j
			if angle > 20.degrees
				vecsum = vec_i + vec_j
				ls_priv.push [:perp_bisector, vecsum] if vecsum.valid?
			end	
		end
	end	
	
	ls_priv
end
	
#DIRECTION: Compute the code along axes if applicable
def direction_code_from_vector(vec, code)
	if vec.parallel?(X_AXIS)
		code = :follow_x
	elsif vec.parallel?(Y_AXIS)
		code = :follow_y
	elsif vec.parallel?(Z_AXIS)
		code = :follow_z
	end
	code
end

#DIRECTION: Compute a vector direction from picking elements in the model	
def direction_vector_from_model(info)
	pt, symb, info_elt = info
	elt, comp, tr = info_elt
	vec = nil
	
	case symb
	when :on_edge, :midpoint
		vec = (tr * elt.start.position).vector_to(tr * elt.end.position)
		code = :follow_edge
	when :on_cline, :end_cline
		vec = G6.transform_vector(elt.direction, tr)
		code = :follow_cline
	when :on_face, :center_face
		vec = G6.transform_vector(elt.normal, tr)
		code = :follow_normal	
	when :on_section_plane
		pt, normal = G6.plane_convert_coords elt.get_plane
		vec = G6.transform_vector(normal, tr)
		code = :follow_section_plane	
	end
	
	return nil unless vec && vec.valid?
	code = direction_code_from_vector(vec, code)
	(vec) ? [vec, code] : nil
end

#DIRECTION: Compute a planar direction from picking elements in the model	
def direction_plane_from_model(info, use_best_normal=true)
	#Use bissector
	if option_implicit_plane_check(:bissector)
		return direction_plane_from_model_bisec(info)
	end

	#Details on picked point
	pt, symb, info_elt, best_normal = info
	elt, comp, tr = info_elt
	vec = nil
	
	#Option is best face
	if false && best_normal && (use_best_normal || option_implicit_plane_check(:best_normal))
		vec = best_normal
		code = :follow_normal
		code = direction_code_from_vector(vec, code)
		return [vec, code]
	end
	
	#Normal cases
	case symb
	when :on_edge, :midpoint
		vec = (tr * elt.start.position).vector_to(tr * elt.end.position)
		vec = vec.reverse if elt.start.edges.length == 1
		code = :follow_edge
	when :on_cline, :end_cline
		vec = G6.transform_vector(elt.direction, tr)
		code = :follow_cline
	when :on_face, :center_face
		vec = G6.transform_vector(elt.normal, tr)
		code = :follow_normal		
	when :on_section_plane
		pt, normal = G6.plane_convert_coords elt.get_plane
		vec = G6.transform_vector(normal, tr)
		code = :follow_section_plane	
	when :center_arc, :center_polygon
		edge = elt
		vec1 = pt.vector_to(tr * edge.start.position)
		vec2 = pt.vector_to(tr * edge.end.position)
		vec = vec1 * vec2
		code = :follow_normal		
	when :vertex
		vx = elt
		edges = vx.edges
		edges_plain = edges.find_all { |e| G6.edge_plain?(e) }
		if vx.faces.length > 0
			lvec = vx.faces.collect { |f| G6.transform_vector(f.normal, tr) }
			vec = G6.vector_nice_average lvec
			code = :follow_normal			
		elsif edges.length == 1
			edge = edges[0]
			code = :follow_edge
			vec = (tr * edge.start.position).vector_to(tr * edge.end.position)
			vec = vec.reverse if edge.start.edges.length == 1
		elsif edges_plain.length == 1	
			edge = edges_plain[0]
			code = :follow_edge
			vec = (tr * edge.start.position).vector_to(tr * edge.end.position)
			vec = vec.reverse if edge.start.edges.length == 1
		elsif edges_plain.length == 2
			edge1, edge2 = edges_plain
			vec1 = (tr * edge1.start.position).vector_to(tr * edge1.end.position)
			vec2 = (tr * edge2.start.position).vector_to(tr * edge2.end.position)
			if edge1.faces.length > 0 && edge2.faces.length > 0
				vec = vec1 * vec2
				code = :plane_of_edges
			else
				vec = G6.vector_straight_average [vec1.normalize, vec2.normalize]
				code = :plane_of_edges
			end
		elsif best_normal
			vec = best_normal
			code = :follow_normal				
		end
		
	end
	
	return nil unless vec && vec.valid?
	
	code = direction_code_from_vector(vec, code)
	(vec) ? [vec, code] : nil
end

#DIRECTION: Compute a planar direction from picking elements in the model, with plane based on bissector	
def direction_plane_from_model_bisec(info)
	pt, symb, info_elt = info
	elt, comp, tr = info_elt
	return nil unless elt && elt.valid?
	vec = nil
	
	case symb
	when :on_edge, :midpoint
		edge = elt
		lvec = edge.faces.collect { |f| G6.transform_vector(f.normal.normalize, tr) }
		if lvec.empty?
			vec = (tr * elt.start.position).vector_to(tr * elt.end.position)
			code = :follow_edge
		else	
			vec = G6.vector_nice_average lvec
			code = :follow_normal
		end	
	when :on_cline, :end_cline
		vec = G6.transform_vector(elt.direction, tr)
		code = :follow_cline
	when :on_face, :center_face
		vec = G6.transform_vector(elt.normal, tr)
		code = :follow_normal	
	when :on_section_plane	
		pt, normal = G6.plane_convert_coords(elt.get_plane)
		vec = G6.transform_vector(normal, tr)
		code = :follow_section_plane
	when :vertex
		vx = elt
		edges = vx.edges
		edges_plain = edges.find_all { |e| G6.edge_plain?(e) }
		if vx.faces.length > 0
			lvec = vx.faces.collect { |f| G6.transform_vector(f.normal, tr) }
			vec = G6.vector_nice_average lvec
			code = :follow_normal			
		elsif edges.length == 1
			edge = edges[0]
			code = :follow_edge
			vec = (tr * edge.start.position).vector_to(tr * edge.end.position)
		elsif edges_plain.length == 1	
			edge = edges_plain[0]
			code = :follow_edge
			vec = (tr * edge.start.position).vector_to(tr * edge.end.position)
		elsif edges_plain.length == 2
			edge1, edge2 = edges_plain
			vec1 = (tr * edge1.start.position).vector_to(tr * edge1.end.position)
			vec2 = (tr * edge2.start.position).vector_to(tr * edge2.end.position)
			vec = vec1 * vec2
			code = :plane_of_edges			
		end
		
	end
	
	return nil unless vec && vec.valid?
	
	code = direction_code_from_vector(vec, code)
	(vec) ? [vec, code] : nil
end

#---------------------------------------------------------
# PROTRACTOR: Manage the protractor
#---------------------------------------------------------

#PROTRACTOR: Check if the protractor is enabled
def protractor_enabled?
	@protractor_enabled
end

#PROTRACTOR: Check if the protractor is displayed
def protractor_shown?
	@protractor_enabled && @protractor_showable
end

#PROTRACTOR: Enable the protractor	
def protractor_enable(flag)
	@protractor_enabled = flag
	protractor_manage_after
end

#PROTRACTOR: Toggle the protractor
def protractor_toggle
	@protractor_enabled = !@protractor_enabled
	protractor_manage_after
end

#PROTRACTOR: Set the plane
def protractor_set_plane(normal)
	@protractor_normal = normal
	protractor_find_plane_direction
	@view.invalidate
end

#PROTRACTOR: Set the direction vector
def protractor_set_direction(vector)
	@protractor_direction = vector
	protractor_find_plane_direction
	@view.invalidate
end

#PROTRACTOR: common tasks after changing parameters
def protractor_manage_after
	if @protractor_enabled 
		#Saving the previous direction information if any
		if @forced_plane_dir
			@protractor_previous_dir_info = nil
			return
		end	
		@protractor_previous_dir_info = [@option_vector]
		@protractor_normal = @protractor_vector = nil
		protractor_find_plane_direction
	else	
		if @protractor_previous_dir_info != nil
			option_vector, = @protractor_previous_dir_info
			if option_vector
				option_vector_set_dir :no
			else	
				option_planar_set_dir :no
			end	
		end
		@view.invalidate
		return
	end	
	
	#Computing the parameters of the protractor
	@view.invalidate
end

#PROTRACTOR: compute the VCB display for angle
def protractor_angle_text(sign=false)
	return nil unless protractor_shown?
	return nil unless @pt_origin && @pt_target && @pt_origin != @pt_target
	vec = @pt_origin.vector_to @pt_target
	angle = @protractor_vector.angle_between vec
	angle = -angle if (@protractor_vector * vec) % @protractor_normal < 0
	sign = (sign) ? "°" : 'd'
	sprintf("%0.1f", angle.radians) + sign
end

#PROTRACTOR: Store the normal and vector from the Input Point
def protractor_compute_from_ip(view, x, y)
	normal = vector = edges = nil
	face, edge, vx, tr = [@ip.face, @ip.edge, @ip.vertex, @ip.transformation]
	edges = vx.edges if vx
	ptxy = Geom::Point3d.new x, y, 0
	
	#Computing the normal
	if face
		normal = G6.transform_vector face.normal, tr
		pt2d = @view.screen_coords @ip.position
		ray = @view.pickray pt2d.x, pt2d.y
		normal = normal.reverse if (normal % ray[1] > 0)
	elsif vx
		e1, e2 = edges
		if edges.length == 2
			vec1 = (tr * e1.start.position).vector_to(tr * e1.end.position)
			vec2 = (tr * e2.start.position).vector_to(tr * e2.end.position)
			normal = vec1 * vec2
		end
	end
	
	#Finding the closest edge
	if edges && !vector
		dmin = 200
		edge_min = nil
		edges.each do |edge|
			pt1_2d = view.screen_coords tr * edge.start.position
			pt2_2d = view.screen_coords tr * edge.end.position
			pt1_2d.z = pt2_2d.z = 0
			d = ptxy.distance_to_line [pt1_2d, pt2_2d]
			if d < dmin
				dmin = d
				edge_min = edge
			end	
		end
		vector = (tr * edge_min.start.position).vector_to(tr * edge_min.end.position) if edge_min
	end
	
	@protractor_from_ip = [normal, vector]
end

#PROTRACTOR: try to find a plane
def protractor_find_plane_direction	
	@protractor_showable = false
	return unless @pt_origin
	
	#Information on origin
	origin, type, info_comp = @origin_info
	elt, comp, tr = info_comp
	normal_ip, vector_ip = @protractor_from_ip
	
	#Computing the Plane normal
	normal = nil
	code = :protractor
	if @forced_plane_dir
		normal = @forced_plane_dir
		code = @forced_plane_code
	elsif normal_ip
		normal = normal_ip
	elsif elt.instance_of?(Sketchup::Face)
		normal = G6.transform_vector elt.normal, tr
		pt2d = @view.screen_coords(tr * elt.vertices[0].position)
		ray = @view.pickray pt2d.x, pt2d.y
		normal = normal.reverse if (normal % ray[1] > 0)	
	else
		normal = @protractor_normal
	end
	
	#Computing the direction
	vector = @protractor_direction
	unless vector
		if vector_ip
			vector = vector_ip
		elsif elt.instance_of?(Sketchup::Edge)
			vector = (tr * elt.start.position).vector_to(tr * elt.end.position)
		elsif elt.instance_of?(Sketchup::ConstructionLine)
			vector = G6.transform_vector(elt.direction, tr)
		end
	end	
	if vector && vector.valid?
		normal = vector.axes[0] unless normal && normal.valid?
	else
		normal = @protractor_normal unless normal && normal.valid?
		normal = Z_AXIS unless normal
		vector = normal.axes[0]
	end	
	return unless normal && normal.valid?
	vector = G6.project_vector_to_plane(vector, [ORIGIN, normal])
	return unless vector.valid?
	
	#Storing the information and configuring the protractor
	@protractor_normal = normal
	@protractor_code = code
	@protractor_vector = vector
	
	@protractor.set_placement @pt_origin, normal, vector	
	@protractor_showable = true
end

#PROTRACTOR: Compute the best plane and direction
def protractor_compute	
	@protractor_showable = false
	return unless @pt_origin && @protractor_enabled
	return unless @protractor_normal && @protractor_vector
	
	@forced_model_vector, @forced_model_code = [@protractor_normal, @protractor_code]
	@option_planar = true
	@option_vector = false
	direction_changed
	@protractor_showable = true
end

#ROTATION: Executing the rotation
def protractor_target(view, x, y)
	return nil unless protractor_shown?
	return nil unless @pt_origin
	return nil if @forced_vector_dir
	
	#Normal target is on the plane of protractor
	pt, type = @target_info
	if pt && pt.on_plane?([@pt_origin, @protractor_normal]) && 
		case type
		when :vertex, :cpoint, :midpoint, /center/, /inter/, /bbox/
			return @target_info
		end	
	end	
	
	#Calculating the angle
	normal = @protractor_normal
	center = @pt_origin
	vecdir = @protractor_vector
	
	target = Geom.intersect_line_plane view.pickray(x, y), [center, normal]
	vec = center.vector_to target
	angle = vecdir.angle_between vec
	angle = -angle if angle != 0 && (vecdir * vec) % normal < 0
	
	#Rounding up the angle
	center2d = view.screen_coords center
	center2d.z = 0
	ptxy = Geom::Point3d.new x, y, 0
	d = center2d.distance ptxy
	radius2d = 1.2 * @protractor.get_radius2d
	factor = d / radius2d
	adjusted_angle = @protractor.adjust_angle(angle.abs, factor)
	adjusted_angle = -adjusted_angle if angle < 0
	
	#Performing the transformation
	t = Geom::Transformation.rotation center, normal, adjusted_angle
	target = t * center.offset(vecdir, center.distance(target))
	
	[target, :free, nil]
end

#PROTRACTOR: Force the angle from VCB input (angle in radians)
def protractor_force_angle(angle)
	return unless protractor_shown? && @pt_origin
	t = Geom::Transformation.rotation @pt_origin, @protractor_normal, angle
	new_vec = t * @protractor_vector
	set_plane_direction nil
	set_vector_direction new_vec, :protractor
	@direction_lock = true
	onMouseMove_zero
end

#---------------------------------------------------------
# VCB: Helpers for VCB management
#---------------------------------------------------------

#VCB: Store the VCB Information
def vcb_freeze
	return if @vcb_info
	if @remoting_dir
		@vcb_info = [:remote, @pt_origin, @pt_target, @remote_origin, @remoting_dir]
	else
		@vcb_info = [:offset, @pt_origin, @pt_target]
	end	
	@vcb_info
end

#VCB: Reset the VC Information (and save previous)
def vcb_unfreeze
	@vcb_info_prev = @vcb_info if @vcb_info
	@vcb_info = nil
end

#VCB: Check if a Redo is possible
def vcb_redo_possible?
	@vcb_info != nil
end
	
#VCB: Restore the previous VCB information just after a cancel	
def vcb_restore_after_cancel
	@vcb_info = @vcb_info_prev
end

#VCB: Return the effective Label and Length to be displayed in the VCB	
def vcb_label_length
	#Getting the information
	type, ori, targ, rori, rvec = vcb_dynamic_info	
	
	#Computing the label and distance
	len = nil
	case type
	when :remote
		label = @txt_distance
		if !ori || !rori
			len = nil
		elsif targ
			len = targ.distance(rori)
		else
			len = ori.distance(rori)
		end	
	when :offset
		label = @txt_offset
		len = (targ) ? ori.distance(targ) : nil
	end
	
	tx_len = (len) ? Sketchup.format_length(len) : ''
	tx_ang = protractor_angle_text
	tx_len += '   ' + tx_ang if tx_ang
	[label, tx_len]
end

#VCB: Check if origin is in Remoting mode
def vcb_origin_remote?
	@mode == :origin && !@pt_target && @remoting_dir
end

#VCB: Return the information about the offset or distance mode
def vcb_dynamic_info
	ori = @pt_origin
	targ = @pt_target
	rori = rvec = nil
	if @vcb_info
		type, ori, targ, rori, rvec = @vcb_info	
	elsif @remoting_dir
		type = :remote
		rori = @remote_origin
		rvec = @remoting_dir
	else
		type = :offset
	end
	[type, ori, targ, rori, rvec]
end

#VCB: Specify an offset or a distance for the origin or target
def vcb_change_specifications(len, new_origin=nil, new_vec=nil)
	#Getting the information
	type, ori, targ, rori, rvec = vcb_dynamic_info	
	new_ori = ori
	new_targ = targ
	
	#New origin
	if new_origin
		return nil unless ori && targ
		vec = ori.vector_to targ
		return nil unless vec.valid?
		new_targ = new_origin.offset vec, vec.length
		@vcb_info = [:offset, new_origin, new_targ]
		return [new_origin, new_targ]
	end

	#New vector
	if new_vec
		return nil unless ori && new_vec.valid?
		new_targ = ori.offset new_vec, new_vec.length
		@vcb_info = [:offset, ori, new_targ]
		return [ori, new_targ]
	end
	
	#Remote inference
	if type == :remote
		len = rvec.length unless len
		if !targ && new_ori != rori
			rvec = rori.vector_to ori
			new_ori = rori.offset rvec, len
		else
			rvec = rori.vector_to targ
			new_targ = rori.offset rvec, len
		end	
		rvec.length = len
	
	#Regular offset
	elsif ori && targ
		vec = ori.vector_to targ
		len = vec.length unless len
		new_targ = ori.offset vec, len	
	end
	
	#Updating the VCB Info
	@vcb_info = [type, new_ori, new_targ, rori, rvec]
	
	#Returning the effective origin and target
	[new_ori, new_targ]	
end

#---------------------------------------------------------
# REMOTE: Manage remote inferencing
#---------------------------------------------------------

#REMOTE: Disable or Enable the Remote inference
def remote_enable(flag)
	@option_inference_remote = flag
	remote_reset unless flag
end

#REMOTE: Toggle the remote flag
def remote_toggle
	flag = !@option_inference_remote
	remote_enable flag
end
	
#REMOTE: Check if remote is applicable	
def remote_possible?
	return true if @mode == :target
	(@ignore_selection || @preselection || @cumulative_selection)
end

#REMOTE: Check if Remote inference is currently enabled
def remote_enabled? 
	(!@no_inference && @option_inference_remote && remote_possible?)
end
	
#REMOTE: Check if Remote inference is currently enabled
def remote_active_dir
	@remoting_dir
end
	
#REMOTE: Return direction of alignment for origin	
def remote_parallel_vector
	if @parallel_vector
		@parallel_vector
	elsif @forced_vector_dir
		@forced_vector_dir
	elsif @forced_plane_dir
		@forced_plane_dir
	else
		nil
	end	
end
	
#REMOTE: Reset the remote inferences	
def remote_reset
	@view.invalidate if @remote_origin
	@remote_origin = @remote_origin_prev = @remote_vectors = nil
	@remoting_dir = nil
	@parallel_info = @parallel_info_prev = nil
	@parallel_origin = nil
	@parallel_dir = nil
	@remote_info_prev = nil
	@twin_info = nil
	remote_timing_stop
	@remote_last_time = nil
end

#REMOTE: Store and compute the remote inference direction if applicable
def remote_register
	return unless remote_enabled?
	remote_timing_stop
	info = (@mode == :origin) ? @origin_info : @info_drive3d
	return remote_reset unless info
	
	#Not candidate for remote inference or not wait long enough
	return unless @remote_last_time && (Time.now - @remote_last_time) > remote_delay
	target, type, info_comp = info	
	return if target == @pt_origin && @mode == :target

	#Registering the origin and potentials directions for remote reference
	elt, comp, tr = info_comp
	return remote_reset unless !elt || elt.valid?
	
	@parallel_info = nil
	ori = nil
	lvec = []
	case type
	when :vertex
		lvec += direction_privileged_at_vertex(elt, target, tr)
	when :midpoint
		vec_edge = (tr * elt.start.position).vector_to(tr * elt.end.position)
		lvec.push [:follow_edge, vec_edge]
		if elt.faces.length > 0
			elt.faces.each do |face|
				lvec.push [:perp_edge_face, vec_edge * G6.transform_vector(face.normal, tr)]
			end
		else
			normal = G6.edge_alone_best_plane_normal(elt, target, tr)
			lvec.push [:perp_edge_alone, vec_edge * normal] if normal && normal.valid?
		end
	when :end_cline
		lvec.push [:follow_cline, G6.transform_vector(elt.direction, tr)]
		
	when :cpoint, :center_arc, :center_twin
		
	when :bbox_corner, :center_comp, :center_comp_top, :center_group, :center_group_top
		t = comp.transformation * tr
		lvec += [[:follow_x_comp, X_AXIS], [:follow_y_comp, Y_AXIS], [:follow_z_comp, Z_AXIS]].collect do |symb, vec| 
			[symb, G6.transform_vector(vec, t)] 
		end	
		
	when :on_edge
		@parallel_info = info_comp
		@parallel_vector = (tr * elt.start.position).vector_to(tr * elt.end.position)
		@parallel_dir_symb = :to_edge
		
	when :on_face, :center_face
		lvec.push [:follow_normal, G6.transform_vector(elt.normal, tr)]

	when :on_cline
		@parallel_info = info_comp
		@parallel_vector = G6.transform_vector(elt.direction, tr)
		@parallel_dir_symb = :to_cline
		
	else
		if !@remoting_dir && (Time.now - @remote_last_time) > 2 * remote_delay
			remote_reset
		end	
		return	
	end	
	
	#Registering the directions
	if @parallel_info
		return if @parallel_info_prev && @parallel_info_prev[0] == @parallel_info[0]
		@parallel_origin = target
		@parallel_info_prev = @parallel_info
		@remoting_dir = @remote_vectors = nil
		@remote_origin = @remote_origin_prev = nil
	else
		return if @remote_origin_prev == target
		@remote_origin = @remote_origin_prev = target
		lvec.push [:follow_x, X_AXIS], [:follow_y, Y_AXIS], [:follow_z, Z_AXIS]
		@remote_vectors = lvec
		@parallel_info = @parallel_info_prev = @parallel_vector = nil
	end
	
	#Check if there are twin inferences
	twin_compute
	
	#Storing the previous remote inference if any
	@remote_info_prev = twin_store_previous
	
	@view.invalidate
end

#REMOTE: Start the timer
def remote_timing_start
	return unless remote_enabled?
	@remote_last_time = Time.now
	return if @remote_timer
	@remote_timer = UI.start_timer(remote_delay + 0.1) { @remote_timer = nil ; remote_register }
end

#REMOTE: Stop the timer
def remote_timing_stop
	return unless @remote_timer
	UI.stop_timer @remote_timer
	@remote_timer = nil
end
	
#REMOTE: Check alignment with remote origin	
def remote_alignment_check(view, x, y, info)
	return nil unless remote_enabled?
	return nil unless @remote_origin && (!info || @remote_origin != info[0])

	#Checking if remoting can be applied
	@remoting_dir = nil
	type = (info) ? info[1] : :void
	
	#Other cases that can be skipped
	return nil if [:vertex, :midpoint, :bbox_corner, :origin, :cpoint, :text_anchor].include?(type)
	return nil if type.to_s =~ /center|inter/i
	
	#Calculating the reference direction
	size = view.pixels_to_model 100, @remote_origin
	origin2d = view.screen_coords @remote_origin
	ptxy = Geom::Point3d.new x, y, 0
	origin2d.z = 0
	vecref2d = origin2d.vector_to ptxy
	
	#Checking alignment
	ecart_min = @inference_distance
	@remote_vectors.each do |symb, vec3d|
		pt3d = @remote_origin.offset vec3d, size
		pt2d = view.screen_coords pt3d
		pt2d.z = 0
		vec2d = origin2d.vector_to pt2d
		vec2d = vec2d.reverse if vecref2d % vec2d < 0
		next if vec2d.angle_between(vecref2d) > @inference_angle
		d = origin2d.distance ptxy
		pt2d = origin2d.offset vecref2d, d
		ecart = ptxy.distance(pt2d)
		if ecart < ecart_min
			ecart_min = ecart
			@remoting_dir = vec3d
			@remoting_dir_symb = symb
		end		
	end
	
	#No remoting inference
	return nil unless @remoting_dir
	
	#Adjusting the origin
	if @mode == :origin
		return remote_normal_intersections(info, view, x, y)
	end
	
	#Adjusting the target
	remote_normal_intersections(info, view, x, y)
end

#REMOTE: Check intersections with forced direction or privileged directions
def remote_normal_intersections(info, view, x, y)
	lpt = Geom.closest_points [@remote_origin, @remoting_dir], view.pickray(x, y)
	ptinter = lpt.first
	unless ptinter
		@remoting_dir = nil
		return nil
	end	
	type = "remote_#{@remoting_dir_symb}".intern
	if info
		info[0] = ptinter
		info[1] = type
	else
		info = [ptinter, type, nil]
	end	
	info
end

#---------------------------------------------------------
# PARALLEL: Manage the origin
#---------------------------------------------------------

#PARALLEL: Check alignment with remote origin	
def parallel_alignment_check(view, x, y, info)
	return nil unless @mode == :target && remote_enabled?
	return nil unless @parallel_info
	@parallel_dir = nil
	
	#Other cases that can be skipped
	type = (info) ? info[1] : :void
	return nil if [:vertex, :midpoint, :bbox_corner, :origin, :cpoint, :text_anchor].include?(type)
	return nil if type.to_s =~ /center|inter/i
	
	#Calculating the reference direction
	size = view.pixels_to_model 100, @pt_origin
	origin2d = view.screen_coords @pt_origin
	origin2d.z = 0
	ptxy = Geom::Point3d.new x, y, 0
	vecref2d = origin2d.vector_to ptxy
	
	#Checking alignment
	ecart_min = @inference_distance
	vec3d = @parallel_vector
	pt3d = @pt_origin.offset vec3d, size
	pt2d = view.screen_coords pt3d
	pt2d.z = 0
	vec2d = origin2d.vector_to pt2d
	vec2d = vec2d.reverse if vecref2d % vec2d < 0
	return nil if vec2d.angle_between(vecref2d) > @inference_angle
	d = origin2d.distance ptxy
	pt2d = origin2d.offset vecref2d, d
	ecart = ptxy.distance(pt2d)
	if ecart < ecart_min
		ecart_min = ecart
		@parallel_dir = vec3d
	end		
	
	#No remoting inference
	return nil unless @parallel_dir
	
	#Adjusting the target
	parallel_normal_intersections(info, view, x, y)
end

#PARALLEL: Check intersections with forced direction or privileged directions
def parallel_normal_intersections(info, view, x, y)
	lpt = Geom.closest_points [@pt_origin, @parallel_dir], view.pickray(x, y)
	ptinter = lpt.first
	unless ptinter
		@parallel_dir = nil
		return nil
	end	
	type = "parallel_#{@parallel_dir_symb}".intern
	if info
		info[0] = ptinter
		info[1] = type
	else
		info = [ptinter, type, nil]
	end	
	info
end

#---------------------------------------------------------
# TWIN: Manage the double remote inferences
#---------------------------------------------------------
	
#TWIN: Store the current remote inference	
def twin_store_previous
	if @parallel_info
		info = [@parallel_origin, [@parallel_vector]]
	elsif @remote_origin	
		vectors = @remote_vectors.find_all { |a| a[0].to_s !~ /x|y|z/ }
		info = (vectors.empty?) ? nil : [@remote_origin, vectors.collect { |a| a[1] }]
	else
		info = nil
	end	
	info
end

#TWIN: Compute the points where double inferences are concerned
def twin_compute
	return unless @remote_info_prev
	ori, vectors = @remote_info_prev
	
	new_info = twin_store_previous
	return unless new_info
	new_ori, new_vectors = new_info
	
	@twin_info = []
	vectors.each do |vec|
		line = [ori, vec]
		new_vectors.each do |new_vec|
			new_line = [new_ori, new_vec]
			ptinter = Geom.intersect_line_line line, new_line
			@twin_info.push [ptinter, ori, new_ori] if ptinter
		end
	end
	@twin_info = nil if @twin_info.empty?
end
		
#---------------------------------------------------------
# ORIGIN: Manage the origin
#---------------------------------------------------------

#ORIGIN: Pick the target (complex mode with exclusion)
def origin_pick(view, x, y, extended=false, multi_selection=false, no_vertex=false)
	remote_timing_stop
	@origin_privileged_directions = nil
	@ip.pick view, x, y
	protractor_compute_from_ip(view, x, y)
	
	#Storing the remote inference
	remote_register
		
	@origin_info = anchor_pick(view, x, y)	
	@origin_face_info = anchor_picked_face_info(@origin_info)
	
	#Origin with no inference
	unless @origin_info && !@no_inference
		pt = Geom.intersect_line_plane view.pickray(x, y), [@ip.position, view.camera.direction]
		@origin_info = no_inference_compute(view, x, y, :void_origin)
	end
		
	#Adjusting for remote_inference
	unless @no_inference
		info = remote_alignment_check(view, x, y, @origin_info)
		@origin_info = info if info
	end
	
	#Storing the normal for further pivoting
	@pivot_vector_origin = direction_pivot_compute(@origin_info)
	
	#Auto selection
	@active_selection = auto_selection(view, x, y, extended, multi_selection, no_vertex)
	
	#Computing the Protractor parameters
	protractor_find_plane_direction

	#Computing the implicit plane if applicable
	direction_implicit_plane_compute
	
	#Unfreeze the VCB information
	vcb_unfreeze if @remoting_dir
	
	#Common tasks
	origin_after_set	
end

#ORIGIN: Set the origin information
def origin_set_info(origin_info)
	remote_reset
	@pt_origin_prev = @pt_origin
	@origin_privileged_directions = nil
	@origin_info = origin_info
	@pt_origin, = @origin_info
	
	#Common tasks
	origin_after_set	
end
	
#ORIGIN: Compute the origin environment based on origin information	
def origin_after_set
	@forced_vector_dir = @forced_vector_code = nil if @forced_vector_code == :protractor
	@origin_privileged_directions = nil
	compute_privileged_directions(@origin_info) unless @no_inference
	protractor_find_plane_direction

	#Setting the tooltip for the view
	@view_tooltip = anchor_tooltip(@origin_info)
	@view.tooltip = (@view_tooltip) ? '   ' + @view_tooltip : nil

	@pt_origin = @origin_info.first
	
	#Starting timer for remoting
	remote_timing_start
		
	@origin_info
end
	
#ORIGIN: Active selection when picking origin	
def origin_active_selection
	@active_selection
end
	
#ORIGIN: get information on origin	
def origin_info
	@origin_info
end

#TARGET: return the target face info if any
def origin_face_info
	@origin_face_info
end
		
#ORIGIN: get information on the top parent picked in Origin mode	
def origin_info_top_parent
	@info_top_parent
end
		
#---------------------------------------------------------
# TARGET: Manage the target
#---------------------------------------------------------
	
#TARGET: Pick the target (complex mode with exclusion)
def target_pick(view, x, y, with_exclusion=false)	
	remote_timing_stop
	
	#Storing the remote inference
	remote_register
	
	#Picking the target
	@pt_drive2d = Geom::Point3d.new x, y, 0	
	@ip.pick view, x, y
	@target_info = anchor_pick(view, x, y, with_exclusion)
	@pt_target, type, info_comp = @target_info
	
	@pivot_vector_target = direction_pivot_compute(@target_info)

	#No Inference
	if @no_inference
		#do nothing
	
	#Adjusting for parallel and remote inference
	else
		info = parallel_alignment_check(view, x, y, @target_info)
		if info
			@target_info = info
		else	
			info = remote_alignment_check(view, x, y, @target_info)
			@target_info = info if info
		end	
	end	
	
	#Protractor mode
	info = protractor_target(view, x, y)
	if info
		@target_info = @info_drive3d = info
		return target_after_set(view, x, y)
	end
	
	#Adjusting the target based on a possible forced direction or plane
	direction_adjust_target view, x, y
			
	#No target found yet - Intersection with current plane	
	@pt_target, = @target_info
	@pt_drive3d, = @info_drive3d
	if !@target_info || (@no_inference && @forced_plane_dir && !@pt_drive3d.on_plane?([@pt_origin, @forced_plane_dir]) ||
	   (@no_inference && !@forced_plane_dir && ![:on_face].include?(type)))
		@info_drive3d = @target_info = no_inference_compute(view, x, y, :free)
	end	
	@target_face_info = anchor_picked_face_info(@target_info)

	#Common tasks
	target_after_set view, x, y
end

#TARGET: Common operations after picking or setting the target
def target_after_set(view, x, y)
	#Setting the tooltip for the view
	@view_tooltip = anchor_tooltip(@target_info)
	view.tooltip = (@view_tooltip) ? '   ' + @view_tooltip : nil

	@pt_target, type, info_comp = @target_info
	@pt_drive3d, = @info_drive3d

	#Starting timer for remoting
	remote_timing_start
	
	vcb_unfreeze if @pt_target != @pt_origin
	
	#adjusting the track view if applicable
	#####view_follow_target(view, x, y)
	@pt_target_prev = @pt_target
	
	@target_info
end

def view_follow_target(view, x, y)
	return unless @pt_target_prev
	vec = @pt_target_prev.vector_to @pt_target
	return unless vec.valid?
	
	w = view.vpwidth
	h = view.vpheight
	fac = 0.1 ; xmin_on = w * fac ; xmax_on = w * (1 - fac) ; ymin_on = h * fac ; ymax_on = h * (1 - fac)
	fac = 0.3 ; xmin_off = w * fac ; xmax_off = w * (1 - fac) ; ymin_off = h * fac ; ymax_off = h * (1 - fac)
	if x < xmin_on || x > xmax_on || y < xmin_on || y > ymax_on
		@enter_unreal = true
	elsif x > xmin_off && x < xmax_off && y > xmin_off && y < ymax_off
		@enter_unreal = false
	end
	
	if true || @enter_unreal
		camera = @view.camera
		new_eye = camera.eye.offset vec
		new_target = camera.target.offset vec
		camera.set new_eye, new_target, camera.up
	end	
end

#TARGET: get information on target	
def target_info
	@target_info
end

#TARGET: return the target face info if any
def target_face_info
	@target_face_info
end
	
#TARGET: Set the origin information
def target_set_info(target_info)
	remote_reset
	@pt_origin_prev = @pt_origin
	@origin_privileged_directions = nil
	@target_info = target_info
	@pt_target, = @target_info
	@info_drive3d = @target_info
	@pt_drive3d, = @info_drive3d

	#Setting the tooltip for the view
	@view_tooltip = anchor_tooltip(@target_info)
	@view.tooltip = (@view_tooltip) ? '   ' + @view_tooltip : nil
	
	#Common tasks
	vcb_freeze if @pt_target != @pt_origin
	
	@target_info
end
	
#---------------------------------------------------------
# REPERE: Manage the anchor point detection
#---------------------------------------------------------
	
#REPERE: Compute the axes of the plane
def repere_compute_plane
	@repere_axes = nil
	normal = direction_plane_active
	return unless normal && @pt_origin
	
	#Reversing the normal to make it emerging from face
	pt2d = @view.screen_coords @pt_origin
	ray = @view.pickray pt2d.x, pt2d.y
	if (normal % ray[1] > 0)
		normal = normal.reverse
	end	
	
	#Computing the axes
	vecx = (@pt_target && @pt_target != @pt_origin) ? @pt_origin.vector_to(@pt_target) : normal.axes[0]
	vecy = normal * vecx
	vecx, vecy = normal.axes unless vecy.valid?
	@repere_axes = [vecx, vecy, normal]
	@repere_tr = Geom::Transformation.axes @pt_origin, *@repere_axes
end
	
#---------------------------------------------------------
# ANCHOR: Manage the anchor point detection
#---------------------------------------------------------

#ANCHOR: Top level method to find an anchor
def anchor_pick(view, x, y, with_exclusion=false)
	#Picking the point
	@ph.do_pick x, y, 8
	
	#Finding an element which is not part of the selection
	@picked_face_info = nil
	lst_anchor = []
	nelt = @ph.count
	for i in 0..nelt
		path = @ph.path_at(i)
		break unless path && path.length > 0
		elt = path.last	
		next unless filter_check(elt)
		next if elt.class == Sketchup::SectionPlane && nelt > 1
		comp = (path.length == 1) ? nil : path[-2]
		unless with_exclusion && part_of_exclusion?([elt], comp)
			lst_anchor.push [i, elt, comp, @ph.transformation_at(i)]
		end	
	end
	
	#Getting the face if any
	lst_anchor.each do |a|
		i, elt, comp, tr = a
		if elt.instance_of?(Sketchup::Face)
			@picked_face_info = [elt, comp, tr] 
			break
		end	
	end
	
	#puts "\n****************************"
	lst_anchor.each do |a|
		i, elt, comp, tr = a
		#puts "LST ANCHOR #{i} = #{elt}"
	end
	
	#Analysing the anchors
	info = nil
	top_parent = nil
	unless lst_anchor.empty?
		info = anchor_analysis(view, lst_anchor, x, y)
		i = lst_anchor[0][0]
		top_parent = @ph.path_at(i)[0]
		top_parent = nil unless top_parent.instance_of?(Sketchup::ComponentInstance) || top_parent.instance_of?(Sketchup::Group) ||
		                        top_parent.instance_of?(Sketchup::Image)
	end
	
	#Anchor not found or on Text or Dimension (if so, privilege directions)
	info_def = nil
	if info
		pt, symb = info
		info_def = info if symb.to_s =~ /text/ || symb == :dimension || symb == :element
	end	
	
	#Calculating other inferences and taking the closest to camera
	info_void = anchor_in_void(view, x, y, info_def)
	info = anchor_compare_with_void(view, info, info_void)
	info = info_def unless info
	
	#Computing inferences for Center of active face or arc, component or group
	target, symb, info_comp = info
	if info && symb.to_s !~ /center/
		elt, comp, tr = info_comp
		if elt.instance_of?(Sketchup::Face)
			inference_center_face(elt, comp, tr)
		elsif elt.instance_of?(Sketchup::Edge) && 
			inference_center_arc_from_edge(elt, comp, tr)
		elsif elt.instance_of?(Sketchup::Vertex)
			inference_center_arc_from_vertex(elt, comp, tr)
		end
		
		#Computing the centers in the parent group or component
		if ![:center_comp, :center_comp_top, :center_group, :center_group_top, :bbox_corner, :bbox_face_center].include?(info[1])
			@hsh_comp_bbox_info = {} unless @option_inference_comp
			if top_parent != comp
				if comp
					inference_center_grouponent(top_parent, comp, tr)	
					inference_bbox_grouponent(top_parent, comp, tr)
				end	
			elsif comp
				inference_center_grouponent(comp, nil, @tr_id)	
				inference_bbox_grouponent(comp, nil, @tr_id)	
			end
		end			
	end
	
	#Computing the current top_parent and checking if mouse is in top parent	 
	@top_parent = top_parent if top_parent
	info_top = anchor_in_top_parent view, x, y
	@top_parent = nil if (info && !top_parent) || !info_top

	#Updating the last plane
	anchor_last_plane_record info

	#Appending the best face for edges and vertex
	anchor_best_normal(info, view, x, y)
		
	#Returning the info
	info
end

def anchor_picked_face_info(info)
	return @picked_face_info if @picked_face_info
	pt, type, info_comp = info
	(info_comp && info_comp[0].instance_of?(Sketchup::Face)) ? info_comp : nil
end

#ANCHOR: Record the last plane found under the mouse
def anchor_last_plane_record(info)
	pt, type, info_comp = info
	return unless pt
	elt, comp, tr = info_comp
	return unless elt && elt.valid?
	
	case type
	when :on_section_plane
		normal = Geom::Vector3d.new *(elt.get_plane[0..2])
		@last_plane_found = [pt, G6.transform_vector(normal, tr)]
	when :on_face, :center_face
		@last_plane_found = [pt, G6.transform_vector(elt.normal, tr)]
	when :center_arc, :center_polygon
		if elt.instance_of?(Sketchup::Edge)
			edge = elt
		elsif elt.instance_of?(Sketchup::Vertex)
			edge = elt.edges[0]
		end
		if edge
			pt2 = tr * edge.start.position
			pt3 = tr * edge.end.position
			@last_plane_found = G6.plane_by_3_points(pt, pt2, pt3)
		end	
	end	
end

#ANCHOR: Retain the closest point between element and void picking
def anchor_compare_with_void(view, info, info_void)
	return info unless info_void
	return info_void unless info
	pt, type, info_comp = info
	ptvoid,	type_void = info_void
	return info if type_void == :free
		
	eye = view.camera.eye
	(eye.distance(pt) <= eye.distance(ptvoid)) ? info : info_void
end

#ANCHOR: Analyse the anchor and determine its type
def anchor_analysis(view, lst_anchor, x, y)
	#Removing axis if in first position, due to a bug in Sketchup API where axis are always in first position
	axis_first = false
	if lst_anchor.length > 1 
		i0, elt0, comp0, tr0 = lst_anchor.first
		if !comp0 && elt0.bounds.diagonal == 0 && !elt0.instance_of?(Sketchup::ConstructionPoint)
			axis_first = true
			lst_anchor.shift
		end
	end
	
	#Initialization
	type = nil
	nb_anchor = lst_anchor.length
	ptxy = Geom::Point3d.new x, y, 0
	
	i0, elt0, comp0, tr0 = lst_anchor.first
	i1, elt1, comp1, tr1 = lst_anchor[1]
	i2, elt2, comp2, tr2 = lst_anchor[2]
	
	#Check if at center of a face
	info = anchor_at_center_face(view, x, y)
	return info if info
	
	#Check if at center of a arc
	info = anchor_at_center_arc(view, x, y)
	return info if info

	#Center of Component or Group
	info = anchor_at_center_comp(view, x, y)
	return info if info
	
	#Guide Point
	if elt0.instance_of?(Sketchup::ConstructionPoint)
		info = [tr0 * elt0.position, :cpoint, [elt0, comp0, tr0]]
		
	#Section Plane
	elsif elt0.instance_of?(Sketchup::SectionPlane)
		tinv = tr0.inverse
		ray = view.pickray(ptxy.x, ptxy.y)
		line = [tinv * ray[0], G6.transform_vector(ray[1], tinv)]
		ptinter = Geom.intersect_line_plane line, elt0.get_plane
		info = [tr0 * ptinter, :on_section_plane, [elt0, comp0, tr0]]
		
	#Guide Line
	elsif elt0.instance_of?(Sketchup::ConstructionLine)
		if elt1.instance_of?(Sketchup::ConstructionPoint)
			info = [tr1 * elt1.position, :cpoint, [elt1, comp1, tr1]]
		elsif nb_anchor == 1	
			info = anchor_on_guide_line(view, ptxy, elt0, comp0, tr0)
		elsif elt2.instance_of?(Sketchup::ConstructionPoint)
			info = [tr2 * elt2.position, :cpoint, [elt2, comp2, tr2]]
		else
			info = anchor_inter_cline(view, ptxy, elt0, comp0, tr0, elt1, comp1, tr1)		
		end	
		
	#Edge picked	
	elsif elt0.instance_of?(Sketchup::Edge)
		if nb_anchor == 1 || (comp0 == comp1 && elt1.instance_of?(Sketchup::Face) && elt1.edges.include?(elt0))
			info = anchor_on_edge(view, ptxy, elt0, comp0, tr0)
		elsif comp0 == comp1 && elt1.instance_of?(Sketchup::Edge)
			info = anchor_on_edge(view, ptxy, elt0, comp0, tr0)
		elsif elt1.instance_of?(Sketchup::Face) && !elt1.edges.include?(elt0)
			info = anchor_inter_edge_face(view, ptxy, elt0, comp0, tr0, elt1, comp1, tr1)
		else	
			info = anchor_on_edge(view, ptxy, elt0, comp0, tr0)
		end
		
	#Face picked	
	elsif elt0.instance_of?(Sketchup::Face)
		if nb_anchor == 1
			info = anchor_on_face(view, ptxy, elt0, comp0, tr0)
		elsif nb_anchor == 2 && comp0 == comp1 && elt1.instance_of?(Sketchup::Edge) && elt0.edges.include?(elt1)
			info = anchor_on_edge(view, ptxy, elt1, comp1, tr1)
		elsif elt1.instance_of?(Sketchup::Edge) && !elt0.edges.include?(elt1)
			info = anchor_on_face_at_vertex(view, ptxy, elt0, comp0, tr0)
			info = anchor_inter_edge_face(view, ptxy, elt1, comp1, tr1, elt0, comp0, tr0) unless info
		elsif elt1.instance_of?(Sketchup::ConstructionPoint)
			info = [tr1 * elt1.position, :cpoint, [elt1, comp1, tr1]]
		elsif elt1.instance_of?(Sketchup::ConstructionLine)
			info = anchor_inter_cline(view, ptxy, elt1, comp1, tr1, elt0, comp0, tr0)		
		elsif elt1.instance_of?(Sketchup::Text)
			info = anchor_on_text(view, ptxy, elt1, comp1, tr1)
		end	
		info = anchor_on_face(view, ptxy, elt0, comp0, tr0) unless info
	
	#Text picked	
	elsif elt0.instance_of?(Sketchup::Text)
		info = anchor_on_text(view, ptxy, elt0, comp0, tr0)

	#Dimension picked	
	elsif defined?(Sketchup::Dimension) && elt0.is_a?(Sketchup::Dimension)
		pt = @ip.position.project_to_line view.pickray(x, y)
		info = [pt, :dimension, [elt0, comp0, tr0]]

	#Model Axes	and other elements
	elsif elt0.instance_of?(Sketchup::Drawingelement)
		if !comp0 && elt0.bounds.diagonal == 0
			info = anchor_on_axis(view, ptxy)
		else
			pt = @ip.position.project_to_line view.pickray(x, y)
			info = [pt, :dimension, [elt0, comp0, tr0]]
		end	

	else
		pt = @ip.position.project_to_line view.pickray(x, y)
		info = [pt, :element, [elt0, comp0, tr0]]
	end
	
	#Treatment of axis
	if axis_first && info
		info_axis = anchor_on_axis(view, ptxy)
		info = anchor_compare_with_void(view, info, info_axis)
	end
	
	info
end

#ANCHOR: Find best face for edges and vertex
def anchor_best_normal(info, view, x, y)
	return unless info
	pt, type, info_comp = info
	elt, comp, tr = info_comp
	return unless elt && elt.valid?
	
	#Finding the best face based on the picked point for an Edge or a vertex
	if elt.instance_of?(Sketchup::Vertex) || elt.instance_of?(Sketchup::Edge)
		trinv = tr.inverse
		eye = view.camera.eye
		ray = view.pickray(x, y)
		dmin = nil
		face0 = nil
		normal0 = nil
		elt.faces.each do |face|
			plane = [tr * face.vertices[0].position, G6.transform_vector(face.normal, tr)]
			ptinter = Geom.intersect_line_plane ray, plane
			next unless ptinter && [1, 2, 4].include?(face.classify_point(trinv * ptinter))
			d = eye.distance(ptinter)
			if !dmin || d < dmin
				dmin = d
				face0 = face
				normal0 = G6.transform_vector(face0.normal, tr)
			end
		end	
	
	#Using the bbox face for component center or corner
	elsif [:center_comp, :center_group, :center_comp_top, :center_group_top, :bbox_corner].include?(type)
		info_top = anchor_in_bbox(view, x, y, elt, elt.transformation * tr)
		normal0 = info_top[3] if info_top
	end
	
	#Storing the information in the info
	info[3] = normal0
end

#ANCHOR: try to find a point
def anchor_in_void(view, x, y, info_def=nil)
	ptref, type_ref, info_comp_ref = @origin_info
	
	#Check if at center of a face
	info = anchor_at_center_face(view, x, y)
	return info if info
	
	#Check if at center of a arc
	info = anchor_at_center_arc(view, x, y)
	return info if info

	#Check if at center of a Grouponent
	info = anchor_at_center_comp(view, x, y)
	return info if info
	
	#Testing if twin_center if on the face
	info = anchor_at_twin(view, x, y)
	return info if info
	
	#Bounding box of groups and components
	info = anchor_at_bbox_comp(view, x, y)
	return info if info
	
	#No point of reference
	unless ptref || info_def
		view.tooltip = nil
		return nil
	end	
	
	#Forced direction
	return nil if @forced_vector_dir || @forced_plane_dir || @mode == :origin
	
	#Return default direction if any
	return info_def if info_def
	
	#Other cases
	if ptref
		plane = [ptref, Z_AXIS]
		ray = view.pickray x, y
		target = Geom.intersect_line_plane ray, plane
		view.tooltip = nil
		return [target, :free, nil]	
	end
	nil	
end

#ANCHOR: Check if the mouse is within the bounding box of a component or group
def anchor_in_top_parent(view, x, y)
	@info_top_parent = nil
	return nil unless @top_parent && @top_parent.valid?
	anchor_in_bbox view, x, y, @top_parent, @top_parent.transformation
end

#ANCHOR: Check if the mouse is within the bounding box of a component or group
def anchor_in_bbox(view, x, y, comp, tr)
	return nil unless comp && comp.valid?
	bbox = G6.grouponent_definition(comp).bounds
	
	#List of panels of the bounding box
	corners3d = G6.grouponent_corners(comp, tr)
	corners2d = corners3d.collect { |pt| p = view.screen_coords(pt) ; p.z = 0 ; p }
	ls_panels = []
	flat = (bbox.corner(0) == bbox.corner(4))
	if flat
		ls_panels.push [0, 1, 3, 2]
	else
		ls_panels.push [0, 1, 3, 2], [0, 1, 5, 4], [1, 3, 7, 5], [3, 2, 6, 7], [2, 0, 4, 6], [4, 5, 7, 6]
	end
	laxes = [Z_AXIS.reverse, Y_AXIS.reverse, X_AXIS, Y_AXIS, X_AXIS.reverse, Z_AXIS]

	#Checking in 3D
	ray = view.pickray x, y
	eye = view.camera.eye
	
	ptxy = Geom::Point3d.new x, y, 0
	info_inter = nil
	dmin = nil
	normal = nil
	ls_panels.each_with_index do |ipts, ipanel|
		pts = ipts.collect { |i| corners3d[i] }
		pts2d = ipts.collect { |i| corners2d[i] }
		plane = Geom.fit_plane_to_points pts[0], pts[1], pts[2]
		ptinter = Geom.intersect_line_plane ray, plane
		next unless ptinter
		d = eye.distance ptinter
		if (!dmin || d < dmin) && Geom.point_in_polygon_2D(ptxy, pts2d, true)
			info_inter = [ptinter, comp, ipanel, pts, pts2d]
			normal = laxes[ipanel]
			dmin = d
		end
	end
	
	#No intersection found
	return nil unless info_inter
	@info_top_parent = info_inter if comp == @top_parent

	#Otherwise on a bbox face
	normal = G6.transform_vector(normal, tr)
	ptinter, = info_inter
	[ptinter, :bbox_face, [comp, comp, tr], normal]
end

#ANCHOR: Test predefined directions
def anchor_follow_direction(view, x, y, ptref)
	ls_info = @origin_privileged_directions
	return unless ls_info && ptref
	
	#Calculate origin and target in 2D
	ptxy = Geom::Point3d.new x, y, 0
	origin2d = view.screen_coords ptref
	origin2d.z = 0
	vec2d = origin2d.vector_to ptxy
	size = view.pixels_to_model 30, ptref

	#Cycling through the possible direction inference
	ecart_min = @inference_distance
	result = nil
	ls_info.each do |symb, vec3d|
		next unless vec3d.valid?
		pt = ptref.offset vec3d, size
		pt2d = view.screen_coords pt
		pt2d.z = 0
		axis2d = origin2d.vector_to pt2d
		axis2d = axis2d.reverse if axis2d % vec2d < 0
		next if vec2d.angle_between(axis2d) > @inference_angle
		d = origin2d.distance ptxy
		pt2d = origin2d.offset axis2d, d
		ecart = ptxy.distance(pt2d)
		if ecart < ecart_min
			ecart_min = ecart
			result = [symb, vec3d]
		end	
	end
	
	#Calculating the target if there is a direction matching
	if result
		symb, vec3d = result
		lpt = Geom.closest_points [ptref, vec3d], view.pickray(x, y)
		target = lpt.first
		return [target, symb] if target	
	end	
	nil
end

#ANCHOR: Check if the target is at a center of a face
def anchor_at_center_face(view, x, y)
	return nil unless @center_face_info
	center, normal, info_face = @center_face_info
	center2d = view.screen_coords center
	if (center2d.x - x).abs <= @precision && (center2d.y - y).abs <= @precision
		return [center, :center_face, info_face]
	end
	nil
end

#ANCHOR: Check if the target is at a center of an arccurve
def anchor_at_center_arc(view, x, y)
	return nil unless @hsh_arc_center_info
	@hsh_arc_center_info.each do |id, info|
		center, curve, info_comp = info
		center2d = view.screen_coords center
		if (center2d.x - x).abs <= @precision && (center2d.y - y).abs <= @precision
			code = (curve.vertices.first == curve.vertices.last) ? :center_polygon : :center_arc
			return [center, code, info_comp]
		end
	end	
	nil
end

#ANCHOR: Check if the target is at a center of an arccurve
def anchor_at_twin(view, x, y)
	return nil unless @twin_info
	@twin_info.each do |info|
		ptinter, pt1, pt2 = info
		center2d = view.screen_coords ptinter
		if (center2d.x - x).abs <= @precision && (center2d.y - y).abs <= @precision
			return [ptinter, :center_twin, nil]
		end
	end	
	nil
end

#ANCHOR: Check if the target is at a center of Component or Group
def anchor_at_center_comp(view, x, y)
	return nil unless @option_inference_comp && @hsh_comp_center_info
	@hsh_comp_center_info.each do |id, info|
		center, comp, info_comp = info
		top_parent = info_comp[1]
		center2d = view.screen_coords center
		if (center2d.x - x).abs <= @precision && (center2d.y - y).abs <= @precision
			if comp.instance_of?(Sketchup::Group)
				code = (comp == top_parent) ? :center_group_top : :center_group
			else
				code = (comp == top_parent) ? :center_comp_top : :center_comp
			end	
			return [center, code, info_comp]
		end
	end	
	nil
end

#ANCHOR: Check if the target is at a corner of face center of the bounding box of Component or Group
def anchor_at_bbox_comp(view, x, y)
	return nil unless @option_inference_comp && @hsh_comp_bbox_info
	ptxy = Geom::Point3d.new x, y, 0
	
	info_corner = nil
	dmin = @precision + 1
	@hsh_comp_bbox_info.each do |id, info|
		corners, comp, info_comp = info
		corners.each do |pt|
			pt2d = view.screen_coords pt
			pt2d.z = 0
			d = ptxy.distance pt2d
			if d < dmin
				dmin = d
				info_corner = [pt, :bbox_corner, info_comp]
			end	
		end
	end	
	info_corner
end

#ANCHOR: Check if the target is at Origin or on an axis
def anchor_on_axis(view, ptxy)
	#Test at ORIGIN
	origin2d = view.screen_coords ORIGIN
	origin2d.z = 0
	if (origin2d.x - ptxy.x).abs <= @precision  && (origin2d.y - ptxy.y).abs <= @precision
		return [ORIGIN, :origin ]
	end
	
	#Test on axes
	size = view.pixels_to_model 50, ORIGIN
	dmin = @precision
	info = nil
	[[:x, X_AXIS], [:y, Y_AXIS], [:z, Z_AXIS]].each do |symb, vec|
		pt2d = view.screen_coords ORIGIN.offset(vec, size)
		pt2d.z = 0
		d = ptxy.distance_to_line [origin2d, pt2d]
		if d <= dmin
			ptproj = ptxy.project_to_line [origin2d, pt2d]
			lpt = Geom.closest_points [ORIGIN, vec], view.pickray(ptproj.x, ptproj.y)
			if lpt
				info = [lpt.first, symb]
				dmin = d
			end	
		end			
	end
	(info) ? info : nil
end

#ANCHOR: Check if the target is on a GUIDE LINE
def anchor_on_guide_line(view, ptxy, cline, comp, tr)
	#Checking the extremity of Guide line (if any)
	[cline.start, cline.end].each do |pt3d|
		next unless pt3d
		pt2d = view.screen_coords pt3d
		pt2d.z = 0
		if ptxy.distance(pt2d) <= @precision
			return [pt3d, :end_cline, [cline, comp, tr]]
		end
	end
	
	#Otherwise, check on the line
	line = [tr * cline.position, G6.transform_vector(cline.direction, tr)]
	lpt = Geom.closest_points line, view.pickray(ptxy.x, ptxy.y)
	(lpt) ? [lpt.first, :on_cline, [cline, comp, tr]] : nil
end

#ANCHOR: Check if the target is on a GUIDE LINE
def anchor_on_text(view, ptxy, sutext, comp, tr)	
	#Text with no anchor point
	unless sutext.point
		return [@ip.position, :on_text, [sutext, comp, tr]]
	end
	
	#Close to the anchor
	pt_anchor = tr * sutext.point
	pt2d_anchor = view.screen_coords pt_anchor
	if pt2d_anchor.distance(ptxy) <= @precision
		return [pt_anchor, :text_anchor, [sutext, comp, tr]]
	end
	
	#On the Leader line
	pt_lab = pt_anchor.offset G6.transform_vector(sutext.vector, tr)
	pt2d_lab = view.screen_coords pt_lab
	d, = G6.proximity_point_segment(ptxy, pt2d_anchor, pt2d_lab)
	if d <= @precision
		lpt = Geom.closest_points [pt_anchor, pt_lab], view.pickray(ptxy.x, ptxy.y)
		return [lpt[0], :text_leader, [sutext, comp, tr]] if lpt&& lpt[0]
	end
	
	#On the Text label
	plane = [pt_lab, view.camera.direction]
	pt = Geom.intersect_line_plane view.pickray(ptxy.x, ptxy.y), [pt_lab, view.camera.direction]
	(pt) ? [pt, :on_text, [sutext, comp, tr]] : nil
end

#ANCHOR: Check if the target is on an EDGE
def anchor_on_edge(view, ptxy, edge, comp, tr)
	dmin = @precision
	info = nil
	
	#Checking vertices and midpoint of edge
	vx1 = edge.start
	vx2 = edge.end
	pt1 = tr * vx1.position
	pt2 = tr * vx2.position
	ptmid = Geom.linear_combination 0.5, pt1, 0.5, pt2
	[[:vertex, pt1, vx1], [:vertex, pt2, vx2], [:midpoint, ptmid, edge]].each do |symb, pt, e|
		next if symb == :vertex && !anchor_vertex_visible(e)
		pt2d = view.screen_coords(pt)
		pt2d.z = 0
		d = ptxy.distance pt2d
		if d <= dmin
			info = [pt, symb, [e, comp, tr]]
			dmin = d
		end	
	end
		
	#Checking is close to edge line
	unless info
		lpt = Geom.closest_points [pt1, pt2], view.pickray(ptxy.x, ptxy.y)
		pt = lpt[0]
		info = [pt, :on_edge, [edge, comp, tr]] if pt
	end	

	info
end

#Check if an edge is visible
def anchor_edge_visible(e)
	return true if @rendering_options["DrawHidden"] || e.faces.length < 2
	!(e.hidden? || e.soft?)
end

#Check if a vertex is visible
def anchor_vertex_visible(vx)
	return true if @rendering_options["DrawHidden"] || vx.faces.length < 2
	vx.edges.find { |e| anchor_edge_visible(e) }
end

#ANCHOR: Check if the target is on a FACE
def anchor_on_face(view, ptxy, face, comp, tr)
	#Test if close to center of face
	unless !@rendering_options["DrawHidden"] && face.edges.find { |e| e.smooth? || e.soft? || e.hidden? }
		info = anchor_at_center_face(view, ptxy.x, ptxy.y)
		return info if info
	end
	
	#Testing if point picked is on the vertex of a face
	info = anchor_on_face_at_vertex(view, ptxy, face, comp, tr)
	return info if info
	
	#Testing if a Comp center if on the face
	info = anchor_at_center_comp(view, ptxy.x, ptxy.y)
	return info if info

	#Testing if an arc center if on the face
	info = anchor_at_center_arc(view, ptxy.x, ptxy.y)
	return info if info

	#Testing if twin_center if on the face
	info = anchor_at_twin(view, ptxy.x, ptxy.y)
	return info if info
	
	#Position on the face
	pt1 = tr * face.vertices[0].position
	normal = G6.transform_vector(face.normal, tr)
	pt = Geom.intersect_line_plane view.pickray(ptxy.x, ptxy.y), [pt1, normal]
	
	(pt) ? [pt, :on_face, [face, comp, tr]] : nil
end

#ANCHOR: Check if the target is on a FACE at a Vertex
def anchor_on_face_at_vertex(view, ptxy, face, comp, tr)
	dmin = @precision
	ptmin = vx0 = nil
	face.vertices.each do |vx|
		next unless anchor_vertex_visible(vx)
		pt3d = tr * vx.position
		pt2d = view.screen_coords(pt3d)
		d = pt2d.distance ptxy
		if d <= dmin
			vx0 = vx
			ptmin = pt3d
		end	
	end
	(ptmin) ? [ptmin, :vertex, [vx0, comp, tr]] : nil
end

#ANCHOR: Check if the target is at the intersection of an Edge and a FACE
def anchor_inter_edge_face(view, ptxy, edge, compe, tre, face, compf, trf)
	normal = G6.transform_vector(face.normal, trf)
	plane = [trf * face.vertices[0].position, normal]
	pt1e = tre * edge.start.position
	pt2e = tre * edge.end.position
	veced = pt1e.vector_to pt2e
	
	#Edge parallel to face
	point, type = info_edge = anchor_on_edge(view, ptxy, edge, compe, tre)
	if pt1e.on_plane?(plane) && pt2e.on_plane?(plane)
		return info_edge
	elsif type && type != :on_edge
		pt = Geom.intersect_line_plane view.pickray(ptxy.x, ptxy.y), plane
		eye = view.camera.eye
		return info_edge if eye.distance(pt) >= eye.distance(point)
	end
	
	#Intersection Edge with Face
	pt = Geom.intersect_line_plane [pt1e, pt2e], [trf * face.vertices[0].position, normal]
	
	#Checking if real intersection
	if pt
		if [1, 2, 4].include?(face.classify_point(trf.inverse * pt))
			return [pt, :inter_edge_face, [edge, compe, tre], [face, compf, trf]]
		else
			return info_edge
		end
	end	
	
	#Point on face
	anchor_on_face(view, ptxy, face, compf, trf)
end

#ANCHOR: Check if the target is at the intersection of an Edge and a FACE
def anchor_inter_cline(view, ptxy, cline, comp, tr, elt, comp2, tr2)
	line = [tr * cline.position, G6.transform_vector(cline.direction, tr)]
	
	#Intersection Cline with Face
	if elt.instance_of?(Sketchup::Face)
		face = elt
		#normal = G6.transform_vector(face.normal, tr)
		normal = G6.transform_vector(face.normal, tr2)
		plane = [tr2 * face.vertices[0].position, normal]
		pt = Geom.intersect_line_plane line, plane
		if pt && view.screen_coords(pt).distance(ptxy) < 10 && [1, 2, 4].include?(face.classify_point(tr2.inverse * pt))
			return [pt, :inter_cline_face, [cline, comp, tr], [face, comp2, tr2]]
		end
	
	#Intersection Cline with Edge
	elsif elt.instance_of?(Sketchup::Edge)
		edge = elt
		eline = [tr2 * edge.start.position, tr2 * edge.end.position]
		pt = Geom.intersect_line_line line, eline
		if pt
			return [pt, :inter_cline_edge, [cline, comp, tr], [edge, comp2, tr2]]
		end

	#Intersection Cline with Cline
	elsif elt.instance_of?(Sketchup::ConstructionLine) && elt != cline
		pt = Geom.intersect_line_line line, [tr2 * elt.position, G6.transform_vector(elt.direction, tr2)]
		if pt
			return [pt, :inter_cline_cline, [cline, comp, tr], [elt, comp2, tr2]]
		end
	end
	
	#Point on cline (default)
	anchor_on_guide_line(view, ptxy, cline, comp, tr)
end

#ANCHOR: Compute the tooltip based on the anchor type
def anchor_tooltip(info)
	target, type, info_comp = info
	comp = (info_comp.class == Array) ? info_comp[1] : nil

	#Tip for grouponent
	if type == :center_comp_top || type == :center_group_top
		tip_comp = nil
	elsif comp.instance_of?(Sketchup::Group)
		tip_comp = T6[:INFER_InGroup]
	elsif comp.instance_of?(Sketchup::ComponentInstance)
		tip_comp = T6[:INFER_InComponent]
	elsif comp.instance_of?(Sketchup::Image)
		tip_comp = T6[:INFER_InImage]
	elsif type.to_s =~ /remote_(.*)/
		tip_comp = T6[:INFER_FromRemote]
		type = $1.intern
	else	
		tip_comp = nil
	end

	#Tip for element
	tip = nil
	case type
	when :cpoint
		tip = T6[:INFER_Cpoint]
	when :on_cline
		tip = T6[:INFER_Cline]
	when :end_cline
		tip = T6[:INFER_EndCline]
	when :vertex
		tip = T6[:INFER_Endpoint]
	when :midpoint
		tip = T6[:INFER_Midpoint]
	when :center_twin
		tip = T6[:INFER_Twin]
	when :center_face
		tip = T6[:INFER_CenterFace]
	when :center_arc
		tip = T6[:INFER_CenterArc]
	when :center_comp
		tip = T6[:INFER_CenterComp]
	when :center_comp_top
		tip = T6[:INFER_CenterCompTop]
	when :center_group
		tip = T6[:INFER_CenterGroup]
	when :center_group_top
		tip = T6[:INFER_CenterGroupTop]
	when :center_polygon
		tip = T6[:INFER_CenterPolygon]
	when :on_edge
		tip = T6[:INFER_OnEdge] 
	when :on_face
		tip = T6[:INFER_OnFace]
	when :on_section_plane
		tip = T6[:INFER_OnSectionPlane]
	when :origin
		tip = T6[:INFER_AtOrigin]
	when :x
		tip = T6[:INFER_OnXAxis]
	when :y
		tip = T6[:INFER_OnYAxis]
	when :z
		tip = T6[:INFER_OnZAxis]
	when :x, :follow_x
		tip = T6[:INFER_AlongXAxis]
	when :y, :follow_y
		tip = T6[:INFER_AlongYAxis]
	when :z, :follow_z
		tip = T6[:INFER_AlongZAxis]
	when :follow_z_comp
		tip = T6[:INFER_AlongZAxisComp]
	when :follow_x_comp
		tip = T6[:INFER_AlongXAxisComp]
	when :follow_y_comp
		tip = T6[:INFER_AlongYAxisComp]
	when :perp_edge_face
		tip = T6[:INFER_PerpEdgeFace]
	when :perp_edge_alone
		tip = T6[:INFER_PerpEdgeAlone]
	when :perp_bisector
		tip = T6[:INFER_PerpBisector]
	when :follow_edge
		tip = T6[:INFER_AlongEdgeDirection]
	when :plane_of_edges
		tip = T6[:INFER_PerpTwoEdges]
	when :follow_cline
		tip = T6[:INFER_AlongCline]
	when :follow_bbox
		tip = T6[:INFER_AlongAxesBbox]
	when :follow_normal
		tip = T6[:INFER_AlongNormalFace]
	when :follow_section_plane
		tip = T6[:INFER_AlongSectionPlane]
	when :inter_edge_face
		tip = T6[:INFER_InterEdgeFace]
	when :inter_cline_face
		tip = T6[:INFER_InterClineFace]
	when :inter_cline_edge
		tip = T6[:INFER_InterClineEdge]
	when :inter_cline_cline
		tip = T6[:INFER_InterClineCline]
	when :inter_vector_face
		tip = T6[:INFER_InterVectorFace]
	when :inter_vector_section
		tip = T6[:INFER_InterVectorSectionPlane]
	when :inter_vector_edge
		tip = T6[:INFER_InterVectorEdge]
	when :inter_vector
		tip = T6[:INFER_InterVector]
	when :inter_vector_remote
		tip = T6[:INFER_InterVectorRemote]
	when :inter_plane_remote
		tip = T6[:INFER_InterPlaneRemote]
	when :inter_plane_edge
		tip = T6[:INFER_InterPlaneEdge]
	when :text_anchor
		tip = T6[:INFER_TextAnchor]
	when :on_text
		tip = T6[:INFER_OnTextLabel]
	when :text_leader
		tip = T6[:INFER_OnTextLeader]
	when :dimension
		tip = T6[:INFER_OnDimension]
	when :element
		tip = T6[:INFER_OnDrawingElement]
	when :bbox_corner
		tip = T6[:INFER_AtCornerBbox]
	when :bbox_face
		tip = T6[:INFER_OnFaceBbox]
	when :parallel_to_edge
		tip = T6[:INFER_ParallelToEdge]
	when :parallel_to_cline
		tip = T6[:INFER_ParallelToCLine]
	when :point
		tip = T6[:INFER_SpecifiedPoint]
	when :protractor
		tip = T6[:INFER_ProtractorAngle]
	end
	
	#Full tip text
	if tip
		tip = tip + ((tip_comp) ? " [#{tip_comp}]" : "")
	end
	
	tip
end

#---------------------------------------------------------
# INFERENCE: Manage inferences
#---------------------------------------------------------

def inference_center_face(face, comp, tr)
	#@center_face_info = nil
	return unless face.valid?
	unless !@rendering_options["DrawHidden"] && face.edges.find { |e| e.smooth? || e.soft? || e.hidden? }
		center, = G6.face_centroid_area(face)
		@center_face_info = [tr * center, G6.transform_vector(face.normal, tr), [face, comp, tr]]
	end	
end

#INFERENCE: Store the center of arc curves for display and inference when specified by an edge
def inference_center_arc_from_edge(edge, comp, tr)
	curve = edge.curve
	return unless curve.instance_of?(Sketchup::ArcCurve)
	
	@hsh_arc_center_info = {} unless @hsh_arc_center_info
	
	id = "#{curve.entityID}-#{tr.to_a}"
	@hsh_arc_center_info[id] = [tr * curve.center, curve, [edge, comp, tr]]
end

#INFERENCE: Store the center of arc curves for display and inference when specified by a vertex
def inference_center_arc_from_vertex(vx, comp, tr)
	vx.edges.each { |edge| inference_center_arc_from_edge(edge, comp, tr) }
end

#INFERENCE: Store the center of component or group
def inference_center_grouponent(top_parent, comp, tr)
	@hsh_comp_center_info = {} unless @hsh_comp_center_info
	
	if top_parent && !top_parent.instance_of?(Sketchup::Image)
		id = top_parent.entityID
		@hsh_comp_center_info[id] = [top_parent.bounds.center, top_parent, [top_parent, top_parent, @tr_id]]
	end
	
	if comp && comp != top_parent && !comp.instance_of?(Sketchup::Image)
		id = comp.entityID
		t = tr * comp.transformation.inverse
		@hsh_comp_center_info[id] = [t * comp.bounds.center, comp, [comp, top_parent, t]]
	end
end

#INFERENCE: Store the corner of bounding boxes of component or group
def inference_bbox_grouponent(top_parent, comp, tr)
	@hsh_comp_bbox_info = {} unless @hsh_comp_bbox_info && @option_inference_comp
	
	if top_parent && !top_parent.instance_of?(Sketchup::Image)
		id = top_parent.entityID
		corners = G6.grouponent_corners(top_parent, top_parent.transformation)
		@hsh_comp_bbox_info[id] = [corners, top_parent, [top_parent, top_parent, tr]]
	end
	
	if comp && comp != top_parent && !comp.instance_of?(Sketchup::Image)
		id = comp.entityID
		corners = G6.grouponent_corners(comp, tr)
		@hsh_comp_bbox_info[id] = [corners, comp, [comp, top_parent, tr]]
	end
end

#INFERENCE: Reset the position of the centers at their original position
def inference_update_centers
	#Face center
	if @center_face_info
		center, normal, face_info = @center_face_info
		inference_center_face(*face_info)
	end

	#Centers for Arc curves
	if @hsh_arc_center_info
		@hsh_arc_center_info.each do |id, info|
			center, edge, curve_info = info
			inference_center_arc_from_edge(*curve_info)
		end
	end

	#Centers for Components and Groups
	if @hsh_comp_center_info
		hsh = {}
		@hsh_comp_center_info.each do |id, info|
			center, comp, comp_info = info
			comp, parent, tr = comp_info
			id = "#{comp.entityID} - #{tr.to_a}"
			hsh[id] = [tr * comp.bounds.center, comp, comp_info]
		end
		@hsh_comp_center_info = hsh
	end
end

#INFERENCE: Remove all defined centers
def inference_reset
	@center_face_info = @hsh_arc_center_info = @hsh_comp_center_info = @hsh_comp_bbox_info = nil
	remote_reset
end

#---------------------------------------------------------
# DRAW: Manage drawing
#---------------------------------------------------------

#DRAW: Draw top method
def draw_all(view)		
	#Draw the directions if applicable
	draw_hi_curve view
	draw_plane_direction view
	draw_vector_direction view if @option_vector
	draw_protractor view
	draw_parallel view
	draw_remote view
	draw_twin view
	
	#Drawing the origin
	draw_origin view, (@mode == :origin)
	
	#Drawing the target
	if @mode == :target
		draw_target view, false	
		draw_origin_target view
	end
	draw_distance view	
end

#DRAW: Draw the origin, targets and joining line	
def draw_origin(view, flg_ip=true)
	draw_centers view
	draw_highlight_edge(view, @origin_info) unless @pt_target
	
	#Drawing the origin
	if flg_ip && @pt_origin && @ip_origin && @ip_origin.valid? && @ip_origin.display?
		@ip_origin.draw view 
	elsif @origin_info
		mark_draw view, @origin_info 
	end
end

#DRAW: Draw the origin, targets and joining line	
def draw_target(view, flg_ip=true)
	draw_centers view
	
	#Drawing the origin
	if flg_ip && @pt_target && @ip_target && @ip_target.valid? && @ip_target.display?
		@ip_target.draw view 
	elsif @pt_target
		target, symb_target, info_target = @target_info
		drive, symb_drive, info_drive = @info_drive3d
		draw_highlight_edge(view, @info_drive3d)
		if symb_target != symb_drive
			mark_draw view, [@pt_target, symb_target, info_target]
		end	
		mark_draw view, [@pt_drive3d, symb_drive, info_drive]
	end
end
	
#DRAW: Draw the origin, targets and joining line	
def draw_origin_target(view)
	return unless @pt_origin && @pt_target

	#Drawing the line between the origin and target
	vec = @pt_origin.vector_to @pt_target
	return unless vec.valid? 
	
	origin2d = view.screen_coords @pt_origin
	target2d = view.screen_coords @pt_target
	origin2d.z = target2d.z = 0
	if (@forced_vector_dir || @forced_plane_dir || @remoting_dir) && @pt_drive2d != target2d
		code = (@forced_vector_dir) ? @forced_vector_code : @forced_plane_code
		color = @hsh_color_vector[code]
		view.line_width = 1
		view.line_stipple = '.'
		view.drawing_color = color
		view.draw2d GL_LINE_STRIP, [@pt_drive2d, target2d]				
	end
	pt, symb = @info_drive3d
	if @forced_vector_dir
		view.line_width = 3
		code = @forced_vector_code
	elsif symb.to_s =~ /follow/
		view.line_width = 2
		code = symb
	else	
		view.line_width = @line_width
		code = (@target_info) ? @target_info[1] : 'free'
	end	

	#Drawing the line from origin to target
	view.drawing_color = @hsh_color_vector[code]
	if @line_mode
		view.drawing_color = @line_color if code.to_s == 'free'
		view.line_stipple = (@line_stipple) ? @line_stipple : ''
		view.draw GL_LINE_STRIP, [@pt_origin, @pt_target]
	else
		view.line_stipple = '_'
		view.draw2d GL_LINE_STRIP, [origin2d, target2d]
	end	
end
	
#DRAW: Draw the distance
def draw_distance(view)
	return if @hide_distance
	return unless @pt_origin
	if @remoting_dir && @remote_origin
		pt = (@mode == :origin) ? @pt_origin : @pt_target
		return unless pt
		d = @remote_origin.distance pt
		bkcolor, frcolor = @color_box_distance_remote
		pt2d = (@pt_drive2d) ? @pt_drive2d : view.screen_coords(@remote_origin)
		pt2d.z = 0
	elsif !@pt_target || !@pt_drive2d
		return
	else
		if @pt_origin && @pt_target && @pt_origin == @pt_target && @pt_origin_prev
			d = @pt_origin_prev.distance @pt_target
		else	
			d = @pt_origin.distance @pt_target
		end	
		bkcolor, frcolor = @color_box_distance
		pt2d = @pt_drive2d
	end	
	tip = G6.format_length_general(d) + ' '
	tx_angle = protractor_angle_text true
	tip += '  ' + tx_angle if tx_angle
	G6.draw_rectangle_text(view, pt2d.x, pt2d.y, tip, bkcolor, frcolor)
end

#DRAW: Draw the Vector direction if any
def draw_vector_direction(view)
	return unless @origin_info && @forced_vector_dir
	ptori, code = @origin_info

	#Computing the guide line in 2D
	pt = ptori.offset @forced_vector_dir, 1
	pt1_2d = view.screen_coords ptori
	pt2_2d = view.screen_coords pt
	vec2d = pt1_2d.vector_to pt2_2d
	pt1 = pt1_2d.offset vec2d, view.vpwidth * 2
	pt2 = pt1_2d.offset vec2d, -view.vpwidth * 2
	
	#Drawing the guide line
	view.line_width = (@pt_target) ? 1 : 2
	view.line_stipple = '_'
	symb = @forced_vector_code
	view.drawing_color = @hsh_color_vector[symb]
	view.draw2d GL_LINE_STRIP, [pt1, pt2]
end

#DRAW: Draw the visual plane if applicable
def draw_plane_direction(view)
	draw_vector_direction(view)
	return unless @origin_info
	if @forced_plane_dir && @forced_plane_code
		normal = @forced_plane_dir
		code = @forced_plane_code
	elsif protractor_shown?
		normal = @protractor_normal
		code = @protractor_code	
	#elsif @implicit_plane_normal && (@option_planar || @implicit_show_plane)
	elsif @implicit_plane_normal && @implicit_show_plane
		normal = @implicit_plane_normal
		code = :implicit_plane
	else
		return
	end	
	ptori, = @origin_info
	
	#Size and axes of the plane
	axes = normal.axes
	if @pt_target && ptori != @pt_target && code != :implicit_plane
		size = ptori.distance(@pt_target) * 1.2	
	else
		size = view.pixels_to_model(50, ptori)
	end
	size2 = size * 2
	
	#Computing the points
	vecx = axes[0]
	vecy = axes[1]
	pt1 = ptori.offset(vecx, size).offset(vecy, size)
	pt2 = pt1.offset vecx, -size2
	pt3 = pt2.offset vecy, -size2
	pt4 = pt3.offset vecx, size2
	pts = [pt1, pt2, pt3, pt4]
	
	#Drawing the plane
	view.line_width = 2
	view.line_stipple = ''
	if G6.su_capa_color_polygon
		view.drawing_color = @hsh_bkcolor_plane[code]
		view.draw GL_POLYGON, pts
	end	
	view.drawing_color = @hsh_frcolor_plane[code]
	view.draw GL_LINE_LOOP, pts
end

#DRAW: Draw the curves when selection is a single vertex
def draw_hi_curve(view)
	return unless @hi_curve_pts
	view.line_width = 2
	view.line_stipple = ''
	view.drawing_color = 'orchid'
	@hi_curve_pts.each do |pts|
		view.draw GL_LINE_STRIP, pts.collect { |pt| G6.small_offset(view, pt) }
	end	
end

#DRAW: Draw marks on edges under target
def draw_highlight_edge(view, info)
	target, symb, info_comp = info
	return unless target 
	return unless symb == :on_edge || symb == :midpoint || symb == :vertex

	if symb == :vertex
		vx, comp, tr = info_comp
		return unless vx.valid?
		lvx = []
		vx.edges.each do |edge|
			lvx.push edge.other_vertex(vx)
		end			
	else
		edge, comp, tr = info_comp
		return unless edge.valid?
		lvx = [edge.start, edge.end]
	end
	
	quads = []
	eye = view.camera.eye
	lvx.each do |vx|
		pt = tr * vx.position
		size = view.pixels_to_model 3, pt
		size2 = 2 * size
		vecx, vecy = (eye.vector_to(pt)).axes
		pt0 = pt.offset vecx, size
		pt1 = pt0.offset vecy, size
		pt2 = pt1.offset vecx, -size2
		pt3 = pt2.offset vecy, -size2
		pt4 = pt3.offset vecx, size2
		quads += [pt1, pt2, pt3, pt4].collect { |pt| G6.small_offset view, pt }
	end
	
	color = 'lightblue'
	view.drawing_color = color
	view.draw GL_QUADS, quads
end

#DRAW: Draw the center of the active face and arc curves if applicable
def draw_centers(view)
	#Draw the top_axes
	draw_top_axes view

	#Draw the bounding boxes for components and groups
	if @hsh_comp_bbox_info
		@hsh_comp_bbox_info.each do |id, info|
			corners, comp, info_comp = info
			next unless comp.valid?
			comp, top_parent, tr = info_comp
			lix = (corners.length == 4) ? @bbox_lines_indexing[0..7] : @bbox_lines_indexing
			lines = lix.collect { |i| corners[i] }
			view.line_width = 1
			view.line_stipple = ''
			view.drawing_color = 'gray'
			view.draw GL_LINES, lines				
		end
	end		
		
	#Draw the center of face
	center_face, normal = @center_face_info
	if center_face && !@no_inference && G6.check_screen_coords(view, center_face)
		size = view.pixels_to_model 6, center_face
		vecx, vecy, = normal.axes
		pts = [center_face.offset(vecx, -size), center_face.offset(vecx, size), center_face.offset(vecy, -size), center_face.offset(vecy, size)]
		view.line_width = 2
		view.line_stipple = ''
		view.drawing_color = 'blue'
		pts = pts.collect { |pt| view.screen_coords pt }	
		view.draw2d GL_LINES, pts unless pts.empty?	
	end
	
	#Draw the center of arc curves
	if @hsh_arc_center_info && !@no_inference
		@hsh_arc_center_info.each do |id, info|
			center, curve, info_comp = info
			next if center == center_face
			return unless center
			mark_draw view, [center, :center_arc_mark, info_comp]
		end	
	end
	
	#Draw the centers for Component or groups
	if @option_inference_comp && @hsh_comp_center_info && !@no_inference
		@hsh_comp_center_info.each do |id, info|
			center, comp, info_comp = info
			top_parent = info_comp[1]
			next if center == center_face
			return unless center
			code = (top_parent == comp) ? :center_comp_mark_top : :center_comp_mark
			mark_draw view, [center, code, info_comp]
		end	
	end		
end

#DRAW: draw the top axes for component
def draw_top_axes(view)
	return unless @top_axes && @top_parent && @top_parent.instance_of?(Sketchup::ComponentInstance)
	tr = @top_parent.transformation
	center = tr * @top_parent.definition.insertion_point
	center2d = view.screen_coords center
	center2d.z = 0
	size = view.pixels_to_model 50, center
	view.line_stipple = ''
	view.line_width = 3
	[[X_AXIS, 'red'], [Y_AXIS, 'green'], [Z_AXIS, 'blue']].each do |vec, color|
		pt = center.offset G6.transform_vector(vec, tr), size
		pt2d = view.screen_coords pt
		pt2d.z = 0
		view.drawing_color = color
		view.draw2d GL_LINE_STRIP, [center2d, pt2d]
	end
end

#DRAW: draw the top axes for component
def draw_protractor(view)
	return unless @protractor_enabled && @protractor_showable
	@protractor.draw view	

	#Computing the guide line in 2D for remoting
	ori2d = view.screen_coords @pt_origin
	pt = @pt_origin.offset @protractor_vector, 10
	pt2d = view.screen_coords pt
	pt2d.z = ori2d.z = 0
	vec2d = ori2d.vector_to pt2d
	pt1 = ori2d.offset vec2d, view.vpwidth * 2
	pt2 = ori2d.offset vec2d, -view.vpwidth * 2
	
	#Drawing the guide line
	[[X_AXIS, 'tomato'], [Y_AXIS, 'green'], [Z_AXIS, 'royalblue']].each do |vec, color|
		if @protractor_vector.parallel?(vec)
			view.line_width = 2
			view.line_stipple = '_'
			view.drawing_color = color
			view.draw2d GL_LINE_STRIP, [pt1, pt2]						
		end
	end	
	view.line_width = 1
	view.line_stipple = ''
	view.drawing_color = @hsh_color_vector[:protractor]
	view.draw2d GL_LINE_STRIP, [pt1, pt2]			
end

#DRAW: Draw the remote vertex inference if applicable
def draw_remote(view)
	return if @no_inference
	return unless remote_enabled? && @remote_origin
	ori2d = view.screen_coords @remote_origin
	pts = G6.pts_square ori2d.x, ori2d.y, 6
	view.drawing_color = @color_remote
	view.draw2d GL_POLYGON, pts
	view.line_width = 1
	view.line_stipple = ''
	view.drawing_color = @color_remote_fr
	view.draw2d GL_LINE_LOOP, pts
	
	return unless @remoting_dir

	#Computing the guide line in 2D for remoting
	pt = @remote_origin.offset @remoting_dir, 10
	pt2d = view.screen_coords pt
	vec2d = ori2d.vector_to pt2d
	pt1 = ori2d.offset vec2d, view.vpwidth * 2
	pt2 = ori2d.offset vec2d, -view.vpwidth * 2
	
	#Drawing the guide line
	view.line_width = 2
	view.line_stipple = '_'
	view.drawing_color = @color_remote_fr
	view.draw2d GL_LINE_STRIP, [pt1, pt2]	
	if [:follow_x, :follow_y, :follow_z].include?(@remoting_dir_symb)
		color = @hsh_color_vector[@remoting_dir_symb]
		view.line_stipple = ''
		view.line_width = 1
		view.drawing_color = color
		view.draw2d GL_LINE_STRIP, [pt1, pt2]	
	end	
end

#DRAW: Draw the parallel edge or cline inference if applicable
def draw_parallel(view)
	return if @no_inference
	return unless @parallel_info
	elt, comp, tr = @parallel_info
	
	view.line_width = 2
	view.drawing_color = @color_remote_fr
	if elt.instance_of?(Sketchup::Edge)
		pt1 = tr * elt.start.position
		pt2 = tr * elt.end.position
		view.line_stipple = ''
		view.draw GL_LINE_STRIP, [pt1, pt2].collect { |pt| G6.small_offset view, pt }
	elsif elt.instance_of?(Sketchup::ConstructionLine)
		pt1 = tr * elt.position
		pt2 = pt1.offset @parallel_vector, 100
		pt1_2d = view.screen_coords pt1
		pt2_2d = view.screen_coords pt2
		pt1_2d.z = pt2_2d.z = 0
		vec2d = pt1_2d.vector_to pt2_2d
		pt1 = pt1_2d.offset vec2d, view.vpwidth * 2
		pt2 = pt1_2d.offset vec2d, -view.vpwidth * 2
		view.line_stipple = '_'
		view.draw2d GL_LINE_STRIP, [pt1, pt2]
	else
		return
	end	
	
	return unless @parallel_dir

	#Computing the guide line in 2D for remoting
	ori2d = view.screen_coords @pt_origin
	ori2d.z = 0
	pt = @pt_origin.offset @parallel_dir, 10
	pt2d = view.screen_coords pt
	vec2d = ori2d.vector_to pt2d
	pt1 = ori2d.offset vec2d, view.vpwidth * 2
	pt2 = ori2d.offset vec2d, -view.vpwidth * 2
	
	#Drawing the guide line
	view.line_width = 2
	view.line_stipple = '_'
	view.drawing_color = @color_remote_fr
	view.draw2d GL_LINE_STRIP, [pt1, pt2]	
end

#DRAW: Draw the marks for twin information	
def draw_twin(view)
	return unless @twin_info
	cross2d = []
	lines3d = []
	@twin_info.each do |info|
		ptinter, pt1, pt2 = info
		ptinter2d = view.screen_coords ptinter
		ptinter2d.z = 0
		cross2d += G6.pts_cross ptinter2d.x, ptinter2d.y, 4	
		lines3d.push ptinter, pt1, ptinter, pt2	
	end
	color = @color_twin
	view.drawing_color = color
	view.line_width = 2
	view.line_stipple = '_'
	view.draw GL_LINES, lines3d
	view.line_stipple = ''
	view.draw2d GL_LINES, cross2d
end

#---------------------------------------------------------
# SELECTION: Manage selection and auto-selection
#---------------------------------------------------------

#SELECTION: Auto selection where there is no pre-selection	
def auto_selection(view, x, y, extended, multi_selection, no_vertex)
	return @preselection if @preselection
	
	#There is a top parent identified
	if @top_parent
		suselection_clear unless multi_selection || @cumulative_selection
		suselection_add @top_parent
		@active_selection = @selection.to_a
		preselection_cumulate if multi_selection
		return @active_selection
	end
	
	#Picking the elements
	@ph.do_pick x, y, @precision
	
	hsh_elt = {}
	lst_element = []
	edge = face = elt = cpoint = nil
	@ph.all_picked.each do |e|
		if e.instance_of?(Sketchup::Edge)
			unless hsh_elt[:edge]
				lst_element.push e
				hsh_elt[:edge] = e
				edge = e
			end	
		elsif e.instance_of?(Sketchup::Face)
			unless hsh_elt[:face]
				lst_element.push e
				hsh_elt[:face] = e
				face = e
			end	
		elsif e.instance_of?(Sketchup::ConstructionPoint)
			unless hsh_elt[:cpoint]
				lst_element.push e
				hsh_elt[:cpoint] = e
				cpoint = e
			end	
		elsif (e.instance_of?(Sketchup::Group) || e.instance_of?(Sketchup::ComponentInstance))
			unless hsh_elt[:comp]
				lst_element.push e
				hsh_elt[:comp] = e
			end	
		elsif !hsh_elt[:elt]
			if !(e.instance_of?(Sketchup::Drawingelement) && e.bounds.diagonal == 0)	#discard model axes
				lst_element.push e
				hsh_elt[:elt] = e
				elt = e
			end			
		end	
	end
	
	#Retaining the good element picked
	sel = lst_element.first
	
	#Special case when edge or face are at the surface of a component
	[cpoint, edge, face, elt].each do |e|
		if e && e.parent == Sketchup.active_model
			sel = e
			break
		end	
	end	
	
	#Vertex only
	@hi_curve_pts = nil
	if sel.instance_of?(Sketchup::Edge) && !no_vertex && !extended && !@cumulative_selection && !multi_selection
		edge = sel
		vx = @ip.vertex
		if vx
			curve = edge.curve
			unless curve && vx.curve_interior?
				lpts = []
				vx.edges.each do |ee|
					lvx = (ee.curve) ? ee.curve.vertices : [ee.start, ee.end]
					lpts.push.push lvx.collect { |v| v.position }
				end	
				@hi_curve_pts = lpts
				suselection_clear
				@active_selection = [vx]
				return @active_selection
			end	
		end	
	end	
	
	#Extending edges to curve
	if sel.instance_of?(Sketchup::Edge)
		if sel.curve
			sel = sel.curve.edges
		elsif extended
			anglemax = 40.degrees
			sel = [edge] + G6.curl_follow_extend(sel, anglemax, true)
		else
			sel = [sel]
		end	
	elsif sel.instance_of?(Sketchup::Face)
		sel = (@rendering_options["DrawHidden"] && !extended) ? [sel] : G6.face_neighbours(sel)
	elsif sel
		sel = [sel]
	end	
		
	#Finding centers of faces
	unless sel
		info = anchor_at_center_face(view, x, y)
		if info
			center, symb, info_face = info
			face, comp, tr = info_face
			if comp
				sel = [comp]
			elsif face
				sel = [face]
			end	
		end
	end

	#Finding centers of Component or Group
	unless sel
		info = anchor_at_center_comp(view, x, y)
		info = anchor_at_bbox_comp(view, x, y) unless info
		if info
			center, symb, info_comp = info
			comp, parent, tr = info_comp
			sel = (parent) ? [parent] : [comp]
		end
	end

	#Finding centers of arcs
	unless sel
		info = anchor_at_center_arc(view, x, y)
		if info
			center, symb, info_arc = info
			edge, comp, tr = info_arc
			if comp
				sel = [comp]
			elsif edge && edge.curve
				sel = edge.curve.edges
			end	
		end
	end
	
	#Replacement of additive mode depending on Shift	
	suselection_clear if !multi_selection
	suselection_add @cumulative_selection if @cumulative_selection
	
	#Updating the selection	
	if sel
		ledges = []
		sel.each do |face|
			face_edges_visible(face).each { |e| ledges.push e unless sel.include?(e) } if face.instance_of?(Sketchup::Face)
		end
		suselection_add sel + ledges
	end
	@active_selection = @selection.to_a
	@active_selection = nil if @active_selection.empty?
	
	@active_selection
end

#SUSELECTION: find the edges of a face when visible
def face_edges_visible(face)
	ledges = face.edges
	return ledges if @rendering_options["DrawHidden"]
	ledges.find_all { |e| e.faces.length == 1 || G6.edge_plain?(e) }
end
	
#SUSELECTION: Clear the Sketchup Selection
def suselection_clear
	return if @ignore_selection
	@selection.clear
end

#SUSELECTION: Manage the Sketchup Selection
def suselection_add(lentities, clear_before=false)
	return if @ignore_selection
	@selection.clear if clear_before
	return unless lentities
	lentities = [lentities] unless lentities.class == Array
	return unless lentities && lentities.length > 0
	
	#Adding the top-model elements to the selection
	lentities = lentities.find_all { |e| e.valid? }
	return if lentities.empty?
	@selection.add lentities
	
	#Computing the box inferences for groups and components
	ls_grouponents = lentities.grep(Sketchup::Group) + lentities.grep(Sketchup::ComponentInstance)	
	ls_grouponents.each do |comp|
		inference_bbox_grouponent(comp, nil, @tr_id)
	end

	@selection
end

#---------------------------------------------------------
# OPTIONS: Manage Vector and Planar options
#---------------------------------------------------------
	
#PARAMETER: Update a hash array with the parameters
def option_set_multi(hsh)
	@option_planar = hsh.fetch(:option_planar, @option_planar)
	@option_planar_code = hsh.fetch(:option_planar_code, @option_planar_code)
	@option_vector = hsh.fetch(:option_vector, @option_vector)
	@option_vector_code = hsh.fetch(:option_vector_code, @option_vector_code)
	@option_inference_comp = hsh.fetch(:option_inference_comp, @option_inference_comp)
	@option_inference_remote = hsh.fetch(:option_inference_remote, @option_inference_remote)
end
	
#PARAMETER: Update a hash array with the parameters
def option_update_hash(hsh)
	hsh[:option_planar] = @option_planar
	hsh[:option_planar_code] = @option_planar_code
	hsh[:option_vector] = @option_vector
	hsh[:option_vector_code] = @option_vector_code
	hsh[:option_inference_comp] = @option_inference_comp
	hsh[:option_inference_remote] = @option_inference_remote
end

#PARAMETER: Status for Plane and vector modes
def option_vector? ; @option_vector ; end
def option_planar? ; @option_planar ; end
	
#PARAMETER: Setting the direction parameters
def option_planar_toggle
	@option_planar = !@option_planar
	@option_vector = false if @option_planar
	@forced_model_vector = @forced_model_code = nil
	direction_changed
end

def option_vector_toggle
	@option_vector = !@option_vector
	@option_planar = false if @option_vector
	@forced_model_vector = @forced_model_code = nil
	direction_changed
end

def option_planar_set_dir(code)
	@option_planar = true
	@option_planar_code = code
	@option_vector = false
	@option_vector_code = :no
	@forced_model_vector = @forced_model_code = nil
	@direction_lock = false
	direction_changed
end

def option_vector_set_dir(code)
	@option_vector = true
	@option_vector_code = code
	@option_planar = false
	@option_planar_code = :no
	@forced_model_vector = @forced_model_code = nil
	@direction_lock = false
	direction_changed
end

#OPTION: Manage parameters from the arrow keys
def option_from_arrow(key, shift_down, ctrl_down)	
	#Resetting
	
	#Computing the new code
	case key
	when VK_UP
		code = :follow_z
	when VK_LEFT
		code = :follow_y
	when VK_RIGHT
		code = :follow_x
	when VK_DOWN
		code = :no
		if @direction_lock
			direction_toggle_lock
			return
		end
		return direction_cancel_all if @forced_model_vector
	end
	
	#Direction was locked
	if @direction_lock
		@direction_lock = !@direction_lock
		@forced_model_vector = @forced_model_code = nil
		decide_planar = @option_planar
	
	#Deciding if planar or vector mode should be set
	else
		if ctrl_down
			decide_planar = true
		elsif shift_down
			decide_planar = false
		elsif @option_planar
			decide_planar = (@option_planar_code != code) ? true : false		
		else
			decide_planar = (@option_vector_code != code) ? false : true		
		end
	end
	
	#Setting the direction and mode
	if decide_planar
		option_planar_set_dir code
	else
		option_vector_set_dir code
	end
end	

#OPTION: Toggle the flag for inference on components
def toggle_inference_comp
	@option_inference_comp = !@option_inference_comp
end

#----------------------------------------------------------------------------------------------------
# FILTER: Manage element filter
#----------------------------------------------------------------------------------------------------

#FILTER: set a filter for picking elements
def filter_set(lst_classes)
	lst_classes = [lst_classes] unless lst_classes.class == Array
	@filter = lst_classes
end
	
#FILTER: Check an element for filter
def filter_check(elt)
	return true unless @filter
	@filter.include?(elt.class)
end
		
#----------------------------------------------------------------------------------------------------
# DIRECTION: Manage directions
#----------------------------------------------------------------------------------------------------

#DIRECTION: Manage the change of direction
def direction_changed
	if @option_vector
		direction_compute_vector
	else
		direction_compute_plane
	end	
	protractor_find_plane_direction
	direction_implicit_plane_compute
	#remote_reset
end

#Unset the axes directions
def direction_no_axes
	@option_planar_code = @option_vector_code = :no
end
	
#DIRECTION: unset all directions	
def direction_cancel_all
	set_plane_direction nil
	set_vector_direction nil
	@forced_model_vector = @forced_model_code = nil
	@direction_lock = false
	direction_no_axes
	direction_changed
end

#DIRECTION: Pick the direction or plane from the underlying element
def direction_pick_from_model(bissector=false)
	if @mode == :origin
		info = @origin_info
	elsif @target_info
		info = @target_info
	end	
	if @option_vector
		vec, code = direction_vector_from_model(info)
		if @forced_model_vector && vec && vec.parallel?(@forced_model_vector) && code == @forced_model_code
			@option_vector = false
			@option_planar = true
		elsif !vec && @remoting_dir
			vec = @remoting_dir
			code = :remote
		end	
	else
		vec, code = direction_plane_from_model(info)
		if !option_implicit_plane_check(:bissector) && @forced_model_vector && vec && vec.parallel?(@forced_model_vector) && code == @forced_model_code
			@option_vector = true
			@option_planar = false
		else	
			return option_planar_set_dir(:follow_z) unless vec
		end	
	end
	@forced_model_vector, @forced_model_code = [vec, code]

	direction_changed
end

#DIRECTION: Compute the implicit plane from the origin
def direction_implicit_plane_compute
	@implicit_plane_normal = nil
	best_normal = (@origin_info) ? @origin_info[3] : nil
	if @forced_plane_dir
		@implicit_plane_normal = @forced_plane_dir
	elsif best_normal && option_implicit_plane_check(:best_normal)
		pt, type, info_comp = @origin_info
		elt, comp, tr = info_comp
		@implicit_plane_normal = best_normal
	else	
		@implicit_plane_normal, = direction_plane_from_model(@origin_info, true)
	end
	notify_caller :plane
end

#DIRECTION: Compute the vector based on specifications
def direction_compute_vector
	code = @option_vector_code
	vec = nil
	case code
	when :follow_x
		vec = X_AXIS
	when :follow_y
		vec = Y_AXIS
	when :follow_z
		vec = Z_AXIS
	end

	#Locking of direction is for vector
	shift_down = notify_caller :shift_down, false
	if @mode == :target && @pt_target && (@direction_lock || shift_down)
		vec = @pt_origin.vector_to @pt_target
		pt, code = @target_info
		direction_no_axes
		
	#Lock of direction	
	elsif @forced_model_vector			
		vec, code = [@forced_model_vector, @forced_model_code]
		direction_no_axes
	end

	#Setting the directions
	set_plane_direction nil
	set_vector_direction vec, code
end

#DIRECTION: Compute the plane based on specifications
def direction_compute_plane
	code = @option_planar_code
	vec = nil
	case code
	when :follow_x
		vec = X_AXIS
	when :follow_y
		vec = Y_AXIS
	when :follow_z
		vec = Z_AXIS	
	end
	
	#Locking of direction is for vector
	shift_down = notify_caller :shift_down, false
	if @mode == :target && (@direction_lock || shift_down)
		@direction_no_axes
		return direction_compute_vector
	end
	
	#Direction picked from model takes precedence
	if @pt_origin && @forced_model_vector	
		vec, code = [@forced_model_vector, @forced_model_code]
		direction_no_axes
	end	

	#Setting the direction
	set_vector_direction nil
	set_plane_direction vec, code
end

#DIRECTION: Toggle the lock of current direction picked on screen
def direction_toggle_lock
	@direction_lock = !@direction_lock
	direction_changed
end

#DIRECTION: Compute a planar direction from picking elements in the model	
def direction_pivot_compute(info)
	return direction_plane_from_model(info, true)
	#####pt, symb, info_elt = info
	elt, comp, tr = info_elt
	vec = nil
	
	case symb
	when :on_edge, :midpoint
		vec = (tr * elt.start.position).vector_to(tr * elt.end.position)
		vec = vec.reverse if elt.start.edges.length == 1
		code = :follow_edge
	when :on_cline, :end_cline
		vec = G6.transform_vector(elt.direction, tr)
		code = :follow_cline
	when :on_face, :center_face
		vec = G6.transform_vector(elt.normal, tr)
		code = :follow_normal	
	when :on_section_plane
		vec = Geom::Vector3d.new *(elt.get_plane[0..2])
		code = :follow_section_plane
	when :vertex
		vx = elt
		edges = vx.edges
		edges_plain = edges.find_all { |e| G6.edge_plain?(e) }
		if edges.length == 1
			edge = edges[0]
			code = :follow_edge
			vec = (tr * edge.start.position).vector_to(tr * edge.end.position)
			vec = vec.reverse if elt.start.edges.length == 1
		elsif edges_plain.length == 1	
			edge = edges_plain[0]
			code = :follow_edge
			vec = (tr * edge.start.position).vector_to(tr * edge.end.position)
			vec = vec.reverse if elt.start.edges.length == 1
		elsif edges_plain.length == 2
			edge1, edge2 = edges_plain
			vec1 = (tr * edge1.start.position).vector_to(tr * edge1.end.position)
			vec2 = (tr * edge2.start.position).vector_to(tr * edge2.end.position)
			vec = vec1 * vec2
			code = :plane_of_edges			
		end
		
	end
	
	return nil unless vec && vec.valid?
	
	code = direction_code_from_vector(vec, code)
	(vec) ? [vec, code] : nil
end

#DIRECTION: Compute the direction for pivoting at origin
def direction_pivot_origin(info=nil)
	return nil unless @pt_origin
	@pivot_vector_origin
end

#DIRECTION: Compute the direction for pivoting at target
def direction_pivot_target
	return nil unless @pt_origin
	return [@remoting_dir, :remote] if @remoting_dir
	@pivot_vector_target
end

#---------------------------------------------------------
# PALETTE: Palette Contribution for axes
#---------------------------------------------------------
	
#PALETTE: Buttons for Plane Direction
def palette_contribution_planar(palette, hsh=nil)
	symb_master = :pal_planar	
	
	bk_color = @color_but_title
	passive_title = false
	if hsh.class == Hash
		passive_title = hsh.fetch :passive_title, true
		bk_color = hsh.fetch :bk_color, bk_color
	end	
	
	#Creating the title button
	hsh_sepa = { :passive => true, :draw_proc => :separator_V, :width => 8  }
	hgt = 16
	title = T6[:T_TXT_Plane]
	val_proc_title = proc { @option_planar }
	hsh = { :type => 'multi_free', :passive => passive_title, :text => title, :tooltip => T6[:TIP_Planar_Title], :height => hgt,
			:bk_color => bk_color, :hi_color => @color_but_hi, :value_proc => val_proc_title }
	palette.declare_button(symb_master, hsh) { option_planar_toggle }
		
	#Creating the buttons
	hshp = { :parent => symb_master, :height => hgt, :width => 24, :hi_color => 'white', :draw_proc => @draw_local}
	
	value_proc = proc { @option_planar_code == :no }
	hsh = { :text => 'N', :tooltip => @tip_plane_no, :bk_color => 'silver', :value_proc => value_proc  }
	palette.declare_button("#{symb_master}__no".intern, hshp, hsh) { option_planar_set_dir :no }
	
	value_proc = proc { @option_planar_code == :follow_x }
	hsh = { :text => 'X', :tooltip => @tip_plane_x, :bk_color => 'tomato', :value_proc => value_proc  }
	palette.declare_button("#{symb_master}__x".intern, hshp, hsh) { option_planar_set_dir :follow_x }

	value_proc = proc { @option_planar_code == :follow_y }
	hsh = { :text => 'Y', :tooltip => @tip_plane_y, :bk_color => 'limegreen', :value_proc => value_proc  }
	palette.declare_button("#{symb_master}__y".intern, hshp, hsh) { option_planar_set_dir :follow_y }

	value_proc = proc { @option_planar_code == :follow_z }
	hsh = { :text => 'Z', :tooltip => @tip_plane_z, :bk_color => 'skyblue', :value_proc => value_proc  }
	palette.declare_button("#{symb_master}__z".intern, hshp, hsh) { option_planar_set_dir :follow_z }	
end

#PALETTE: Buttons for Vector Direction
def palette_contribution_vector(palette)
	symb_master = :pal_vector
	
	#Creating the title button
	hsh_sepa = { :passive => true, :draw_proc => :separator_V, :width => 8  }
	hgt = 16
	title = T6[:T_TXT_Vector]
	val_proc_title = proc { @option_vector }
	hsh = { :type => 'multi_free', :text => title, :tooltip => T6[:TIP_Vector_Title], :height => hgt,
			:bk_color => @color_but_title, :hi_color => @color_but_hi, :value_proc => val_proc_title }
	palette.declare_button(symb_master, hsh) { option_vector_toggle }

	#Creating the buttons
	hshp = { :parent => symb_master, :height => hgt, :width => 24, :hi_color => 'white' }
	
	value_proc = proc { @option_vector_code == :no }
	hsh = { :text => 'N', :tooltip => @tip_vector_no, :bk_color => 'silver', :value_proc => value_proc  }
	palette.declare_button("#{symb_master}__no".intern, hshp, hsh) { option_vector_set_dir :no }

	value_proc = proc { @option_vector_code == :follow_x }
	hsh = { :text => 'X', :tooltip => @tip_vector_x, :bk_color => 'tomato', :value_proc => value_proc  }
	palette.declare_button("#{symb_master}__x".intern, hshp, hsh) { option_vector_set_dir :follow_x }

	value_proc = proc { @option_vector_code == :follow_y }
	hsh = { :text => 'Y', :tooltip => @tip_vector_y, :bk_color => 'limegreen', :value_proc => value_proc  }
	palette.declare_button("#{symb_master}__y".intern, hshp, hsh) { option_vector_set_dir :follow_y }

	value_proc = proc { @option_vector_code == :follow_z }
	hsh = { :text => 'Z', :tooltip => @tip_vector_z, :bk_color => 'skyblue', :value_proc => value_proc  }
	palette.declare_button("#{symb_master}__z".intern, hshp, hsh) { option_vector_set_dir :follow_z }	
end

#PALETTE: Contribution button
def palette_contribute_button(code, palette, symb_master, hshp)
	draw_local = self.method "draw_button_opengl"
	@palette = palette

	case code
	
	#Remove all inference marks
	when :reset_all_marks
		symb = "#{symb_master}_#{code}".intern
		hsh = { :tooltip => @tip_reset_inference, :draw_proc => draw_local }
		palette.declare_button(symb, hshp, hsh) { inference_reset }

	#Toggle hidden geometry	
	when :toggle_show_hidden	
		symb = "#{symb_master}_#{code}".intern
		value_proc = proc { @rendering_options["DrawHidden"] }
		hsh = { :value_proc => value_proc, :draw_proc => :std_draw_hidden, :tooltip => T6[:T_TIP_DrawHidden] }
		palette.declare_button(symb, hshp, hsh) { @rendering_options["DrawHidden"] = !@rendering_options["DrawHidden"] }

	#Inferences on Components
	when :inference_on_comp
		symb = "#{symb_master}_#{code}".intern
		value_proc = proc { @option_inference_comp }
		hsh = { :value_proc => value_proc, :draw_proc => draw_local, :tooltip => T6[:TIP_ToggleInferenceComp] }
		palette.declare_button(symb, hshp, hsh) { toggle_inference_comp }

	#Remote Inferences
	when :inference_remote
		symb = "#{symb_master}_#{code}".intern
		value_proc = proc { @option_inference_remote }
		hsh = { :value_proc => value_proc, :tooltip => T6[:TIP_ToggleInferenceRemote], :draw_proc => draw_local }
		palette.declare_button(symb, hshp, hsh) { remote_toggle }
			
	#Remote Inferences
	when :no_inference
		symb = "#{symb_master}_#{code}".intern
		value_proc = proc { !@no_inference }
		hsh = { :value_proc => value_proc, :tooltip => @tip_no_inference, :draw_proc => draw_local }
		palette.declare_button(symb, hshp, hsh) { no_inference_toggle }
			
	end	
end

#PALETTE: Custom drawing of buttons
def draw_button_opengl(symb, dx, dy, main_color, frame_color, selected)
	code = symb.to_s
	lst_gl = []
	dx2 = dx / 2
	dy2 = dy / 2
	grayed = @palette.button_is_grayed?(symb)
	color = (grayed) ? 'gray' : frame_color
	
	case code

	when /inference_remote/
		lpti = [[1, 0], [dx-1, dy]]
		pts = lpti.collect { |x, y| Geom::Point3d.new(x, y, 0) }
		lst_gl.push [GL_LINES, pts, 'gray', 4, '']	
		lst_gl.push [GL_LINES, pts, @color_remote_fr, 3, '-']	
		pts = G6.pts_square dx-5, 3, 3
		lst_gl.push [GL_POLYGON, pts, @color_remote_fr]	
		lst_gl.push [GL_LINE_LOOP, pts, 'black', 1, '']			
	
	when /inference_on_comp/
		pts = G6.pts_rectangle 1, 1, dx-1, dy-1
		lst_gl.push [GL_LINE_LOOP, pts, 'gray', 1, '']	
		t = Geom::Transformation.translation Geom::Vector3d.new(dx2, dy2, 0)
		@hsh_marks[:center_comp_mark].each do |ll|
			ll2 = ll.clone
			ll2[1] = ll[1].collect { |pt| t * pt }
			lst_gl.push ll2
		end	

	when /reset_all_marks/
		t = Geom::Transformation.translation Geom::Vector3d.new(dx2/2, dy2, 0)
		@hsh_marks[:center_face].each do |ll|
			ll2 = ll.clone
			ll2[1] = ll[1].collect { |pt| t * pt }
			lst_gl.push ll2
		end	
		t = Geom::Transformation.translation Geom::Vector3d.new(dx2 + dx2/2, dy2, 0)
		@hsh_marks[:bbox_corner].each do |ll|
			ll2 = ll.clone
			ll2[1] = ll[1].collect { |pt| t * pt }
			lst_gl.push ll2
		end	
		lpti = [[0, 0], [dx, dy], [0, dy], [dx, 0]]
		pts = lpti.collect { |x, y| Geom::Point3d.new(x, y, 0) }
		lst_gl.push [GL_LINES, pts, 'red', 2, '-']	
	
	when /no_inference/i
		color = 'lightgrey'
		frcolor = 'gray'
		xc = dx-dy2-1
		yc = dy2
		n = 12
		dec = 4
		pts1 = G6.pts_circle(xc, yc, dy2, n)[0..6]
		pts2 = G6.pts_circle(xc, yc, dy2-dec, n)[0..6]
		quads = []
		for i in 0..pts1.length-2
			quads.push pts1[i], pts1[i+1], pts2[i+1], pts2[i]
		end
		lst_gl.push [GL_QUADS, quads, color]	
		lst_gl.push [GL_LINE_STRIP, pts1, frcolor, 1, '']	
		lst_gl.push [GL_LINE_STRIP, pts2, frcolor, 1, '']	
		
		[0, -1].each do |i|
			pt1 = pts1[i]
			pt2 = pts2[i]
			len = -dx2+dec
			pts = [pt1, pt1.offset(X_AXIS, len), pt2.offset(X_AXIS, len), pt2]
			lst_gl.push [GL_POLYGON, pts, color]
			lst_gl.push [GL_LINE_STRIP, pts[0..3], frcolor, 1, '']
			pts = [pts[1], pts[1].offset(X_AXIS, -dec), pts[2].offset(X_AXIS, -dec), pts[2]]
			lst_gl.push [GL_POLYGON, pts, 'red']
			lst_gl.push [GL_LINE_STRIP, pts[0..3], frcolor, 1, '']
		end
		
	end	#case code
	
	lst_gl
end

#---------------------------------------------------------
# MENU: Contextual Contribution for axes
#---------------------------------------------------------
		
#MENU: Contextual menu
def menu_contribution(cxmenu)	
	#Reset inferences
	cxmenu.add_sepa
	cxmenu.add_item(@tip_reset_inference) { inference_reset }
	cxmenu.add_sepa
	cxmenu.add_item(@tip_no_inference) { no_inference_toggle }

	#Privileged directions
	cxmenu.add_sepa
	if @option_planar
		ls = [[:no, @tip_plane_no], [:follow_x, @tip_plane_x], [:follow_y, @tip_plane_y], [:follow_z, @tip_plane_z]]
	elsif @option_vector
		ls = [[:no, @tip_vector_no], [:follow_x, @tip_vector_x], [:follow_y, @tip_vector_y], [:follow_z, @tip_vector_z]]
	end	
	ls.each { |code, tip| menu_direction_add(cxmenu, code, tip) }
	
	#Pick from model
	if @mode == :origin && @origin_info
		cxmenu.add_sepa
		tip = (@option_planar) ? @tip_pick_plane : @tip_pick_vector
		cxmenu.add_item(tip) { direction_pick_from_model }
	end	
end

#MENU: Add the menu item for forced direction
def menu_direction_add(cxmenu, code, tip)
	if @option_vector
		cxmenu.add_item(tip) { option_vector_set_dir code } unless code == @option_vector_code
	else	
		cxmenu.add_item(tip) { option_planar_set_dir code } unless code == @option_planar_code
	end	
end
	
end	#class OriginTargetPicker

end	#End Module Traductor
