=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# © Copyright May 2014
# Designed and developped May 2014 by Fredo6
#
# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for commercial purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:   Lib6StencilIntersector.rb
# Original Date	:   18 May 2014
# Description	:   Manage Intersections for stencils
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end

module Traductor

#=============================================================================================
#=============================================================================================
# Class Intersector: Manage intersections of pseudo solid and stencils
#=============================================================================================
#=============================================================================================

class StencilIntersector

#Structure describing a tube with all associated info
Tubage = Struct.new :pshape, :tube_group, :bounds, :tr_entities, :invalid

#Structure describing a pseudo solid in the model to be intersected with
PseudoSolid = Struct.new :type, :comp, :tr, :entities, :lst_ents, :mirrored

#Structure describing a tube constructed for intersection
PseudoShape = Struct.new :stencil_shape, :loops_info, :tr_axe, :anchor, :bbox, :ignored

#INIT: Class instance initialization
def initialize__(*hargs)
	#Initialization
	@model = Sketchup.active_model
	@selection = @model.selection
	@top_entities = @model.active_entities
	@view = @model.active_view
	@tr_id = Geom::Transformation.new
	
	#Initialization of attribute for edges and faces
	@dico_hole = "LibFredo6_Intersector6"
	@attr_tube = "tube"
	@attr_shape_min = 'shape_min'
	@attr_shape_max = 'shape_max'
	@attr_line = 'line'
	@attr_quad = 'quad'
	@attr_solid = 'solid'

	#Parsing the arguments
	hargs.each do |harg|	
		harg.each { |key, value|  parse_args(key, value) } if harg.class == Hash
	end
end

#INIT: Parse the parameters of the Face Picker
def parse_args(key, value)
	skey = key.to_s
	#case skey

	#end	
end	

#--------------------------------------------------------------
# TOP: Top execution
#--------------------------------------------------------------

#TOP: Top execution method
def top_execute(selection_info, *hargs)
	begin
		status = top_execute_protected(selection_info, *hargs)
	rescue Exception => e
		status = e
	end	
	@status_execution = status unless @suops
	status
end

#TOP: Get the status of execution (useful for asynchronous mode)
def get_status_execution
	@status_execution
end
	
#TOP: Top execution in error-protected mode	
def top_execute_protected(selection_info, *hargs)
	@selection_info = selection_info
	
	#Parsing the parameters
	hargs.each do |harg|	
		harg.each { |key, value|  parse_args_execute(key, value) } if harg.class == Hash
	end
	@cosinus_smooth = (@deviation_smooth) ? Math.cos(@deviation_smooth) : nil
	@cosinus_stop = (@deviation_stop) ? Math.cos(@deviation_stop) : nil
		
	#Synchronous or asynchronous execution
	hsh_vbar = { :delay => 3 }
	@status_execution = false
	case @cut_style
	when :thru
		nticks = 4 * @selection_info.length + 1
		status = (@suops) ? @suops.launch_execution(nticks, nil, hsh_vbar) { thru_intersect_robot} : thru_intersect_synchronous
	else
		nticks = 5
		status = (@suops) ? @suops.launch_execution(nticks, nil, hsh_vbar) { stamp_intersect_robot} : stamp_intersect_synchronous
	end
	status
end

#TOP: Compute the solids, shapes and positioning of shapes on the surface
def execution_prepare
	#Registering the selection into pseudo solids
	@lst_solids = []
	@out_group = nil
	@selection_info.each { |comp, tr| solid_register(comp, tr) }
	return false if @lst_solids.empty?
		
	#Preparing the solids
	solid_prepare
		
	#Getting information on the stencil	and computing the anchors
	stencil_pshape_prepare
end

#TOP: Last step of execution: restore top-level geometry from groups
def finish_execute
	solid_restore_top_level
	@status_execution
end

#TOP: Parse the arguments for execution
def parse_args_execute(key, value)
	case key.to_s
	when /repere_tr/i
		@repere_tr = value
	when /stencil/i
		@stencil = value
	when /thru_nb_tube/i
		@thru_nb_tube = value
	when /thru_no_face/i
		@thru_no_face = value
	when /thru_mark/i
		@thru_mark = value
	when /stamp_punch/i
		@stamp_punch = value
	when /stamp_mark/i
		@stamp_mark = value
	when /carve_flat/i
		@carve_flat = value
	when /carve_plan/i
		@carve_plan = value
	when /cut_in_group/i
		@cut_in_group = value
	when /cut_out_group/i
		@cut_out_group = value
	when /cut_style/i
		@cut_style = value
	when /carve_offset/
		@carve_offset = value
	when /surfloc/
		@surfloc = value
	when /custom_material/
		@custom_material = value
	when /deviation_stop/
		@deviation_stop = value
	when /deviation_smooth/
		@deviation_smooth = value
	when /suops/
		@suops = value
	end	
end	

#--------------------------------------------------------------
# STENCIL: Management of pseudo shape in the stencil
#--------------------------------------------------------------

#STENCIL: Prepare the stencil
def stencil_pshape_prepare
	@lst_pshapes = []
	@stencil.pseudoshape_list.each do |stencil_shape|
		pshape = PseudoShape.new
		pshape.stencil_shape = stencil_shape
		pshape.anchor = @stencil.pseudoshape_get_anchor stencil_shape
		pshape.loops_info = @stencil.pseudoshape_get_loops_info(stencil_shape)
		pts, = pshape.loops_info[0]
		bbox = pshape.bbox = Geom::BoundingBox.new
		bbox.add pts
		@lst_pshapes.push pshape
	end	
	
	#Compute the axes for each shape
	if @surfloc
		return nil unless stencil_compute_axes
	else
		origin = @repere_tr * ORIGIN
		axes = [X_AXIS, Y_AXIS, Z_AXIS].collect { |v| @repere_tr * v }
		@lst_pshapes.each do |pshape| 
			pshape.ignored = true unless stencil_compute_axes_proj(pshape, origin, axes)
		end	
	end
	
	#Keeping only the relevant pshapes
	@lst_pshapes = @lst_pshapes.find_all { |pshape| !pshape.ignored }
	
	#Everything is OK
	(@lst_pshapes.length > 0)
end

#STENCIL: Compute the axes for each individual shape on the surface to be cut (projection)
def stencil_compute_axes_proj(pshape, origin, axes)
	#Compute the direction for the anchor in pshape coordinates
	anchor = pshape.anchor
	vec_anchor = ORIGIN.vector_to anchor
	unless vec_anchor.valid?
		pshape.tr_axe = Geom::Transformation.axes origin, *axes
		return true
	end		
	d_anchor = vec_anchor.length
	angle = X_AXIS.angle_between vec_anchor
	angle = -angle if angle != 0 && (X_AXIS * vec_anchor) % Z_AXIS < 0
	
	#Compute the transformed axes
	vecx, vecy, vecz = axes
	trot = Geom::Transformation.rotation origin, vecz, angle
	new_vecx = trot * vecx
	new_vecy = trot * vecy
	new_axes = [new_vecx, new_vecy, vecz]
	
	group_contour = surfloc_master_contour(origin, new_axes)
	entities = group_contour.entities

	#Compute the section along X (i.e. Y axis the normal)
	vec0 = new_axes[0]
	plane = [origin, vec0]
	dtarget_plus = dtarget_minus = 0
	ibeg, pts = surfloc_compute_contour(origin, vec0, group_contour, d_anchor, d_anchor, plane)
	
	#Erasing the group contour
	group_contour.erase!
	
	#Pshape is out of range
	return false unless ibeg

	#Compute the target point on surface
	plane = [origin.offset(vec0, d_anchor), vec0]
	target = surfloc_transfo_on_contour_proj(origin, ibeg, pts, d_anchor, plane)	
	return false unless target
	
	#Compute the final axes transformation
	tt_anchor = Geom::Transformation.translation anchor.vector_to(ORIGIN)
	tr_axe = Geom::Transformation.axes target, vecx, vecy, vecz
	pshape.tr_axe = tr_axe * tt_anchor
	
	#Everything is OK
	true
end

#STENCIL: Compute the axes for each individual shape on the surface to be cut (general)
def stencil_compute_axes
	origin = @repere_tr * ORIGIN
	axes = [X_AXIS, Y_AXIS, Z_AXIS].collect { |v| @repere_tr * v }
	
	#Creating the section contour in X
	group_contour = surfloc_master_contour(origin, axes)
	entities = group_contour.entities

	#Compute the section along X (i.e. Y axis the normal)
	vec0 = axes[0]
	plane = (@surfloc) ? nil : [origin, vec0]
	dtarget_plus = 1
	dtarget_minus = -1
	@lst_pshapes.each do |pshape|
		d = pshape.anchor.x
		dtarget_plus = d if d >= 0 && d > dtarget_plus		
		dtarget_minus = d if d < 0 && d < dtarget_minus	
	end
	ibeg, pts = surfloc_compute_contour(origin, vec0, group_contour, dtarget_plus, dtarget_minus.abs, plane)
	return nil unless ibeg
	pts.each { |pt| entities.add_cpoint pt }
	
	#Computing the transformation for X and Z axis for each pshape
	vecx, vecy, vecz = axes
	@lst_pshapes.each do |pshape|
		anchor = pshape.anchor
		d = anchor.x
		if @surfloc
			target, vec = surfloc_transfo_on_contour_surf(origin, ibeg, pts, d)	
			unless target
				pshape.ignored = true
				next
			end	
			angle = vecx.angle_between vec
			angle = -angle if angle != 0 && (vecx * vec) % vecy < 0
			trot = Geom::Transformation.rotation target, vecy, angle
			new_vecz = trot * vecz
			new_vecx = vec
			new_vecy = vecy
		else
			plane = [origin.offset(vecx, d), vecx]
			target = surfloc_transfo_on_contour_proj(origin, ibeg, pts, d, plane)	
			unless target
				pshape.ignored = true
				next
			end	
			new_vecx = vecx
			new_vecz = vecz
		end
		entities.add_cline target, new_vecz
		tt_anchor = Geom::Transformation.translation Geom::Vector3d.new(-d, 0, 0)
		tr_axe = Geom::Transformation.axes target, new_vecx, vecy, new_vecz
		pshape.tr_axe = tr_axe * tt_anchor
	end
	
	#Erasing the group contour
	group_contour.erase!
	
	#Computing the transformation in Y
	@lst_pshapes.each do |pshape|
		next if pshape.ignored
		pshape.ignored = true unless stencil_compute_axes_in_Y pshape
	end	
	
	#Everything is OK
	true
end

#SURFLOC: Compute the origin and transformation following Y direction
def stencil_compute_axes_in_Y(pshape)
	#Coordinates and axes for the anchor in natural and in real model
	tr_axe = pshape.tr_axe
	anchor = pshape.anchor
	origin = tr_axe * Geom::Point3d.new(anchor.x, 0, 0)
	vecx, vecy, vecz = [X_AXIS, Y_AXIS, Z_AXIS].collect { |v| tr_axe * v }
	
	#Creating the section contour in Y
	axes = [vecy, vecx.reverse, vecz]
	group_contour = surfloc_master_contour(origin, axes)
	entities = group_contour.entities

	#Computing the contours
	vec0 = axes[0]
	plane = (@surfloc) ? nil : [origin, vec0]
	d = anchor.y
	ibeg, pts = surfloc_compute_contour(origin, vec0, group_contour, d.abs, d.abs, plane)
	
	#Erasing the group contour
	group_contour.erase!
	
	#Shape anchor is out of range
	return nil unless ibeg

	#Finding the new orientation on surface
	if @surfloc
		target, vec = surfloc_transfo_on_contour_surf(origin, ibeg, pts, d)	
		return nil unless target
		angle = vecy.angle_between vec
		angle = -angle if angle != 0 && (vecy * vec) % vecx < 0
		trot = Geom::Transformation.rotation target, vecx, angle
		new_vecz = trot * vecz
		new_vecy = vec
		new_vecx = vecx
	else
		plane = [origin.offset(vecy, d), vecy]
		target = surfloc_transfo_on_contour_proj(origin, ibeg, pts, d, plane)	
		return nil unless target
		new_vecx = vecx
		new_vecy = vecy
		new_vecz = vecz
	end

	#Computing the axes transformation
	tt_anchor = Geom::Transformation.translation Geom::Vector3d.new(-anchor.x, -anchor.y, 0)
	tr_axe = Geom::Transformation.axes target, new_vecx, new_vecy, new_vecz
	pshape.tr_axe = tr_axe * tt_anchor #* pshape.tr_axe	
	
	#Everything is OK
	true
end

#--------------------------------------------------------------
# SURFLOC: Surface locator
#--------------------------------------------------------------

#SURFLOC: Compute the local transformation (Projection mode)
def surfloc_transfo_on_contour_proj(origin, ibeg, pts, dtarget, plane)	
	#Locating the starting segment
	n = pts.length - 1
	ptinter = nil
			
	#Toward right
	if dtarget >= 0
		for i in ibeg..n-1
			line = [pts[i], pts[i+1]]
			pt = Geom.intersect_line_plane line, plane
			next unless pt
			ptinter = pt
			return ptinter if G6.point_within_segment?(ptinter, *line)
		end
		
	#Toward left
	else
		ibeg = ibeg+1 unless origin == pts[ibeg]
		for i in 0..ibeg-1
			j = ibeg - i
			line = [pts[j], pts[j-1]]
			pt = Geom.intersect_line_plane line, plane
			next unless pt
			ptinter = pt
			return ptinter if ptinter && G6.point_within_segment?(ptinter, *line)
		end
	end
	
	#No point found
	nil
end

#SURFLOC: Compute the local transformation (Follow Surface mode)
def surfloc_transfo_on_contour_surf(origin, ibeg, pts, dtarget)	
	#Locating the starting segment
	n = pts.length - 1
	found = false
			
	#Toward right
	if dtarget >= 0
		iseg1 = ibeg
		target = pts[ibeg]
		dtot = -origin.distance(target)
		for i in ibeg+1..n
			dtot += target.distance pts[i]
			iseg1 = i - 1
			if dtot >= dtarget
				found = true
				break
			end	
			target = pts[i]
		end
		
	#Toward left
	else
		iseg1 = 0
		ibeg = ibeg-1 if origin == pts[ibeg]
		target = pts[ibeg+1]
		dtot = -origin.distance(target)
		for i in 0..ibeg
			j = ibeg - i
			dtot += target.distance pts[j]
			iseg1 = j
			if dtot >= -dtarget
				found = true
				break
			end	
			target = pts[j]
		end
	end
	
	#No target found
	return nil unless found
	
	#Computing the segment beg and end points
	iseg2 = iseg1 + 1
	pt1 = pts[iseg1]
	pt2 = pts[iseg2]
	d = pt1.distance(pt2)
	r = (dtot - dtarget.abs) / d
	r = 1-r if dtarget < 0
	target = Geom.linear_combination r, pt1, 1-r, pt2
	vec = pt1.vector_to pt2
	
	#Calculating the tangent
	if @cosinus_smooth
		dev = nil
		if r < 0.5 && pts[iseg2+1]
			dev = surfloc_deviation_smooth pt1, pt2, pts[iseg2+1]
		elsif r > 0.5 && iseg1 > 0
			dev = surfloc_deviation_smooth pts[iseg1-1], pt1, pt2
		end
		if dev
			r = (1 - 2 * r).abs
			vec = Geom.linear_combination r, dev, 1-r, vec
		end	
	end
	
	#Returning the target point and tangent vector
	[target, vec]
end

#SURFLOC: Compute the mean vector in smooth situations
def surfloc_deviation_smooth(pt1, ptc, pt2)	
	vec1 = pt1.vector_to(ptc).normalize
	vec2 = ptc.vector_to(pt2).normalize
	ps = vec1 % vec2
	return nil unless ps >= @cosinus_smooth
	vec1 + vec2
end

#SURFLOC: Compute the bi-directional contour from the origin
def surfloc_compute_contour(origin, vec0, group_contour, dtarget_plus, dtarget_minus, plane)
	entities = group_contour.entities
	ledges = entities.grep(Sketchup::Edge)
	dtarget_plus *= 1.1
	dtarget_minus *= 1.1
	
	#Finding the starting edge
	edge0 = ledges.find { |e| G6.point_within_segment?(origin, e.start.position, e.end.position) }
	
	#Trying with proximity if no exact solution
	unless edge0
		dmin = nil
		ledges.each do |e|
			d, = G6.proximity_point_segment(origin, e.start.position, e.end.position)
			if !dmin || d < dmin
				edge0 = e
				dmin = d
			end	
		end	
	end	
	return [] unless edge0
	
	#Finding the bounding vertices and most appropriate starting edge
	vx = [edge0.start, edge0.end].find { |vx| vx.position == origin }
	vedges = (vx) ? vx.edges : [edge0]
	psmax = nil
	vx_left = nil
	vedges.each do |e|
		vec = e.start.position.vector_to e.end.position
		ps = vec0 % vec
		if !psmax || ps.abs > psmax
			psmax = ps.abs
			vx_left = (ps < 0) ? e.end : e.start
			edge0 = e
		end
	end
	vx_right = edge0.other_vertex(vx_left)
	vec = vx_left.position.vector_to vx_right.position
	
	#Getting the contour on the right and left
	pts_right = surfloc_pursue_contour(origin, edge0, vx_right, vec, dtarget_plus, 'green')
	pts_left = surfloc_pursue_contour(origin, edge0, vx_left, vec.reverse, dtarget_minus, 'lightblue')
	
	#Returning the contour and position of origin
	ibeg = pts_left.length - 1
	pts = pts_left.reverse + pts_right
	[ibeg, pts]
end

#SURFLOC: Pursue the contour in the direction up to the specified distance, along the surface
def surfloc_pursue_contour(origin, edge0, vx0, vec, dtarget, color)
	dtot = origin.distance(vx0.position)
	edge_ini = edge0
	pts = [vx0.position]
	plane = (@surfloc) ? nil : [origin, vec]
	
	finished = 0
	n = 0
	while true
		break if n > 100
		n += 1
		edge0 = surfloc_pursue_next_edge(edge0, vx0)
		if edge0 && edge0 != edge_ini
			vx0 = edge0.other_vertex(vx0)
			d = origin.distance vx0.position
			vec = origin.vector_to vx0.position
			origin = vx0.position
			pts.push origin
			break if plane && origin.distance_to_plane(plane) > dtarget
			break if finished > 0
			finished += 1 if dtarget && dtot > dtarget
			dtot += d
		else
			break			
		end
	end
	
	pts
end
				
#SURFLOC: Compute the next edge in the specified edge from vertex vx
def surfloc_pursue_next_edge(edge, vx)
	edges = vx.edges.find_all { |e| e != edge}
	return nil if edges.length == 0
	
	#Computing the deviation
	new_edge = nil
	psmax = @cosinus_stop
	edges.each do |e|
		vx1 = edge.other_vertex vx
		vx3 = e.other_vertex vx
		vec1 = vx1.position.vector_to(vx.position).normalize
		vec2 = vx.position.vector_to(vx3.position).normalize
		ps = vec1 % vec2
		if !psmax || ps > psmax
			psmax = ps
			new_edge = e
		end	
	end
	new_edge
end
	
#SURFLOC: Construct the master contour by intersection of a plane and the solids	
def surfloc_master_contour(origin, axes)	
	#Estimating the plane size
	size = 0
	bbtot = Geom::BoundingBox.new
	@lst_solids.each do |psolid|
		comp = psolid.comp
		tr = psolid.tr * comp.transformation.inverse
		bb = comp.bounds
		bbtot.add tr * bb.min
		bbtot.add tr * bb.max
	end	
	size = 2 * bbtot.diagonal
	
	#Creating the fake plane for intersection
	vecx, vecy, vecz = axes
	group_plane = @top_entities.add_group
	gent = group_plane.entities
	pt1 = origin.offset(vecx, -0.5 * size).offset(vecz, 0.5 * size)
	pt2 = pt1.offset vecx, size
	pt3 = pt2.offset vecz, -size
	pt4 = pt3.offset vecx, -size
	
	face = gent.add_face [pt1, pt2, pt3, pt4]
	
	#Intersection of the plane with each solid put in a group
	group_contour = @top_entities.add_group
	grent = group_contour.entities
	all_ledges = []
	@lst_solids.each do |psolid|
		tr = psolid.tr
		entities = psolid.entities
		solid_id = psolid.comp.entityID
		edges = entities.intersect_with false, tr, grent, @tr_id, false, group_plane
		edges.each { |e| tube_mark_solid(e, solid_id) }
		all_ledges += edges
	end

	#Deleting the plane
	group_plane.erase!
	
	#Returning the contour
	group_contour
end
	
#--------------------------------------------------------------
# SOLID: Management of pseudo solids
#--------------------------------------------------------------

#SOLID: Register a solid	
def solid_register(comp, tr)
	psolid = PseudoSolid.new
	if comp.class == Array
		psolid.comp = nil
		psolid.entities = @top_entities
		psolid.lst_ents = comp
		psolid.type = :flat
	else
		comp.make_unique
		psolid.comp = comp
		psolid.entities = G6.grouponent_entities comp
		psolid.type = :comp
	end
	psolid.tr = tr
	psolid.mirrored = G6.transformation_is_mirrored?(tr)
	
	@lst_solids.push psolid
	psolid	
end

#SOLID: Prepare the solids - Creating a fake group for entities at top level
def solid_prepare
	@lst_solids.each do |psolid|
		if psolid.type == :flat
			psolid.comp = @top_entities.add_group psolid.lst_ents
			psolid.tr = psolid.comp.transformation
			psolid.entities = psolid.comp.entities
		end
	end
end

#SOLID: Exploding back the pseudo-group for entities at top level
def solid_restore_top_level
	@lst_solids.each { |psolid| psolid.comp.explode if psolid.type == :flat }
end

#--------------------------------------------------------------
# TUBAGE: Management of tube information
#--------------------------------------------------------------

#TUBAGE: Create a Tubage structure with base parameters
def tubage_create(pshape, bounds, tr_entities)
	tubage = Tubage.new
	tubage.pshape = pshape
	tubage.bounds = bounds
	tubage.tr_entities = tr_entities
	tubage
end

#--------------------------------------------------------------
# TUBE: Management of Tubes for intersection
#--------------------------------------------------------------

#TUBE: Methods to mark tube elements
def tube_mark(e, val) ; e.set_attribute @dico_hole, @attr_tube, val ; end
def tube_marked?(e, val) ; e.get_attribute(@dico_hole, @attr_tube) == val ; end
def tube_marked_any?(e) ; e.get_attribute(@dico_hole, @attr_tube) ; end
def tube_unmark(e) ; e.delete_attribute(@dico_hole, @attr_tube) ; end
def tube_mark_solid(e, id) ; e.set_attribute(@dico_hole, @attr_solid, id) ; end
def tube_marked_solid?(e, id) ; e.get_attribute(@dico_hole, @attr_solid) == id ; end
def tube_marked_shape?(e) ; [@attr_shape_min, @attr_shape_max].include?(e.get_attribute(@dico_hole, @attr_tube)) ; end
def tube_mark_transfer(e1, e2) 
	e2.set_attribute @dico_hole, @attr_tube, e1.get_attribute(@dico_hole, @attr_tube)
	e2.set_attribute @dico_hole, @attr_solid, e1.get_attribute(@dico_hole, @attr_solid)
end	

#TUBE: Calculating the tube dimension
def tube_boundaries(lst_solids, tr_axe)
	tinv = tr_axe.inverse
	zmin = zmax = nil

	lst_solids.each do |psolid|
		tr = psolid.tr
		ent = (psolid.type == :comp) ? psolid.entities : psolid.lst_ents
		faces = ent.grep(Sketchup::Face)
		hedges = {}
		faces.each do |face|
			face.edges.each { |e| hedges[e.entityID] = e }
		end	
		edges = hedges.values
		edges.each do |edge|
			[edge.start, edge.end].each do |vx|
				pt = tinv * tr * vx.position
				z = pt.z
				zmax = z if !zmax || z > zmax
				zmin = z if !zmin || z < zmin
			end
		end
	end

	[zmin, zmax]
end

#TUBE: Constructing the tube
def tube_build(entities, tr_entities, lst_solids, tr_solid, tubages, mirrored)

	#Creating the individual tubes for each shape loop of the stencil
	@lst_pshapes.each do |pshape|
		tr_axe = pshape.tr_axe
		
		#Computing the boundaries for the tube based on solids specified
		zmin0, zmax0 = tube_boundaries(lst_solids, tr_axe)
		next unless zmin0
		d = [(zmax0 - zmin0) * 0.01, 1].max
		zmin = zmin0 - d
		zmax = zmax0 + d
		
		#Computing the master transformation for the tube
		t = tr_solid * tr_axe
		ptmin = Geom::Point3d.new 0, 0, zmin
		ptmax = Geom::Point3d.new 0, 0, zmax
		tmin = t * Geom::Transformation.translation(ptmin)
		tmax = t * Geom::Transformation.translation(ptmax)
		
		#Creating the tubes
		pshape.loops_info.each do |info|
			tubage = tubage_create pshape, [zmin0, zmax0], tr_entities
			tubages.push tubage
			
			#Creating the main tube
			pts, type, interior = info		
			tube_group = tube_construct_geometry entities, pts, tmin, tmax, interior, mirrored
			tubage.tube_group = tube_group			
		end	
	end	
end

#TUBE: Create a single tube out of a specified contour
def tube_construct_geometry(entities, pts, tmin, tmax, interior=nil, mirrored=false)
	tube_group = entities.add_group
	gent = tube_group.entities
	
	pts = pts.reverse if mirrored
	ptsmin = pts.collect { |pt| tmin * pt }
	ptsmax = pts.collect { |pt| tmax * pt }

	#Computing the quads for the faces
	quads = []
	n = pts.length - 1
	n -= 1 if pts.first == pts.last
	for i in 0..n
		j = (i == n) ? 0 : i+1
		quads.push [ptsmin[i], ptsmin[j], ptsmax[j], ptsmax[i]]
	end	
	
	#Creating the faces
	quads.each do |quad| 
		f = gent.add_face(quad)
		tube_mark f, @attr_quad
		f.material = @custom_material unless @custom_material == :same
	end	

	#Creating the shape edges and line edges
	for i in 0..n
		j = (i == n) ? 0 : i+1
		edge = gent.add_line [ptsmin[i], ptsmin[j]]
		tube_mark edge, @attr_shape_min
		edge = gent.add_line [ptsmax[i], ptsmax[j]]
		tube_mark edge, @attr_shape_max
		edge = gent.add_line [ptsmin[i], ptsmax[i]]
		edge.soft = edge.smooth = true if interior && interior[i]
		tube_mark edge, @attr_line
	end
	
	#Returning the created group
	tube_group
end

#TUBE: Extend the faces to all faces within the stencil shape
def tube_solid_faces_inside(psolid, lfaces, pshape_or_info)
	lfaces = lfaces.find_all { |f| f.valid? }
	return {} if lfaces.empty?
	
	if pshape_or_info.class == PseudoShape
		pshape = pshape_or_info
		stencil_shape = pshape.stencil_shape
		t = pshape.tr_axe.inverse * psolid.tr
		inside_proc = proc { |face| @stencil.pseudoshape_face_inside?(stencil_shape, face, t) }
	else
		tr_solid = psolid.tr
		inside_proc = proc { |face| stamp_face_inside_bottom?(face, tr_solid, pshape_or_info) }
	end	
	
	#Identifying edges at the borders of the intersection
	hfaces = {}
	hedges = {}
	lfaces.each do |f| 
		hfaces[f.entityID] = true
		f.edges.each { |e| hedges[e.entityID] = e }
	end	
	hsh_edges_ini = {}
	hedges.values.each do |e| 
		lf = e.faces.find_all { |f| hfaces[f.entityID] }
		hsh_edges_ini[e.entityID] = true if lf.length > 1
	end		
	
	#Exploration of the faces from the intersecting edges
	hsh_faces_used = {}
	hsh_faces_inside = {}
	while lfaces.length > 0
		hedges = {}
		lfaces.each do |face|
			next unless face.valid?
			face_id = face.entityID
			next if hsh_faces_used[face_id]			
			hsh_faces_used[face_id]	= true
			next if tube_marked?(face, @attr_quad)

			#Check if face is within the stencil
			next unless inside_proc.call(face)	
			
			#Faces is part of the shape. Getting its edges for extension
			hsh_faces_inside[face_id] = face
			face.edges.each do |e| 
				next if hsh_edges_ini[e.entityID]
				hedges[e.entityID] = e if e.faces.find { |f| !hsh_faces_used[f.entityID] } 
			end	
		end
		
		#Computing the next faces from these edges
		hfaces_next = {}
		hedges.each do |eid, e| 
			e.faces.each { |f| hfaces_next[f.entityID] = f unless hsh_faces_used[f.entityID] } if e.valid?
		end	
		lfaces = hfaces_next.values
	end
	
	#Returning the list of faces
	hsh_faces_inside
end

#TUBE: Clean-up the solid to create holes by erasing faces with the stencil projection
def tube_solid_cleanup(psolid, lfaces, pshape)
	hfaces_erase = tube_solid_faces_inside(psolid, lfaces, pshape)
	
	#Selecting the edges bordering the faces
	ledges_erase = []
	hedges = {}
	hfaces_erase.each do |fid, face|
		face.edges.each do |edge|
			eid = edge.entityID
			next if hedges[eid]
			hedges[eid] = edge
			ledges_erase.push edge unless edge.faces.find { |f| !hfaces_erase[f.entityID] }
		end
	end
	
	#Erasing the faces and edges
	lerase = hfaces_erase.values + ledges_erase
	psolid.entities.erase_entities lerase unless lerase.empty?
end

#TUBE: Erasing the edges if any left standalone or bordering a face not within Stencil
def tube_solid_cleanup_lonely_edges(psolid, ledges, pshape)
	stencil_shape = pshape.stencil_shape
	t = pshape.tr_axe.inverse * psolid.tr
	lerase = []
	ledges.each do |e|
		next unless e.valid?
		next unless e.faces.length == 0
		ptmid = Geom.linear_combination 0.5, e.start.position, 0.5, e.end.position
		lerase.push e if @stencil.pseudoshape_point_inside?(stencil_shape, t * ptmid, false) ||
						(@stencil.pseudoshape_point_inside?(stencil_shape, t * e.start.position, true) && 
						 @stencil.pseudoshape_point_inside?(stencil_shape, t * e.end.position, true))			 
	end	
	psolid.entities.erase_entities lerase unless lerase.empty?
end

#TUBE: Merge tube groups when based on the same pshape
def tube_merge_all(tube_groups)
	hshapes = {}
	tube_groups.each do |tube_group, pshape, tr|
		idshape = pshape.object_id
		info = hshapes[idshape]
		if info
			main_group = info[0]
			tube_merge main_group, tube_group
			tube_group.erase!
		else
			hshapes[idshape] = [tube_group, pshape, tr]
		end	
	end
	hshapes.values
end

#TUBE: Recreate the tube group into the main group
def tube_merge(main_group, tube_group)
	mgent = main_group.entities
	tube_group.entities.grep(Sketchup::Face).each do |face|
		pts = face.outer_loop.vertices.collect { |vx| vx.position }
		f = mgent.add_face pts
		G6.face_transfer_properties(face, f)
	end
	tube_group.entities.grep(Sketchup::Edge).each do |edge|
		e = mgent.add_line edge.start.position, edge.end.position
		G6.edge_transfer_properties(edge, e)
		tube_mark_transfer edge, e
	end
	tube_group.entities.grep(Sketchup::ConstructionLine).each do |cline|
		mgent.add_cline cline.start, cline.end
	end
end

#TUBE: Remove the collinear edges in the solid
def tube_remove_coplanar_edges(entities)
	lfaces = entities.grep(Sketchup::Face).find_all { |f| tube_marked?(f, @attr_quad) }
	hedges = {}
	lerase = []
	lfaces.each { |f| tube_unmark(f) }
	lfaces.each do |face|
		face.edges.each do |e|
			eid = e.entityID
			next if hedges[eid]
			lerase.push e if edge_coplanar?(e)
		end
	end
	entities.erase_entities lerase unless lerase.empty?
end

#TUBE: Convert the edges to construction lines
def tube_convert_to_mark(tube_group)
	gent = tube_group.entities
	edges = gent.grep(Sketchup::Edge)
	edges.each do |edge|
		pt1 = edge.start.position
		pt2 = edge.end.position
		gent.add_cline pt1, pt2
	end
	gent.erase_entities edges
end

#TUBE: Construct a main group to store intersections
def tube_make_group_out(tube_groups)
	#Main group
	if tube_groups.length == 1 && @lst_solids.length == 1
		out_ent = @top_entities
	else	
		@out_group = @top_entities.add_group unless @out_group
		out_ent = @out_group.entities
	end
	
	#Recreating the tube groups within the main group
	tube_groups.each do |tube_group, pshape, tr|
		tent = tube_group.entities
		g = out_ent.add_group
		gent = g.entities
		
		#Transferring the faces
		G6.copy_faces_from tent.grep(Sketchup::Face), gent, tr
		
		#Transferring edges
		G6.copy_edges_from tent.grep(Sketchup::Edge), gent, tr
		
		#Transferring Construction lines
		G6.copy_clines_from tent.grep(Sketchup::ConstructionLine), gent, tr
		
		#Erasing the original group
		tube_group.erase!
	end
end

#--------------------------------------------------------------
# STAMP: Intersection with stamp
#--------------------------------------------------------------

#STAMP: Top method for Stamping, Punching, Carving and Embossing
def stamp_intersect_synchronous
	#Preparing the solid and shapes
	#return finish_execute unless execution_prepare
	
	#Prepare the tubages
	stamp_geometry_build_tubes
	
	#Intersecting the component with the tubes and putting the intersection lines in the Tube group
	stamp_geometry_fill_tubes
	
	#Filtering portions in the tube and keeping only valid tubes and erasing the bad ones
	stamp_geometry_analyze_tubes
	
	#Building the tube info with intersections in the solids and executing the stamping operation with cleanup
	stamp_geometry_perform_stamping
end

#STAMP: Process Stamp execution in asynchronous mode
def stamp_intersect_robot
	begin
		stamp_intersect_robot_protected
	rescue Exception => e
		@status_execution = e
		@suops.abort_execution
	end	
end

#STAMP: Process Stamp execution in asynchronous mode (within error capture)
def stamp_intersect_robot_protected
	while(action, *param = @suops.current_step) != nil	
		case action
		
		when :_init
			next_step = (execution_prepare) ? [:build_tubes, 0] : :finish
		
		when :finish
			finish_execute
			next_step = nil
		
		else
			return if @suops.yield?
			@suops.countage
			case action
			when :build_tubes
				stamp_geometry_build_tubes
				next_step = :fill_tubes
				
			when :fill_tubes
				stamp_geometry_fill_tubes
				next_step = :analyze_tubes
			
			when :analyze_tubes
				stamp_geometry_analyze_tubes
				next_step = :perform_stamping
				
			when :perform_stamping
				stamp_geometry_perform_stamping
				next_step = :finish
			end
			
			@status_execution = true if @lst_tubages.length > 0
		end	
		
		break if @suops.next_step(*next_step)
	end
end

#STAMP: Prepare the main tubages
def stamp_geometry_build_tubes
	#Offset parameter
	@offset = (@cut_style == :carve) ? -@carve_offset.abs : @carve_offset.abs

	#Creating the Tubes at top level
	@lst_tubages = []
	tube_build @top_entities, @tr_id, @lst_solids, @tr_id, @lst_tubages, false
end

#STAMP: Intersecting the component with the tubes and putting the intersection lines in the Tube group
def stamp_geometry_fill_tubes
	@lst_solids.each do |psolid|
		tr = psolid.tr
		entities = psolid.entities
		solid_id = psolid.comp.entityID
		@lst_tubages.each do |tubage|
			tube_group = tubage.tube_group
			edges = entities.intersect_with false, tr, tube_group.entities, @tr_id, false, tube_group
			edges.each { |e| tube_mark_solid(e, solid_id) }
		end
	end
end

#STAMP: Analyzing and filitering the right contours
def stamp_geometry_analyze_tubes
	good_tubages = []
	bad_tubages = []
	@lst_tubages.each do |tubage|
		pshape = tubage.pshape
		tube_group = tubage.tube_group
		bounds = tubage.bounds
		if section_analyze tubage, @tr_id
			good_tubages.push tubage
		else
			bad_tubages.push tubage
		end	
	end	
	@lst_tubages = good_tubages
	bad_tubages.each { |tubage| tubage.tube_group.erase! }
end

#STAMP: Perform the stamping and cleanup
def stamp_geometry_perform_stamping
	#Building the tube info with intersections in the solids
	tube_groups_info = stamp_make_tube_info @lst_tubages
	
	#Stamp and marking with Guide lines
	if @cut_style == :stamp
		if @stamp_mark
			stamp_mode_mark tube_groups_info
			
		elsif @stamp_punch
			stamp_mode_punch tube_groups_info
			
		else
			stamp_mode_stamp tube_groups_info
		end
	end
	
	#Carve: Creating fully formed tube groups with carve and emboss shape
	if @cut_style == :carve || @cut_style == :emboss
		stamp_mode_carve_emboss tube_groups_info
	end	
	
	#Un-marking all faces and erasing collinear edges
	@lst_solids.each do |psolid|
		tube_remove_coplanar_edges(psolid.entities)
	end
	
	#Erasing the original tubes
	@top_entities.erase_entities @lst_tubages.collect { |tubage| tubage.tube_group }	
end


#STAMP: Manage the mode MARK
def stamp_mode_carve_emboss(tube_groups_info)
	#Init of info
	final_tubes_info = []
	cleanup_info_faces = []
	cleanup_info_edges = []
	
	#Carve: Creating fully formed tube groups with carve and emboss shape
	tube_groups_info.each do |psolid, edges, pshape|	
	
		#Faces forming the inside of hole
		t = pshape.tr_axe.inverse * psolid.tr
		hfaces = {}
		edges.each { |e| e.faces.each { |f| hfaces[f.entityID] = f } if e.valid? }
		hfaces = tube_solid_faces_inside(psolid, hfaces.values, pshape)
		lfaces = hfaces.values
		
		#Creating the tube with carving / emboss
		next if lfaces.empty?
		g, proj_info = stamp_create_final_tubes(psolid, edges, lfaces, pshape)
		final_tubes_info.push [g, psolid, pshape, proj_info]
		
		#Clean-up the hole faces
		hedges = {}
		lfaces.each { |f| f.edges.each { |e| hedges[e.entityID] = e } }	
		cleanup_info_faces.push [psolid, lfaces, pshape]
		cleanup_info_edges.push [psolid, hedges.values, pshape]
	end	

	#Re-intersecting the final tubes with the solids
	unless @cut_out_group
		#Performing the clean-ups for faces
		cleanup_info_faces.each { |info| tube_solid_cleanup *info }
		
		#Re-intersecting the final tubes with the solids
		final_tubes_info.each do |tube_group, psolid0, pshape, proj_info|
			tr0_inv = psolid0.tr.inverse
			@lst_solids.each do |psolid|
				entities = psolid.entities
				t = tr0_inv * psolid.tr
				edges = entities.intersect_with(false, t, entities, t, false, tube_group)
				next if edges.empty?	
				hfaces = {}
				edges.each { |e| e.soft = e.smooth = false } 
				edges.each { |e| e.faces.each { |f| hfaces[f.entityID] = f } }
				tube_solid_cleanup(psolid, hfaces.values, proj_info)
			end	
		end
		
		#Performing the clean-up of lonely edges in the solid
		cleanup_info_edges.each { |info| tube_solid_cleanup_lonely_edges *info }
	end
	

	#Exploding the tubes (unless Group option is on)
	unless @cut_in_group || @cut_out_group
		final_tubes_info.each do |tube_group, psolid, pshape|
			edges = G6.grouponent_explode(tube_group).grep(Sketchup::Edge)
		end
	end	
	
	#GROUP OUT: creating the master group
	if @cut_out_group
		lst_groups = []
		final_tubes_info.each do |tube_group, psolid, pshape|
			lst_groups.push [tube_group, pshape, psolid.tr]
		end		
		tube_make_group_out lst_groups
		stamp_erase_intersections tube_groups_info
	end	
end

#STAMP: Manage the mode MARK
def stamp_mode_mark(tube_groups_info)
	#Converting the intersection edges into Guide lines in groups
	lst_tubes = []
	tube_groups_info.each do |psolid, edges, pshape|
		hfaces = {}
		edges.each { |e| e.faces.each { |f| hfaces[f.entityID] = f } if e.valid? }
		entities = psolid.entities
		g = entities.add_group
		lst_tubes.push [g, pshape, psolid.tr]
		gent = g.entities
		edges.each do |edge|
			gent.add_cline edge.start.position, edge.end.position
		end
	end	
	
	#Creating the groups if mode is group
	if @cut_out_group
		tube_make_group_out lst_tubes
	elsif !@cut_in_group
		lst_tubes.each { |tg| tg[0].explode }
	end
	
	#Erasing the intersection edges in the solid
	stamp_erase_intersections(tube_groups_info)
end

#STAMP: Manage the mode PUNCH
def stamp_mode_punch(tube_groups_info)
	lst_tubes = []
	tube_groups_info.each do |psolid, edges, pshape|
		if @cut_in_group || @cut_out_group
			entities = psolid.entities
			g = entities.add_group
			lst_tubes.push [g, pshape, psolid.tr]
			gent = g.entities
			edges.each do |edge|
				gent.add_line edge.start.position, edge.end.position
			end
		end
		
		unless @cut_out_group
			hfaces = {}
			edges.each { |e| e.faces.each { |f| hfaces[f.entityID] = f } if e.valid? }
			tube_solid_cleanup(psolid, hfaces.values, pshape)
		end			
	end	
	
	#Creating the groups if mode is OUT group
	if @cut_out_group
		tube_make_group_out lst_tubes
		stamp_erase_intersections(tube_groups_info)
	end
end

#STAMP: Manage the mode STAMP
def stamp_mode_stamp(tube_groups_info)
	#NO Group mode
	unless @cut_in_group || @cut_out_group
		if @custom_material != :same
			tube_groups_info.each do |psolid, edges, pshape|
				hfaces = {}
				edges.each { |e| e.faces.each { |f| hfaces[f.entityID] = f } if e.valid? }
				hfaces = tube_solid_faces_inside(psolid, hfaces.values, pshape)
				hfaces.values.each { |f| f.material = f.back_material = @custom_material }	
			end	
		end
		return
	end
	
	#GROUP mode. Creating tubes with the stamp intersections and colorizing with custom material if required
	lst_tubes = []
	tube_groups_info.each do |psolid, edges, pshape|
		hfaces = {}
		edges.each { |e| e.faces.each { |f| hfaces[f.entityID] = f } if e.valid? }
		hfaces = tube_solid_faces_inside(psolid, hfaces.values, pshape)
		lfaces = hfaces.values
		
		entities = psolid.entities
		g = entities.add_group
		lst_tubes.push [g, pshape, psolid.tr]
		gent = g.entities
		
		#Transferring faces 
		new_faces = G6.copy_faces_from(lfaces, gent, @tr_id)
		new_faces.each { |f| f.material = f.back_material = @custom_material } if @custom_material
		
		#Transferring edges
		hedges = {}
		lfaces.each { |f| f.edges.each { |e| hedges[e.entityID] = e } }
		G6.copy_edges_from(hedges.values, gent, @tr_id)
		
		#Cleanup the hole in the solid
		if @cut_in_group
			tube_solid_cleanup(psolid, lfaces, pshape)
		end
	end	
	
	#Creating the groups if mode is OUT group
	if @cut_out_group
		tube_make_group_out lst_tubes
		stamp_erase_intersections(tube_groups_info)
	end
end

#STAMP: Erase the intersections made in the solid (to leave it unchanged)
def stamp_erase_intersections(tube_groups_info)
	tube_groups_info.each do |psolid, edges, pshape|
		entities = psolid.entities
		ledges = edges.find_all { |e| e.valid? && G6.edge_coplanar?(e) }
		entities.erase_entities ledges
	end	
end

#STAMP: Merging the tube info
def stamp_make_tube_info(tubages)
	tube_groups_info = []
	@lst_solids.each do |psolid|
		tr = psolid.tr
		entities = psolid.entities
		solid_id = psolid.comp.entityID
		
		#Creating the intersections in the solid
		tubages.each do |tubage|
			pshape = tubage.pshape
			tube_group = tubage.tube_group
			mledges = tube_group.entities.grep(Sketchup::Edge).find_all { |e| tube_marked_solid?(e, solid_id) }
			next if mledges.empty?
			edges = entities.intersect_with(false, tr, entities, tr, false, tube_group)
			tube_groups_info.push [psolid, edges, pshape] unless edges.empty?
		end
	end

	hshapes = {}
	tube_groups_info.each do |psolid, edges, pshape|
		idshape = pshape.object_id
		info = hshapes[idshape]
		if info
			info[1].concat edges
		else
			hshapes[idshape] = [psolid, edges, pshape]
		end	
	end
	hshapes.values
end

#STAMP: Check if a bottom face is within the stencil and should be erased
def stamp_face_inside_bottom?(face, tr_solid, proj_info)
	vecdir, base_loops = proj_info
	face.vertices.each do |vx|
		pt3d = tr_solid * vx.position
		return false unless base_loops.find { |loop_info| stamp_point_inside_single_loop?(pt3d, vecdir, loop_info) }
	end
	true
end

#STAMP: Check if a point of the bottom faces is within the stencil and should be erased
def stamp_point_inside_single_loop?(pt3d, vecdir, loop_info)
	normal, loops = loop_info
	origin = loops[0][0]
	
	#Projecting the point along the direction of extrusion
	plane = [origin, normal]
	ptinter = Geom.intersect_line_plane [pt3d, vecdir], plane
	return false unless ptinter || pt3d.on_plane?(plane?)
	tr_axe = Geom::Transformation.axes origin, *(normal.axes)
	
	#Checking if the projected point is within the polygon formed by the loops
	t = tr_axe.inverse
	pt2d = t * ptinter
	polys = loops.collect { |loop| loop.collect { |pt| t * pt } }
	return false unless Geom.point_in_polygon_2D(pt2d, polys[0], true)
	polys[1..-1].each do |poly|
		return false if Geom.point_in_polygon_2D(pt2d, poly, false)
	end
	true
end

#STAMP: Create the full tubes with carving or embossing
def stamp_create_final_tubes(psolid, edges, lfaces, pshape)
	#Creating the tube group
	g = psolid.entities.add_group
	gent = g.entities
	tr_solid = psolid.tr
	solid_id = psolid.comp.entityID
	
	#Registering the edges and faces
	hedges = {}
	edges.each { |e| hedges[e.entityID] = true if e.valid? }	
	hfaces = {}
	lfaces.each { |f| hfaces[f.entityID] = true }
	hedges_border = {}
	lfaces.each do |f|
		f.edges.each do |e|
			eid = e.entityID
			lf = e.faces.find_all { |ff| hfaces[ff.entityID] }
			hedges_border[e.entityID] = e if lf.length == 1
		end	
	end
		
	#Computing the average direction
	stencil_shape = pshape.stencil_shape
	tv = tr_solid.inverse * pshape.tr_axe
	tvinv = tv.inverse
	vecref = G6.transform_vector Z_AXIS, tv
	tu = tr_solid
	tuinv = tu.inverse
	
	if @carve_plan
		vecdir = tr_solid.inverse * @repere_tr * Z_AXIS
		anchor = tv * pshape.anchor
	else
		anchor, vecdir2 = stamp_average_direction_anchor(lfaces, vecref, tu)
		vecdir = G6.transform_vector vecdir2, tuinv
	end
	
	#Reference plane for flat mode
	vec = G6.transform_vector vecdir, tvinv
	vec.length = @offset
	vecdir = tv * vec
	t = Geom::Transformation.translation vecdir
	origin = anchor
	origin = origin.offset(vecdir) if vecdir.valid? 
	plane = [origin, vecdir]
	
	#Creating the carving / embossing
	good_faces = []
	lfaces.each do |face|
		face.loops.each_with_index do |loop, iloop|
			lvx = loop.vertices
			
			#Creating the bottom faces
			if @carve_flat
				new_pts = lvx.collect { |vx| Geom.intersect_line_plane [vx.position, vecdir], plane }
			else
				new_pts = lvx.collect { |vx| t * vx.position }
			end	
			begin
				f = gent.add_face new_pts
			rescue
				next
			end	
			if iloop > 0 && hedges[loop.edges[0].entityID]
				f.erase!
			else
				good_faces.push face
				if @custom_material != :same
					f.material = f.back_material = @custom_material
				else	
					f.material = face.material
					f.back_material = face.back_material
				end
				lvx2 = f.vertices
				n = lvx.length-1
				for i in 0..n
					j = (i == n) ? 0 : i+1
					e0 = lvx[i].common_edge lvx[j]
					e2 = lvx2[i].common_edge lvx2[j]
					next unless e2 && e0
					G6.edge_transfer_properties(e0, e2) if (e0.faces.find_all { |ff| !hfaces[ff.entityID] }).length != 1
				end	
			end	
			
			#Creating the borders
			n = lvx.length - 1
			for i in 0..n
				j = (i == n) ? 0 : i+1
				vxi = lvx[i]
				vxj = lvx[j]
				edge = vxi.common_edge vxj
				
				#Create borders only for intersecting edges
				next unless hedges_border[edge.entityID]
				pti = vxi.position
				ptj = vxj.position
				pti2 = new_pts[i]
				ptj2 = new_pts[j]
				pts = [pti, ptj, ptj2, pti2] 
				
				#Creating the border face
				begin
					f = gent.add_face pts
				rescue
					next
				end	
				next unless f
				if @custom_material != :same
					f.material = f.back_material = @custom_material
				else	
					f.material = face.material
					f.back_material = face.back_material
				end
				tube_mark f, @attr_quad
			
				#Edge on border: need to be smooth and soft?
				status = @stencil.pseudoshape_line_interior_status(stencil_shape, [tvinv * pti, Z_AXIS])
				vx1 = f.vertices.find { |vx| vx.position == pti }
				edge = vx1.edges.find { |e| e.other_vertex(vx1).position == pti2 }
				next unless edge
				if status == :unknown
					edge.soft = true
				elsif status == true
					edge.soft = true
				end	
				edge.smooth = G6.edge_would_smooth(edge)
			end		
		end
	end

	#Creating the base faces information for future cleanup
	base_faces_info = []
	good_faces.each do |face|
		loops_info = face.loops.collect { |loop| loop.vertices.collect { |vx| tr_solid * vx.position } }
		base_faces_info.push [G6.transform_vector(face.normal, tr_solid), loops_info]
	end
	
	#Cleaning up coplanar edges
	lerase = gent.grep(Sketchup::Edge).find_all { |edge| edge_coplanar?(edge) }
	gent.erase_entities lerase unless lerase.empty?
	
	#Returning the group and projection info
	[g, [G6.transform_vector(vecdir, tr_solid), base_faces_info]]
end

#Check if an edge is coplanar
def edge_coplanar?(edge)
	dotmax = 0.99999999999999
	faces = edge.faces
	(faces.length != 2) ? false : (faces.first.normal % faces.last.normal).abs > dotmax
end

#STAMP: Calculate the average direction of a set of faces
def stamp_average_direction_anchor(lfaces, vecref, t=nil)
	vecref = G6.transform_vector vecref, t if t
	info_areas = lfaces.collect { |face| G6.face_centroid_area(face) + [face.normal] }
	xc = yc = zc = 0
	xv = yv = zv = 0
	info_areas.each do |centroid, area, normal|
		normal = G6.transform_vector normal, t if t
		normal = normal.reverse if normal % vecref < 0
		xv += normal.x * area
		yv += normal.y * area
		zv += normal.z * area
		xc += centroid.x
		yc += centroid.y
		zc += centroid.z
	end
	n = info_areas.length
	xc /= n
	yc /= n
	zc /= n
	anchor = Geom::Point3d.new(xc, yc, zc)
	vecdir = Geom::Vector3d.new xv, yv, zv
	[anchor, vecdir]
end

#--------------------------------------------------------------
# THRU: Intersection Through Solid with a tube
#--------------------------------------------------------------

#THRU: Process Thru execution in synchronous mode
def thru_intersect_synchronous
	#Preparing the solid and shapes
	return finish_execute unless execution_prepare
			
	#Loops on all solids for drilling
	@lst_solids.each_with_index do |psolid|
		#Creating the Tubes
		thru_geometry_prepare_tubages(psolid)
		
		#Preparing and filtering the tubes
		thru_geometry_analyze_tubages(psolid)

		#Intersecting the component with the tubes
		thru_geometry_intersect_tubages(psolid)
		
		#Merging the tube groups by pshape if groups are preserved
		thru_geometry_grouping
	end
	@status_execution = true
	finish_execute	
end

#THRU: Process Thru execution in asynchronous mode
def thru_intersect_robot
	begin
		thru_intersect_robot_protected
	rescue Exception => e
		@status_execution = e
		@suops.abort_execution
	end	
end

#THRU: Process Thru execution in asynchronous mode (within error capture)
def thru_intersect_robot_protected
	while(action, *param = @suops.current_step) != nil	
		case action
		
		when :_init
			next_step = (execution_prepare) ? [:prepare, 0] : :finish
		
		when :finish
			finish_execute
			next_step = nil
			
		else	
			return if @suops.yield?
			isolid, = param
			psolid = @lst_solids[isolid]
			action = nil unless psolid
			@suops.countage

			case action
			when :prepare
				thru_geometry_prepare_tubages psolid
				next_step = [:analyze, isolid]
				
			when :analyze
				thru_geometry_analyze_tubages psolid
				next_step = [:intersect, isolid]
				
			when :intersect
				thru_geometry_intersect_tubages psolid
				next_step = [:grouping, isolid]
				
			when :grouping
				thru_geometry_grouping
				next_step = [:prepare, isolid + 1]
				
			else
				next_step = :finish
			end
			
		end	
		
		break if @suops.next_step(*next_step)
	end
end

#THRU: Creating the tubes for the solid
def thru_geometry_prepare_tubages(psolid)
	@lst_tubages = []
	tube_build psolid.entities, psolid.tr, [psolid], psolid.tr.inverse, @lst_tubages, psolid.mirrored
end

#THRU: Analysing and filtering intersections in the tubes
def thru_geometry_analyze_tubages(psolid)
	entities = psolid.entities
	good_tubages = []
	@lst_tubages.each do |tubage|
		tube_group = tubage.tube_group
		
		#Intersection put in the main Tube group
		entities.intersect_with false, @tr_id, tube_group.entities, @tr_id, false, tube_group
		
		#Adjusting the tube portions depending on specs (with/without tube, keep 1 tube)
		if section_analyze tubage, psolid.tr, @thru_nb_tube, @thru_no_face
			good_tubages.push tubage
		else
			tubage.invalid = true
			tube_group.erase!
		end	
	end
	@lst_tubages = good_tubages
end

#THRU: Intersecting the component with the tubes
def thru_geometry_intersect_tubages(psolid)
	entities = psolid.entities
	all_edges = []
	
	#Intersection of each tube with the solid
	@lst_tubages.each do |tubage|
		tube_group = tubage.tube_group

		#Replace edges by guide line in the tube for Mark option
		if @thru_mark
			tube_convert_to_mark(tube_group)
		end
		
		#Intersection put in the pseudo solid
		unless @cut_out_group || @thru_mark
			all_edges = all_edges + entities.intersect_with(false, @tr_id, entities, @tr_id, false, tube_group)
		end
		
		#Exploding the tube group within the solid	
		unless @cut_in_group || @cut_out_group
			all_edges += tube_group.explode.grep(Sketchup::Edge)	
		end	
	end
	
	#No intersection
	@status_execution = true unless all_edges.empty? && !@thru_mark && !@cut_out_group
	
	#Cleaning the pseudo solid in the hole
	unless @cut_out_group || @thru_mark
		hfaces = {}
		hedges_used = {}
		all_edges.each do |e| 
			next unless e.valid?
			eid = e.entityID
			next if hedges_used[eid]
			hedges_used[eid] = true
			e.faces.each { |f| hfaces[f.entityID] = f unless tube_marked?(f, @attr_quad) } 
		end		
		@lst_tubages.each do |tubage|
			tube_solid_cleanup psolid, hfaces.values, tubage.pshape
		end
		
		#Un-marking all faces
		entities.grep(Sketchup::Face).each { |f| tube_unmark(f) }
	end	
end

#THRU: Managing the grouping
def thru_geometry_grouping
	return if @lst_tubages.empty?
	
	#Merging the tube groups by pshape if groups are preserved
	if @cut_in_group || @cut_out_group
		tube_groups = @lst_tubages.collect { |tubage| [tubage.tube_group, tubage.pshape, tubage.tr_entities] }
		tube_groups = tube_merge_all(tube_groups)
	end	
	
	#Making the Group out
	if @cut_out_group
		tube_make_group_out(tube_groups)
	end
end

#--------------------------------------------------------------
# SECTION: Analyse the split of contour in tubes
#--------------------------------------------------------------

#SECTION: Erasing the tube sections according to specs
def section_analyze(tubage, tr_solid, nb_tube=nil, no_face=false)	
	#Initialization
	tube_group = tubage.tube_group
	entities = tube_group.entities
	pshape = tubage.pshape
	
	#Getting all edges in the tube group
	ledges = entities.grep(Sketchup::Edge)
	return nil if ledges.empty?
	
	#Transformations and axis of tube
	anchor = pshape.anchor
	tr_axe = pshape.tr_axe
	tinv = tr_axe.inverse * tr_solid	
	t = tinv.inverse	
	normal = t * Z_AXIS
	angle_stop = 45.degrees
	colors = ['red', 'orange', 'lightblue', 'lightgreen', 'purple', 'thistle', 'brown', 'pink', 'yellow']
	ncolor = colors.length

	#Checking if the anchor for the shape is within the tube boundaries
	zanchor = anchor.z
	zmin, zmax = tubage.bounds
	return nil if zanchor < zmin
	
	#Boundary planes
	edge_max = ledges.find { |e| tube_marked?(e, @attr_shape_max) }
	plane_max = [edge_max.start.position, normal]
	
	#Eliminating the contours which are well behind in the tube group
	fac = 0.5
	zbehind = anchor.z + pshape.bbox.diagonal * fac
	return nil unless section_eliminate_behind(tube_group, tinv, zbehind)
	
	#Classifying the edges
	ledges = entities.grep(Sketchup::Edge)
	hedges_contour = {}
	hvx = {}
	ledges.each do |e|
		unless tube_marked?(e, @attr_line)
			[e.start, e.end].each { |vx| hvx[vx.entityID] = vx }
			hedges_contour[e.entityID] = e
		end	
	end
	
	#No more intersection in the tube
	return nil if hedges_contour.length == 0
	
	#Tracing lines at each vertex and intersecting them with contour edges
	hvx_num = {}
	hedge_num = {}
	hvx.each do |vxid, vx|
		next if !vx.valid? || hvx_num[vxid]
		line = [vx.position, normal]
		line[0] = Geom.intersect_line_plane line, plane_max
		section_intersect_line_edges(line, ledges, hvx_num, hedge_num, normal)
	end

	#Numbering the edges
	ledges.each do |edge|
		eid = edge.entityID
		n_min = [hvx_num[edge.start.entityID], hvx_num[edge.end.entityID]].min
		n_max = [hvx_num[edge.start.entityID], hvx_num[edge.end.entityID]].max
		unless hedge_num[eid]
			hedge_num[eid] = (tube_marked?(edge, @attr_line)) ? n_max : n_min
		end
	end

	#Erasing the border lines, one out of 2 as well as the contours in the back
	ledges_line = ledges.find_all { |e| tube_marked?(e, @attr_line) }
	lerase = []
	ledges_line.each do |edge|
		n = hedge_num[edge.entityID]
		lerase.push edge if n % 2 == 0
		#lerase.push edge if n < ibeg_contour || (n + ibeg_contour) % 2 == 0 #&& edge.faces != 1
		#lerase.push edge if n < ibeg_contour || n % 2 == 0 #&& edge.faces.length != 1
	end
	entities.erase_entities lerase unless lerase.empty?

	#Contour edges considered as pseudo lines
	hedges_contour = {}
	hpseudo_lines = {}
	ledges = entities.grep(Sketchup::Edge)
	ledges.each do |e|
		if tube_marked?(e, @attr_line)
			hpseudo_lines[e.entityID] = e
		else	
			angle = e.start.position.vector_to(e.end.position).angle_between(normal)
			if angle_stop && (angle <= angle_stop || angle >= Math::PI - angle_stop)
				hpseudo_lines[e.entityID] = e 
			else
				hedges_contour[e.entityID] = e
			end
		end	
	end	
	
	#Calculating the contour sequences
	all_sequences = section_calculate_sequences(hedges_contour)
	
	#Adjusting the numbering and collecting the contours by number
	all_contours = []
	all_sequences.each do |led, lvx|
		nmax = (led.collect { |e| hedge_num[e.entityID] }).max
		led.each { |e| hedge_num[e.entityID] = nmax }
		all_contours[nmax] = [] unless all_contours[nmax]
		all_contours[nmax].push led
	end
	return nil if all_contours.empty?
	
	#Adjusting the numbering for lines and pseudo lines
	hpseudo_lines.each do |eid, edge|
		led = (edge.start.edges + edge.end.edges).find_all { |e| !hpseudo_lines[e.entityID] }
		next if led.empty?
		hedge_num[edge.entityID] = (led.collect { |e| hedge_num[e.entityID] }).max
	end
	
	#Finding the front contour
	ibeg_contour = section_front_contour(all_contours, tinv, anchor)
	
	#Erasing the contour in front of the front contour
	lerase = entities.grep(Sketchup::Edge).find_all { |e| n = hedge_num[e.entityID] ; !n || n < ibeg_contour }
	entities.erase_entities lerase unless lerase.empty?
	
	#THRU mode: Keeping the number of requested tube
	if @cut_style == :thru
		section_isolate_tubes(tube_group, ibeg_contour - 1 + 2 * nb_tube, hedge_num) if nb_tube
		
		#Erasing the faces if Stamp mode or Thru mode with no surface
		if no_face
			lerase = entities.grep(Sketchup::Face)
			lerase.concat entities.grep(Sketchup::Edge).find_all { |e| tube_marked?(e, @attr_line) }
			entities.erase_entities lerase unless lerase.empty?
		end	
		
	#STAMP, CARVE, EMBOSS mode: Keeping only the front contour
	else
		#section_isolate_tubes(tube_group, ibeg_contour + 1, hedge_num)

		#Keeping only the front contour
		lerase = entities.grep(Sketchup::Face)
		hedges = {}
		all_contours[ibeg_contour].each { |led| led.each { |e| hedges[e.entityID] = true if e.valid? } }
		lerase.concat entities.grep(Sketchup::Edge).find_all { |e| !hedges[e.entityID] }
		entities.erase_entities lerase unless lerase.empty?
		
		#Removing the pseudo lines which are close to the tube axis by the angle stop
		section_isolate_contours tube_group, [t * anchor, normal], angle_stop
	end
	
	#Return the tube group
	tube_group	
end

#SECTION: Calculate the sequences of contours
def section_calculate_sequences(hedges_contour)
	hedges_used = {}
	all_sequences = []
	hedges_contour.each do |eid, edge|
		next if hedges_used[edge.entityID]
		ls_left = section_contour_sequence(edge, edge.start, hedges_contour, hedges_used)
		ls_right = section_contour_sequence(edge, edge.end, hedges_contour, hedges_used)
		ls = ls_left.reverse + [edge] + ls_right
		hedges_used[edge.entityID] = true
		hvx = {}
		ls.each { |e| [e.start, e.end].each { |vx| hvx[vx.entityID] = vx } }
		all_sequences.push [ls, hvx.values]
	end
	all_sequences
end

#SECTION: Eliminate the contours which are reasonably behind the anchor
def section_eliminate_behind(tube_group, tinv, zbehind)
	entities = tube_group.entities
	
	#Cleaning the two lead shape contours
	section_erase_shape_contours(tube_group)

	#Contour edges
	ledges = entities.grep(Sketchup::Edge)
	hedges_contour = {}
	ledges.each { |e| hedges_contour[e.entityID] = e unless tube_marked?(e, @attr_line) }
	return false unless hedges_contour.length > 0
	
	#Building sequences of contours
	all_sequences = section_calculate_sequences(hedges_contour)
	
	#Eliminating the sequences too much behind
	new_sequences = []
	lerase = []
	all_sequences.each do |lseq, lvx|
		if lvx.find { |vx| (tinv * vx.position).z < zbehind }
			new_sequences.push [lseq, lvx]
		else	
			lerase += lseq
		end	
	end
	return nil if new_sequences.empty?
	entities.erase_entities lerase unless lerase.empty?
	
	#Clean up lonely lines
	lerase = []
	new_sequences.each do |lseq, lvx|
		lvx.each do |vx|
			led_line = vx.edges.find_all { |e| e.other_vertex(vx).edges.length == 1 && tube_marked?(e, @attr_line) }
			lerase += led_line
		end	
	end
	entities.erase_entities lerase unless lerase.empty?
	
	#Intersection contours exist
	true
end

#SECTION: Calculate the contour sequence from an edge and vertex
def section_contour_sequence(edge0, vx0, hedges_contour, hedges_used)
	lseq = []
	ledges = [edge0]
	while edge0
		edges = vx0.edges.find_all { |e| e != edge0 && hedges_contour[e.entityID] && !hedges_used[e.entityID] }
		edge0 = edges[0]
		break unless edge0
		vx0 = edge0.other_vertex vx0
		hedges_used[edge0.entityID] = true
		lseq.push edge0
	end
	lseq
end

#SECTION: Erasing the tube sections outside
def section_erase_shape_contours(tube_group)
	entities = tube_group.entities
	lerase = []
	entities.grep(Sketchup::Edge).each do |e|
		if tube_marked_shape?(e)
			lerase.push e
		else	
			[e.start, e.end].each do |vx|
				lerase.push e if vx.edges.find { |ee| ee != e && tube_marked_shape?(ee) }
			end		
		end
	end	
	entities.erase_entities lerase unless lerase.empty?
end

#SECTION: Compute the front contour index based on proximity with anchor
def section_front_contour(all_contours, tinv, anchor)
	line = [anchor, Z_AXIS]
	dzmin = nil
	ibeg_contour = 0
	all_contours.each_with_index do |contour, icontour|
		next unless contour
		led_min = dmin = ptproj_min = nil
		contour.each do |led|
			lseg_weight = led.collect { |e| [e.start.position, e.end.position, e.length] }
			ptavg = tinv * G6.segment_weighted_average(lseg_weight)
			ptproj = ptavg.project_to_line line
			d = ptavg.distance ptproj
			if !dmin || d < dmin
				led_min = led
				dmin = d
				ptproj_min = ptproj
			end	
		end	
		dz = ptproj_min.distance anchor
		if !dzmin || dz < dzmin
			ibeg_contour = icontour
			dzmin = dz
		end				
	end
	ibeg_contour
end

#SECTION: Intersecting a line and all edges in tube
def section_intersect_line_edges(line, ledges, hvx_num, hedge_num, normal)
	origin, vec_line = line
	
	#Determining the intersection of the line with all edges
	linter = []
	ledges.each do |edge|
		vxbeg = edge.start
		ptbeg = vxbeg.position
		vxend = edge.end
		ptend = vxend.position
		vec_ed = ptbeg.vector_to ptend
		
		#Edge parallel to tube axis: take its two vertices
		if vec_line.parallel?(vec_ed)
			next if hvx_num[vxbeg.entityID] || !ptbeg.on_line?(line)
			linter.push [origin.distance(ptbeg), ptbeg, vxbeg, edge], [origin.distance(ptend), ptend, vxend, edge]
			
		#Intersection in one point: consider whether it is also a vertex	
		else
			ptinter = Geom.intersect_line_line line, [ptbeg, ptend]
			next unless ptinter
			if ptinter == ptbeg
				vxinter = vxbeg
			elsif ptinter == ptend
				vxinter = vxend
			elsif ptinter.vector_to(ptbeg) % ptinter.vector_to(ptend) < 0
				vxinter = nil
			else
				next
			end	
			next if vxinter && hvx_num[vxinter.entityID]
			linter.push [origin.distance(ptinter), ptinter, vxinter, edge]
		end
	end
	
	#Sorting the intersections by distance to base
	linter = linter.sort { |a, b| a.first <=> b.first }
	
	#Numbering the vertices and edges
	ptprev = nil
	j = -1
	linter.each do |a|
		d, ptinter, vxinter, edge = a
		j += 1 if ptprev != ptinter
		ptprev = ptinter
		if vxinter
			hvx_num[vxinter.entityID] = j
		else	
			hedge_num[edge.entityID] = j
		end	
	end	
end

#SECTION: eliminate the edges which are close to the group axis
def section_isolate_contours(tube_group, line_tube, angle_stop)
	entities = tube_group.entities
	ledges = entities.grep(Sketchup::Edge)
	anchor, normal = line_tube
	
	#Removing pseudo-lines
	lerase = []
	ledges.each do |edge|
		ptbeg = edge.start.position
		ptend = edge.end.position
		angle = ptbeg.vector_to(ptend).angle_between(normal)
		lerase.push edge if angle_stop && (angle <= angle_stop || angle >= Math::PI - angle_stop)
	end
	entities.erase_entities lerase unless lerase.empty?
end

#SECTION: Remove tubes beyond the limit nb_tube
def section_isolate_tubes(tube_group, icontour_min, hedge_num)
	entities = tube_group.entities
	ledges = entities.grep(Sketchup::Edge)
	
	#Finding connected geometry
	group_erase = []
	hedges_used = {}
	ledges.each do |edge|
		eid = edge.entityID
		next if hedges_used[eid]
		led = edge.all_connected
		led.each { |e| hedges_used[e.entityID] = true }
		if led.find { |e| e.instance_of?(Sketchup::Edge) && hedge_num[e.entityID] > icontour_min }
			group_erase.concat led
		end	
	end		
	entities.erase_entities group_erase unless group_erase.empty?
end

end	#class StencilIntersector

end #module Traductor