=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# © Copyright May 2014
# Designed and developped May 2014 by Fredo6
#
# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for commercial purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:   Lib6StencilManager.rb
# Original Date	:   18 May 2014
# Description	:   Manage stencils for Selection and Creation
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end

module Traductor

#=============================================================================================
#=============================================================================================
# Class StencilManager: 2D shapes Management and GUI
#=============================================================================================
#=============================================================================================

class StencilManager

#---------------------------------------------------------------------------------------------
# INIT: Initializations
#---------------------------------------------------------------------------------------------

#INIT: History for stencils
@@history = nil unless defined?(@@history)
@@ipos_history = 0 unless defined?(@@ipos_history)

#INIT: Class instance initialization
def initialize__(*hargs)
	#Initialization
	@model = Sketchup.active_model
	@selection = @model.selection
	@rendering_options = Sketchup.active_model.rendering_options
	@entities = @model.active_entities
	@view = @model.active_view
	@tr_id = Geom::Transformation.new
	@duration_long_click = 0.8
	@history_keep = 15
		
	#Parsing the arguments
	hargs.each do |harg|	
		harg.each { |key, value|  parse_args(key, value) } if harg.class == Hash
	end
	
	#Initializing the Key manager
	@keyman = Traductor::KeyManager.new() { |event, key, view| key_manage(event, key, view) }
	
	#Create the Origin-Target Picker
	hsh = { :ignore_selection => true, :hide_distance => true,
			:option_implicit_plane => :best_normal, :notify_proc => self.method("notify_from_otpicker") }
	@otpicker = Traductor::OriginTargetPicker.new @hsh_parameters, hsh
	hsh = {}
	hsh[:option_inference_comp] = true
	hsh[:option_inference_remote] = false
	hsh[:option_planar] = true
	hsh[:option_planar_code] = :no
	hsh[:option_vector] = false
	hsh[:option_vector_code] = :no
	@otpicker.option_set_multi hsh

	#Current stencil
	if @@history
		@stencil_cur = @@history[@@ipos_history]
	else
		fake_init_stencil
	end	
	
	#Other initialization
	init_colors
	init_cursors
	init_messages
	reset
end

#INIT: Parse the parameters of the Face Picker
def parse_args(key, value)
	skey = key.to_s
	case skey
	when /notify_proc/
		@notify_caller_proc = value
	when /history_keep/
		@history_keep = value
	end	
end	

def reset
	@mode = @otpicker.set_mode(:origin)
end

def exit
	picking_set false
end
	
def set_current_stencil(stencil)
	return unless stencil
	@stencil_cur = stencil
	@palette.button_refresh_draw(@symb_button_stencil_show)
	palette_button_show_stencil
	notify_caller :stencil
	stencil
end
	
def current_stencil
	@stencil_cur
end
	
def fake_init_stencil
	dx = 0.4.m
	dy = 0.2.m
	dec = 2.m
	pseudoshapes = []
	
	#Rectangle
	lpti = [[-dx, -dy], [dx, -dy], [dx, dy], [-dx, dy]].reverse
	pts = lpti.collect { |a| Geom::Point3d.new *a }
	t = Geom::Transformation.translation Geom::Point3d.new(-dec, 0, 0)
	loops = [pts.collect { |pt| t * pt }]
	pseudoshapes.push [loops]
	
	#Circle
	pts1 = G6.pts_circle 0, 0, dx, 64
	pts2 = G6.pts_circle 0, 0, dx/3, 64
	t = Geom::Transformation.translation Geom::Point3d.new(dec, 0, 0)
	pts1 = pts1.collect { |pt| t * pt }
	pts2 = pts2.collect { |pt| t * pt }
	interiors = Array.new(pts1.length, true)
	polys = []
	n = pts1.length - 1
	for i in 0..n
		j = (i == n) ? 0 : i+1
		polys.push [pts1[i], pts2[i], pts2[j], pts1[j]]
	end	
	loops = [pts1, pts2]
	pseudoshapes.push [loops, polys, [interiors, interiors]]
	
	#Ring hexagon
	pts1 = G6.pts_circle 0, 0, dx, 6
	pts2 = G6.pts_circle 0, 0, dx/2, 6
	t = Geom::Transformation.translation Geom::Point3d.new(0, 0, 0)
	pts1 = pts1.collect { |pt| t * pt }
	pts2 = pts2.collect { |pt| t * pt }
	polys = []
	n = pts1.length - 1
	for i in 0..n
		j = (i == n) ? 0 : i+1
		polys.push [pts1[i], pts2[i], pts2[j], pts1[j]]
	end	
	loops = [pts1, pts2]
	pseudoshapes.push [loops, polys]

	#Creating the Stencil
	hsh = { :from_specs => pseudoshapes }
	@stencil_cur = Stencil.new
	@stencil_cur.configure(:from_specs, false, hsh)
	history_store
end
	
def init_colors
	@color_hi_faces = 'thistle'
	@color_hi_edges = 'purple'
	@color_hi_comp = 'purple'
end

def init_cursors
	@id_cursor_pick = Traductor.create_cursor "Cursor_Stencil_Pick", 2, 2
end
	
#INIT: Initialize texts and messages
def init_messages
	#Text instruction for picking a stencil shape
	@ogl = Traductor::OpenGL_6.new
	bk_color = 'yellow'
	fr_color = 'blue'
	size = 32
	fac = 2
	pts = G6.pts_rectangle 0, -0.5, size, size
	@pick_instructions_icon = [[GL_POLYGON, pts, bk_color], [GL_LINE_LOOP, pts, fr_color, 2, '']]
	@pick_instructions_icon += @ogl.draw_proc(:std_pick_shape, size, size)
	text = T6[:TIT_StencilManager_TopMessage]
	x = fac * size + 4
	y = 2
	@pick_instructions = G6.rectangle_text_instructions(text, x, y, bk_color, fr_color, 0, 2, 0)
	ts = Geom::Transformation.scaling ORIGIN, fac, -fac, 1
	tt = Geom::Transformation.translation Geom::Vector3d.new(0, fac * size, 0)
	@tr_instructions = tt * ts
	@status_message = T6[:TIT_StencilManager_StatusMessage]
end
	
def show_message
	Sketchup.set_status_text "", SB_VCB_LABEL
	Sketchup.set_status_text "", SB_VCB_VALUE
	Sketchup.set_status_text @status_message
end

#----------------------------------------------------------------------------------------------------
# HISTORY: Call back from Otpicker
#----------------------------------------------------------------------------------------------------

#HISTORY: Store the current stencil in history
def history_store
	@@history = [] unless @@history
	ls = []
	@@history.each do |stencil|
		ls.push stencil unless stencil == @stencil_cur
	end
	ls.unshift @stencil_cur
	@@history = ls[0..@history_keep]
	@@ipos_history = 0
	@stencil_cur
end
	
#HISTORY: Navigate in history	
def history_navigate(incr, beg_end=false)
	n = @@history.length
	return if n == 0
	if beg_end
		@@ipos_history = (incr > 0) ? 0 : n - 1
	else
		@@ipos_history = (@@ipos_history - incr).modulo(n)
	end	
	set_current_stencil @@history[@@ipos_history]
end

#HISTORY: Check if history can be used
def history_enabled?
	@@history && @@history.length > 1
end
	
#----------------------------------------------------------------------------------------------------
# NOTIFY: Call back from Otpicker
#----------------------------------------------------------------------------------------------------

def notify_from_otpicker(event)
	case event
	when :shift_down
		@keyman.shift_down?
	when :ctrl_down	
		@keyman.ctrl_down?
	when :onMouseMove_zero
		onMouseMove_zero
	else
		:no_event
	end	
end
	
#NOTIFY: Call the notification method of the caller class	
def notify_caller(event, default=nil)
	return default unless @notify_caller_proc
	val = @notify_caller_proc.call(event)
	return default if val == :no_event
	val
end
	
#---------------------------------------------------------------------------------------------
# MODE: Manage the picking mode
#---------------------------------------------------------------------------------------------

def picking_toggle
	@picking_mode = !@picking_mode
	picking_after_change
end

def picking_set(mode)
	@picking_mode = mode
	picking_after_change
end

def picking_after_change	
	if @picking_mode
		notify_caller :enter_picking
	else
		notify_caller :exit_picking
	end	
end

def picking?
	@picking_mode
end

#ACTIVATION: Reset the picking environment
def reset_picking
	@mode = @otpicker.set_mode(:origin)
	@pt_origin = @pt_target = nil
	@otpicker.reset
	@keyman.reset
	@dragging = false
end	

#---------------------------------------------------------------------------------------------
# PICKING: Picking Stencil from model
#---------------------------------------------------------------------------------------------

#VIEWPORT: Computing Current cursor
def onSetCursor
	UI.set_cursor @id_cursor_pick
end

#MOVE: Mouse Movements
def onMouseMove_zero(flag=nil) ; onMouseMove(flag, @xmove, @ymove, @view) if @xmove ; end
def onMouseMove(flags, x, y, view)
	#Current mouse coordinates
	@xmove = x
	@ymove = y

	#Dispatching the event
	case @mode
	
	#Pick origin with inference
	when :origin	
		@origin_info = @otpicker.origin_pick view, x, y, false, @shift_down, true
		@pt_origin, @type_origin = @origin_info
		repere_compute_plane
		capture_highlighted
			
	#Tracking the target	
	when :target
		@target_info = @otpicker.target_pick(view, x, y, false)
		@pt_target, = @target_info
		repere_dragging if @pt_origin != @pt_target

	end
	
	show_message
end	

#---------------------------------------------------------------------------------------------
# CLICK: Click Management
#---------------------------------------------------------------------------------------------

#CLICK: Button click DOWN
def onLButtonDown(flags, x, y, view)
	#Storing the reference
	@button_down = true
	@xdown = x
	@ydown = y	
	
	#Changing mode
	if @mode == :origin
		onMouseMove_zero unless @pt_origin
		repere_start
	elsif @mode == :target
		#repere_stop
	end	
	@time_down_start = Time.now
end

#CLICK: Button click UP - Means that we end the selection
def onLButtonUp(flags, x, y, view)
	return unless @button_down
	@button_down = false
	
	#Logic for dragging
	far_from_origin = (@xdown - x).abs > 3 || (@ydown - y).abs > 3
	if @time_down_start && (Time.now - @time_down_start) > @duration_long_click && !far_from_origin
		@otpicker.direction_pick_from_model true
		repere_compute_plane
	elsif @dragging
		if @time_down_start && (Time.now - @time_down_start) > 0.3 && ((@xdown - x).abs > 1 || (@ydown - y).abs > 1)
			repere_stop
		end
	else
		repere_stop
	end	
end

#CLICK: Double Click received
def onLButtonDoubleClick(flags, x, y, view)
	exit	
end

#---------------------------------------------------------------------------------------------
# DRAW: Drawing Methods
#---------------------------------------------------------------------------------------------

#DRAW: Draw top method
def draw(view)	
	draw_hi_faces view
	draw_hi_comp view
	draw_repere view
	@otpicker.draw_all view
	draw_message_stencil_picking view
end

#DRAW: Draw a message for Stencil picking
def draw_message_stencil_picking(view)
	@ogl.process_draw_GL view, @tr_id, @pick_instructions
	@ogl.process_draw_GL view, @tr_instructions, @pick_instructions_icon
end

#DRAW: Draw the axes of the repere
def draw_repere(view)
	return unless @pt_origin && @repere_axes 
	size = view.pixels_to_model 75, @pt_origin
	view.line_width = 2
	view.line_stipple = ''
	vecx, vecy, normal = @repere_axes
	[[normal, 'blue'], [vecx, 'red'], [vecy, 'green']].each do |vec, color|
		view.drawing_color = color
		pt2 = @pt_origin.offset vec, size
		view.draw2d GL_LINE_STRIP, [@pt_origin, pt2].collect { |pt| view.screen_coords pt }
		view.line_width = 1
	end	
end

#DRAW: Draw the highlighted face or surface
def draw_hi_faces(view)
	if G6.su_capa_color_polygon && @hi_face_polys && @hi_face_polys.length > 0
		view.drawing_color = @color_hi_faces
		view.draw GL_TRIANGLES, @hi_face_polys
	end	
	
	if @hi_edge_lines && @hi_edge_lines.length > 0
		view.line_width = 2
		view.line_stipple = ''
		view.drawing_color = @color_hi_edges
		view.draw GL_LINES, @hi_edge_lines.collect { |pt| G6.small_offset(view, pt) }
	end	
end
	
def draw_hi_comp(view)
	return unless @hi_comp
	comp, tr = @hi_comp
	view.line_width = 3
	view.line_stipple = ''
	view.drawing_color = @color_hi_comp
	G6.draw_component(view, comp, tr)
end
	
#---------------------------------------------------------------------------------------------
# KEY: Keyboard handling
#---------------------------------------------------------------------------------------------

#Return key pressed
def onReturn(view)
	@otpicker.direction_pick_from_model true
	repere_compute_plane
end

#KEY: Manage key events
def key_manage(event, key, view)
	case event
	
	#SHIFT key
	when :shift_down
		@otpicker.direction_changed
	when :shift_toggle
	
	when :shift_up
		
	#CTRL key	
	when :ctrl_down	
		onMouseMove_zero

	when :ctrl_up
		onMouseMove_zero

	when :ctrl_other_up
	
	#ALT key
	when :alt_down
		@otpicker.no_inference_toggle
	when :alt_up	
		@otpicker.no_inference_set false
	
	#Arrows	
	when :arrow_down	
		action_from_arrow key
		
	#Backspace	
	when :backspace
		@otpicker.inference_reset

	#TAB	
	when :tab
		
	end
end

#KEY: Handle Key down
def onKeyDown(key, rpt, flags, view)
	@keyman.onKeyDown(key, rpt, flags, view)
end

#KEY: Handle Key up
def onKeyUp(key, rpt, flags, view)
	@keyman.onKeyUp(key, rpt, flags, view)
end

#KEY: Manage parameters from the arrow keys
def action_from_arrow(key)	
	#Computing the new code
	case key
	when VK_UP
		code = :follow_z
	when VK_LEFT
		code = :follow_y
	when VK_RIGHT
		code = :follow_x
	when VK_DOWN
		code = :no
	end
	
	@otpicker.option_planar_set_dir code
	repere_compute_plane
end	

#---------------------------------------------------------------------------------------------
# REPERE: Manage the axes and coordinate framework for projection
#---------------------------------------------------------------------------------------------

#REPERE: Start picking
def repere_start
	@mode = @otpicker.set_mode(:target)
end

#REPERE: Finish picking the shape and exit if OK
def repere_stop
	@action_stopped = true
	repere_compute_plane
	status = capture_shape
	reset_picking
	exit if status
end

#REPERE: Dragging mode for picking to set direction of X axis
def repere_dragging
	@action_stopped = false
	repere_compute_plane
	@dragging = true
end

#REPERE: Compute the axes of the plane
def repere_compute_plane
	return unless @otpicker
	@repere_axes = nil
	normal = @otpicker.direction_plane_active
	return unless normal && @pt_origin
	
	#For component, take the Z axis of the component
	origin, type, info_comp = @origin_info
	elt, comp, tr = info_comp
	if comp && !@keyman.ctrl_down?
		normal = G6.transform_vector Z_AXIS, tr
	end
	
	#Reversing the normal to make it emerging from face
	pt2d = @view.screen_coords @pt_origin
	ray = @view.pickray pt2d.x, pt2d.y
	@repere_reverse = false
	if (normal % ray[1] > 0)
		normal = normal.reverse
		@repere_reverse = true
	end	
	
	#Computing the axes
	vecx = (@pt_target && @pt_target != @pt_origin) ? @pt_origin.vector_to(@pt_target) : normal.axes[0]
	vecy = normal * vecx
	vecx, vecy = normal.axes unless vecy.valid?
	@repere_axes = [vecx, vecy, normal]
	@repere_tr = Geom::Transformation.axes @pt_origin, *@repere_axes
end

#---------------------------------------------------------------------------------------------
# CAPTURE: Capture the shape for stencil
#---------------------------------------------------------------------------------------------

#CAPTURE: Compute the grouponent or faces to be highlighted for the potential stencil
def capture_highlighted
	origin, type, info_comp = @origin_info
	face0, comp, tr = @otpicker.origin_face_info
	@hi_comp = @hi_faces = nil
	@hi_face_polys = @hi_edge_lines = nil
	
	if comp && !@keyman.ctrl_down?
		@hi_comp = [comp, tr]
		return
	end
	
	ledges = []
	lfaces = []
	if face0
		hfaces = {}
		lfaces = (@rendering_options["DrawHidden"]) ? [face0] : G6.face_neighbours(face0, hfaces)
		hsh_edges_used = {}
		@hi_faces = [lfaces, tr]
		lfaces.each do |face|
			face.edges.each do |e|
				eid = e.entityID
				next if hsh_edges_used[eid]
				hsh_edges_used[eid] = true
				lf = e.faces.find_all { |f| hfaces[f.entityID] }
				ledges.push e if lf.length == 1
			end
		end	
	end

	#Calculate the polygons for drawing the highlighted faces
	@hi_face_polys = []
	lfaces.each do |face|
		mesh = face.mesh
		pts = mesh.points
		mesh.polygons.each { |p| @hi_face_polys += p.collect { |i| tr * pts[i.abs-1] } }
	end
	
	@hi_edge_lines = []
	ledges.each do |edge|
		@hi_edge_lines.push tr * edge.start.position, tr * edge.end.position
	end	
end

#CAPTURE: Capture the final shape and create the Stencil
def capture_shape
	stencil = nil
	if @hi_comp
		comp, tr = @hi_comp
		hsh = { :comp => comp, :repere_tr => @repere_tr, :tr_comp => tr }
		stencil = Stencil.new
		stencil.configure :from_comp, false, hsh
	elsif @hi_faces
		lst_faces, tr = @hi_faces
		hsh = { :lst_faces => lst_faces, :repere_tr => @repere_tr, :tr_faces => tr }
		stencil = Stencil.new
		stencil.configure :from_faces, false, hsh
	end	
	
	if stencil && stencil.valid?
		set_current_stencil stencil
		history_store
		return true
	end	
	false
end

#--------------------------------------------------
# Contextual menu
#--------------------------------------------------

#Contextual menu
def getMenu(cxmenu)			
	#Pick Selection Menu
	@otpicker.menu_contribution(cxmenu)
	
	#Abort
	cxmenu.add_sepa
	cxmenu.add_item(T6[:T_BUTTON_Cancel]) { picking_toggle }
end

#--------------------------------------------------------------
#--------------------------------------------------------------
# PALETTE: Palette Management
#--------------------------------------------------------------
#--------------------------------------------------------------

#PALETTE: Floating Palette
def palette_contribution(palette)
	@palette = palette

	#Callback procedures
	@draw_local = self.method "draw_button_opengl"
	draw_zone = self.method "instructions_for_palette"
	zone_proc = self.method "edition_dispatch_event"
	cursor_proc = self.method "edition_cursor"
	
	#Parameters for the layout
	bk_color = 'lightblue'
	hi_color = 'lightgreen'
	val_color = 'gold'
	tit_color = 'thistle'
		
	#Declare the floating palette
	@flpal = flpal = :pal_floating_param
	hsh = { :title => "Stencil Management" }
	@palette.declare_floating flpal, hsh
	
	#Parameters for the palette	
	row = 0
	
	#Selection buttons
	widtot = palette_stencil_select row
	
	#Drawing_zone
	row += 1
	@palette_zone_button = symb = "#{flpal}_drawing_zone".intern
	hidden_proc = proc { @glass_mode }
	zone_proc = self.method "edition_dispatch_event"
	cursor_proc = self.method "edition_cursor"
	hsh = { :floating => flpal, :passive => true, :height => widtot, :width => widtot, :row => row, :bk_color => 'white',
            :hidden_proc => hidden_proc, :draw_proc => draw_zone, 
			:draw_refresh => true, :zone_proc => zone_proc, :cursor_proc => cursor_proc }
	@palette.declare_button symb, hsh	
end

def palette_stencil_select(row)
	lst_buttons = [:pick, nil, :shape_geom, nil, :comp, nil, :in_model, :in_su, :favorite, :last]
	nbut = (lst_buttons.find_all { |a| a }).length
	nsepa = (lst_buttons.find_all { |a| !a }).length
	fsepa = 0.25
	wbut = 32 
	wsepa = fsepa * wbut
	widtot = nbut * wbut + nsepa * wsepa
	hshc = { :floating => @flpal, :row => row, :height => wbut }
	hsh_but = { :width => wbut, :hi_color => 'yellow' }
	hsh_sepa = { :width => wsepa, :draw_proc => :separator_V, :passive => true }
	
	isepa = 0
	lst_buttons.each do |symb|
		hsh = palette_stencil_select_proc_hsh(symb)
		if symb
			pal_symb = "pal_stencil_select_#{symb}".intern
			@palette.declare_button pal_symb, hshc, hsh_but, hsh
		else	
			pal_symb = "pal_stencil_select_sepa_#{isepa}".intern
			isepa += 1
			@palette.declare_button pal_symb, hshc, hsh_sepa
		end
	end	
	widtot
end

def palette_stencil_select_proc_hsh(symb)
	action_proc = proc { stencil_select_notify(:exec, symb) }
	tip_proc = proc { stencil_select_notify(:tip, symb) }
	grayed_proc = proc { stencil_select_notify(:grayed, symb) }
	value_proc = proc { stencil_select_notify(:value, symb) }
	case symb
	when :pick
		draw_proc = :std_pick_shape
	else	
		draw_proc = @draw_local
	end	
	
	{ :action_proc => action_proc, :value_proc => value_proc, :grayed_proc => grayed_proc, :draw_proc => draw_proc }
end

#PALETTE: Button for picking a stencil in model
def palette_button_pick_shape(palette, hsh_keys=nil)
	@palette = palette
	
	tx_key_pick = tx_key_nav = nil
	if hsh_keys
		tx_key_pick = hsh_keys[:pick]
		tx_key_nav = hsh_keys[:nav]
	end
	
	#Previous stencil button
	w = 16
	h = 32
	tip = T6[:TIP_PrevStencil]
	tip += ' ' + tx_key_nav if tx_key_nav
	hshb = { :width => w, :height => h, :main_color => 'lightblue', :frame_color => 'gray' }
	grayed_proc = proc { !history_enabled? }
	hsh = { :grayed_proc => grayed_proc, :tooltip => tip, :draw_proc => :triangle_W1P }
	palette.declare_button(:pal_stencil_man_stencil_prev, hshb, hsh) { history_navigate -1 }
	
	#Pick a stencil
	tip = T6[:TIP_PickStencil]
	tip += ' ' + tx_key_pick if tx_key_pick
	value_proc = proc { picking? }
	hsh = { :value_proc => value_proc, :tooltip => tip, :draw_proc => :std_pick_shape, :hi_color => 'yellow' }
	palette.declare_button(:pal_stencil_man_stencil_select, hsh) { picking_toggle }
	
	#Next Stencil button
	tip = T6[:TIP_NextStencil]
	tip += ' ' + tx_key_nav if tx_key_nav
	hsh = { :grayed_proc => grayed_proc, :tooltip => tip, :draw_proc => :triangle_E1P }
	palette.declare_button(:pal_stencil_man_stencil_next, hshb, hsh) { history_navigate +1 }
end

#PALETTE: Button for displaying the stencil
def palette_button_show_stencil(palette=nil)
	palette = @palette unless palette
	@draw_stencil = self.method "draw_stencil_opengl"
	bkcolor = 'lightblue'
	base = 128
	
	#Computing the dimension
	if @stencil_cur
		w, h = @stencil_cur.bbox_dimensions
		r = w / h
		if r > 1.0
			h = base
			w = [h * r, base * 2].min
			h = [w / r, base * 0.5].max
			w = h * r
		else
			w = base
			h = [w / r, base * 2].min
			w = [h * r, base * 0.5].max
			h = w / r
		end		
	else
		text = T6[:BOX_NoActiveStencil]	
		w, = G6.simple_text_size text
		w += 20
		h = 32
	end
	
	hsh = { :sidepal => nil, :passive => true, :tooltip => T6[:TIP_CurrentStencil], :height => h, :width => w,
	        :draw_proc => @draw_stencil, :bk_color => bkcolor, :text => text, :justif => 'MH' }
		
	@symb_button_stencil_show = :stencil_man_display_stencil
	palette.declare_button(@symb_button_stencil_show, hsh)
end

def stencil_select_notify(code, symb)
	case code
	when :grayed
		false
	when :exec
		case symb
		when :pick
			picking_toggle
		end	
	when :value
		case symb
		when :pick
			@picking_mode
		end	
	
	when :tip
		"Tip #{symb}"
	end	
end

#DRAW: generate instructions for the palette	
def instructions_for_palette(symb, dx, dy, main_color, frame_color, selected, grayed)
	ptmid = Geom::Point3d.new 0.5 * dx, 0.5 * dy, 0
	instructions = []
	
	#Geometry
	#instructions += @instructions2d

	#Final instruction for frame and decoration
	#instructions += @instructions2d_after

	#Superimposed instructions from tools
	#instructions += @edition_instructions if @edition_instructions

	lgl = []
	instructions.each do |a|
		code, pts, color, width, stipple = a
		pts = (code == 'T') ? t * pts : pts.collect { |pt| t * pt }
		lgl.push [code, pts, color, width, stipple]
	end
	lgl
end

#PALETTE: Custom drawing of stencil
def draw_stencil_opengl(symb, dx, dy, main_color, frame_color, selected, grayed)
	(@stencil_cur) ? @stencil_cur.bbox_instructions(dx, dy, main_color, frame_color) : []
end

#PALETTE: Custom drawing of buttons
def draw_button_opengl(symb, dx, dy, main_color, frame_color, selected, grayed)
	code = symb.to_s
	lst_gl = []
	dx2 = dx / 2
	dy2 = dy / 2
	color = (grayed) ? 'gray' : frame_color
	
	case code
	
	when /select_pick/
		#Drawing the shape
		color_shape = 'lightblue'
		w = 2 * dx / 3
		lpti = [[1, dy/3], [1, dy-4], [4, dy-1], [w, dy-1], [w, dy/3], [w/2, dy/6]]
		pts = lpti.collect { |x, y| Geom::Point3d.new(x, y, 0) }
		#pts = G6.pts_rectangle 1, dx/3, 2*dx/3, dx2-1
		lst_gl.push [GL_POLYGON, pts, color_shape, 1, '']	
		lst_gl.push [GL_LINE_LOOP, pts, 'black', 1, '']	
		
		#drawing the arrow
		ptmid = Geom::Point3d.new dx2, dy2
		trot = Geom::Transformation.rotation(ptmid, Z_AXIS, 30.degrees)
		tt = Geom::Transformation.translation Geom::Vector3d.new(3, 0, 0)
		t = tt * trot
		y = 4
		dec = dx2 / 2
		color_arrow = 'gray'
		lpti = [[dx2-dec, y], [dx2, y+dy2], [dx2+dec, y]]
		pts_rec = lpti.collect { |x, y| t * Geom::Point3d.new(x, y, 0) }
		w = 2
		lpti = [[dx2+w, y], [dx2+w, 0], [dx2-w, 0], [dx2-w, y] ]
		pts_tri = lpti.collect { |x, y| t * Geom::Point3d.new(x, y, 0) }
		lst_gl.push [GL_POLYGON, pts_tri, color_arrow]	
		lst_gl.push [GL_POLYGON, pts_rec, color_arrow]	
		poly = pts_tri + pts_rec
		lst_gl.push [GL_LINE_LOOP, poly, 'black', 1, '']	
		
		
	end	#case code
	
	lst_gl
end

#EDITION: Cursor depending on the edition tool
def edition_cursor
	0
end

def edition_dispatch_event(code, button_down, dx, dy, x, y)

end

end	#class StencilManager

end #module Traductor